/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

// BitStream.cpp: implementation of the CBitStream class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "BitStream.h"
#include "MGErrMsg.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CBitStream::CBitStream(CFile* stream, UINT o, int lim)
{
	op = o;
	bit_count = 0;
	m_stream = stream;
	limit = lim;
	limit_reached = false;

	if (op == OP_READ)
		m_bit = 0;
	else
		m_bit = 8;
	m_buffer = 0;
}

CBitStream::~CBitStream()
{
	Flush();
}

void CBitStream::Flush()
{
  // warning: Flush will alter the bitcount
  // if you would like to know the exact bitcount
  // you should call GetBitCount() before Flush()
  try
  {
	  if (op == OP_WRITE)
	  {
		  if (m_bit < 8)
		  {
        bit_count += m_bit;
			  m_stream->Write(&m_buffer, sizeof(m_buffer));
			  m_buffer = 0;		m_bit = 8;
		  }
	  }
	  else
	  {
		  ASSERT(op == OP_READ);
      bit_count += m_bit;
		  m_buffer = 0;		m_bit = 0;
      limit = -1;
	  }
  }
  catch (CFileException *e)
  {
    throw CMGDataException(e);
  }
}

void CBitStream::Reset()
{
  Flush();
  bit_count = 0;
  limit_reached = false;
}

BYTE CBitStream::GetRemainingValue() const
{
  if (op == OP_WRITE)
  {
    BYTE mask = 0xff - (1 << m_bit) - 1;
    return m_buffer & mask;
  }
  else
  {
		ASSERT(op == OP_READ);
    BYTE mask = (1 << m_bit) - 1;
    return m_buffer & mask;
  }
}

void CBitStream::SetBitPosition(BYTE pos)
{
  try
  {
		m_stream->Read(&m_buffer, sizeof(m_buffer));
    m_bit = pos;
  }
  catch (CFileException *e)
  {
    throw CMGDataException(e);
  }
}

void CBitStream::ReadBit(BYTE *res)
{
  try
  {
	  if (limit > 0 && bit_count >= limit)
		  throw EEC_ReadingLimitReached();

	  ++bit_count;

	  if (m_bit == 0)
	  {
		  m_stream->Read(&m_buffer, sizeof(m_buffer));
		  m_bit = 8;
	  }
	  *res = !!(m_buffer & (1 << --m_bit));
  }
  catch (CFileException *e)
  {
    throw CMGDataException(e);
  }
}

void CBitStream::WriteBit(BYTE bit)
{
	if (limit_reached) return;

  try
  {
	  m_buffer |= (bit << --m_bit);

	  if (m_bit == 0)
	  {
		  m_stream->Write(&m_buffer, sizeof(m_buffer));
		  m_buffer = 0;		m_bit = 8;
	  }

	  bit_count++;

	  if (limit > 0 && bit_count >= limit)
	  {
		  limit_reached = true;			Flush();
		  throw EEC_WritingLimitReached();
	  }
  }
  catch (CFileException *e)
  {
    throw CMGDataException(e);
  }
}

void CBitStream::ReadBits(BYTE *res, BYTE count)
{
	ASSERT(count <= 8 && count > 0);

  try
  {
	  BYTE result = 0;
	  BYTE size = count;
	  BYTE bres;

	  for (BYTE bit = 0; bit < size; bit++)
	  {
		  ReadBit(&bres);
		  result |= bres;
		  if (bit != size - 1) result <<= 1;
	  }
	  *res = result;
  }
  catch (CFileException *e)
  {
    throw CMGDataException(e);
  }
}

void CBitStream::WriteBits(BYTE value, BYTE count)
{
	ASSERT(count <= 8 && count > 0);

  try
  {
	  BYTE mask = 1 << (count - 1);
	  BYTE shift = count - 1;

	  for (BYTE bit = 0; bit < count; bit++)
	  {
		  WriteBit((value & mask) >> shift);
		  mask >>= 1;		shift--;
	  }
  }
  catch (CFileException *e)
  {
    throw CMGDataException(e);
  }
}

void CBitStream::ReadBits(WORD *res,BYTE count)
{
	ASSERT(count <= 16 && count > 0);

  try
  {
	  WORD result = 0;
	  BYTE size = count;
	  BYTE bres;

	  for (BYTE bit = 0; bit < size; bit++)
	  {
		  ReadBit(&bres);
		  result |= bres;
		  if (bit != size - 1) result <<= 1;
	  }
	  *res = result;
  }
  catch (CFileException *e)
  {
    throw CMGDataException(e);
  }
}

void CBitStream::WriteBits(WORD value, BYTE count)
{
	ASSERT(count <= 16 && count > 0);

  try
  {
	  WORD mask = 1 << (count - 1);
	  WORD shift = count - 1;

	  for (BYTE bit = 0; bit < count; bit++)
	  {
		  WriteBit((value & mask) >> shift);
		  mask >>= 1;		shift--;
	  }
  }
  catch (CFileException *e)
  {
    throw CMGDataException(e);
  }
}

void CBitStream::ReadBits(DWORD *res, BYTE count)
{
	ASSERT(count <= 32 && count > 0);

  try
  {
	  DWORD result = 0;
	  BYTE size = count;
	  BYTE bres;

	  for (BYTE bit = 0; bit < size; bit++)
	  {
		  ReadBit(&bres);
		  result |= bres;
		  if (bit != size - 1) result <<= 1;
	  }
	  *res = result;
  }
  catch (CFileException *e)
  {
    throw CMGDataException(e);
  }
}

void CBitStream::WriteBits(DWORD value, BYTE count)
{
	ASSERT(count <= 32 && count > 0);

  try
  {
	  DWORD mask = 1 << (count - 1);
	  DWORD shift = count - 1;

	  for (BYTE bit = 0; bit < count; bit++)
	  {
		  WriteBit((BYTE)((value & mask) >> shift));
		  mask >>= 1;		shift--;
	  }
  }
  catch (CFileException *e)
  {
    throw CMGDataException(e);
  }
}

void CBitStream::ReadData(BYTE *res)
{
  try
  {
	  BYTE result = 0;
	  BYTE size = 8 * sizeof(BYTE);
	  BYTE bres;

	  for (BYTE bit = 0; bit < size; bit++)
	  {
		  ReadBit(&bres);
		  result |= bres;
		  if (bit != size - 1) result <<= 1;
	  }
	  *res = result;
  }
  catch (CFileException *e)
  {
    throw CMGDataException(e);
  }
}

void CBitStream::WriteData(BYTE value)
{
  try
  {
	  BYTE size = 8 * sizeof(BYTE);
	  BYTE mask = 1 << (size - 1);
	  BYTE shift = size - 1;

	  for (BYTE bit = 0; bit < size; bit++)
	  {
		  WriteBit((value & mask) >> shift);
		  mask >>= 1;		shift--;
	  }
  }
  catch (CFileException *e)
  {
    throw CMGDataException(e);
  }
}

void CBitStream::ReadData(WORD *res)
{
  try
  {
	  WORD result = 0;
	  BYTE size = 8 * sizeof(WORD);
	  BYTE bres;

	  for (BYTE bit = 0; bit < size; bit++)
	  {
		  ReadBit(&bres);
		  result |= bres;
		  if (bit != size - 1) result <<= 1;
	  }
	  *res = result;
  }
  catch (CFileException *e)
  {
    throw CMGDataException(e);
  }
}

void CBitStream::WriteData(WORD value)
{
  try
  {
	  BYTE size = 8 * sizeof(WORD);
	  WORD mask = 1 << (size - 1);
	  BYTE shift = size - 1;

	  for (BYTE bit = 0; bit < size; bit++)
	  {
		  WriteBit((value & mask) >> shift);
		  mask >>= 1;		shift--;
	  }
  }
  catch (CFileException *e)
  {
    throw CMGDataException(e);
  }
}

void CBitStream::ReadData(DWORD *res)
{
  try
  {
	  DWORD result = 0;
	  BYTE size = 8 * sizeof(DWORD);
	  BYTE bres;

	  for (BYTE bit = 0; bit < size; bit++)
	  {
		  ReadBit(&bres);
		  result |= bres;
		  if (bit != size - 1) result <<= 1;
	  }
	  *res = result;
  }
  catch (CFileException *e)
  {
    throw CMGDataException(e);
  }
}

void CBitStream::WriteData(DWORD value)
{
  try
  {
	  BYTE size = 8 * sizeof(DWORD);
	  DWORD mask = 1 << (size - 1);
	  BYTE shift = size - 1;

	  for (BYTE bit = 0; bit < size; bit++)
	  {
		  WriteBit((BYTE)((value & mask) >> shift));
		  mask >>= 1;		shift--;
	  }
  }
  catch (CFileException *e)
  {
    throw CMGDataException(e);
  }
}
