/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

// MGGridCodec.h: interface for the CMGGridCodec class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_MGGRIDCODEC_H__85E0A823_97E8_4B6D_8F77_A9974CF76CD7__INCLUDED_)
#define AFX_MGGRIDCODEC_H__85E0A823_97E8_4B6D_8F77_A9974CF76CD7__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "BasicTypes.h"
#include "cmatrix.h"
#include "Codec3D.h"
#include "Wavelet3D.h"
#include "WaveletShort3D.h"
#include "WaveletSmooth3D.h"
#include "MGGrid.h"

class CBitStream;
class MGStreamHeader;

///////////////////////////////////////////////////////////////////////////////////////////////
// class CGridCodec

// the notations of the 8 subbands of the wavelet decomposition
//
//       W         HLL -------- HLH  
//     /           /|           /|   
//   /           /  |         /  |   
//   ----> U   LLL -------- LLH  |   
//  |           |   |        |   |   
//  |           |  HHL ------|- HHH  
//  V           |  /         |  /    
//              |/           |/
//             LHL -------- LHH

enum CodecSpecific {          // the 8 subbands of the wavelet decomposition
      LLL = 0,
      LLH = 1,
      LHL = 2,
      LHH = 3,
      HLL = 4,
      HLH = 5,
      HHL = 6,
      HHH = 7
};

template <class TYPE>
class CMGGridCodec
{
public:
  virtual ~CMGGridCodec() {};

  static int  GetNrLevelsG(const CPointUVW &nrSlices);
  static bool GetNrTiles(const CPointUVW &nrSlices, CPointUVW &szTile, CPointUVW &nrTiles);
  void  CreateTiles(int levelG, int levelM);

protected:
	void	CreateSubbands(CCubeIArray *subbands);
	int *	CreateSubbands(const CPointUVW &nrSlices, CPointUVW &sizeL, CPointUVW &sizeH, int &nrSubbands);

	void	CreateSubbandTiles(int levelG, const CPointUVW &nrSlices, CPointUVW szTile);
  void  MapROIToTiles(const Cube &roi, const CPointUVW &sizeL, const CPointUVW &sizeH, int *subbands, int nrSubbands, CCubeArray &cubes);

protected:
	CPointUVW 		  m_nrSlices;             // size of the grid
	CPointUVW		    m_dimWave;              // dimension of the wavelet buffer to optimise memory access
  int             m_nrLevelsM;            // number of resolution levels of the mesh
  CPointUVW       m_nrDecLevels;          // number of decomposition levels for each (u,v,w) dimension
	int             m_nrLevelsG;            // maximum number of resolution levels of the grid
	double				  m_cr;                   // compression ratio for coding the grid
	CMGGrid			  * m_pGrid;                // pointer to the grid object
	CTileArray    * m_subbandTiles;         // array of tiles for each subband of the wavelet image

  MGStreamDescr         * m_pStreamDescr; // pointer to the stream description object
  CMGStreamParser       * m_pParser;      // pointer to the parser object
	CCodec3D<TYPE>		    * m_pCodec;       // pointer to the grid codec object
	CWaveletCodec<TYPE>	  * m_pPrecode;     // pointer to the wavelet transform object
};

template <class TYPE>
int CMGGridCodec<TYPE>::GetNrLevelsG(const CPointUVW &nrSlices)
{
  CPointUVW decLevels;
  CWaveletCodec<TYPE>::GetNrDecLevels(nrSlices, decLevels);

  // The minimum grid size that will be encoded as subbands is MIN_SPLIT_SIZE = 9
  // The number of levels = 1 as long as the grid size is less than MIN_SPLIT_SIZE
  // independent of the number of decomposition levels.
  // In the case levels = 1 the grid will be encoded as one Tile
	return max(CWaveletCodec<TYPE>::GetMaxDecLevel(decLevels) - 1, 1);
}

template <class TYPE>
bool CMGGridCodec<TYPE>::GetNrTiles(const CPointUVW &nrSlices, CPointUVW &szTile, CPointUVW &nrTiles)
{
  // when the nr of slices is less than MIN_SPLIT_SIZE we do not apply the tiling
  // since it is not efficient; we encode it a single cube of size equal to nr slices
  if (nrSlices.u < MIN_SPLIT_SIZE &&
      nrSlices.v < MIN_SPLIT_SIZE &&
      nrSlices.w < MIN_SPLIT_SIZE)
  {
    nrTiles = 1;  szTile = nrSlices;   return false;
  }

  // compute the number of ROIs for each direction,
  // since ROI size might not be a multiple of nrSlices, the last ROI might be larger
  nrTiles.u = max(nrSlices.u / szTile.u, 1);
  nrTiles.v = max(nrSlices.v / szTile.v, 1);
  nrTiles.w = max(nrSlices.w / szTile.w, 1);

  return true;
}

template <class TYPE>
void CMGGridCodec<TYPE>::CreateSubbands(CCubeIArray *subbands)
{
  const Z_SPLIT = 4;
  const Y_SPLIT = 2;
  const X_SPLIT = 1;

	CPointUVW decLevels = m_nrDecLevels;

  // find the number of subbands for each decomposition level
	for (int level = m_nrLevelsG - 1; level >= 1; level--)
	{
		int levelCounter = 1;
		if (--decLevels.w >= 0)		levelCounter <<= 1;
		if (--decLevels.v >= 0)		levelCounter <<= 1;
		if (--decLevels.u >= 0)		levelCounter <<= 1;

		// subtract the cube that will be further subdived
		subbands[level].SetSize(levelCounter - 1);
	}

  // first subband will always be coded as one cube. 
	subbands[0].SetSize(1);

	// temporary stacks of cubes that are subdivided
	int idx1 = 0, idx2 = 0;
	CubeI * stack1 = new CubeI[8];
	CubeI * stack2 = new CubeI[8];
	
	CubeI c(0, 0, 0, m_nrSlices.u, m_nrSlices.v, m_nrSlices.w, 0);
	stack1[idx1++] = c;

	decLevels = m_nrDecLevels;

	for (level = m_nrLevelsG - 1; level >= 1; level--)
	{
		if (--decLevels.w >= 0)
		{
			ASSERT(idx1 == 1 && idx2 == 0);

			CubeI tmp;
			const CubeI &cc = stack1[0];
			tmp.u = cc.u;			tmp.width  = cc.width;
			tmp.v = cc.v;			tmp.height = cc.height;

			int depth = (cc.depth + 1) >> 1;

			tmp.index = cc.index + Z_SPLIT;
      tmp.w = depth;		tmp.depth = cc.depth - depth;
			stack2[idx2++] =  tmp;

			tmp.index = cc.index;
			tmp.w = cc.w;			tmp.depth  = depth;
			stack2[idx2++] =  tmp;
			
			idx1 = idx2;		idx2 = 0;
			CubeI *swap = stack1;		stack1 = stack2;		stack2 = swap;
		}

		if (--decLevels.v >= 0)
		{
			ASSERT(idx2 == 0);

			for (int i = 0; i < idx1; i++)
			{
				CubeI tmp;
				const CubeI &cc = stack1[i];
				tmp.w = cc.w;			tmp.depth  = cc.depth;
				tmp.u = cc.u;			tmp.width  = cc.width;

				int height = (cc.height + 1) >> 1;

			  tmp.index = cc.index + Y_SPLIT;
				tmp.v = height;		tmp.height = cc.height - height;
				stack2[idx2++] =  tmp;

			  tmp.index = cc.index;
				tmp.v = cc.v;			tmp.height = height;
				stack2[idx2++] =  tmp;
			}
			
			idx1 = idx2;		idx2 = 0;
			CubeI *swap = stack1;		stack1 = stack2;		stack2 = swap;
		}

		if (--decLevels.u >= 0)
		{
			ASSERT(idx2 == 0);

			for (int i = 0; i < idx1; i++)
			{
				CubeI tmp;
				const CubeI &cc = stack1[i];
				tmp.v = cc.v;			tmp.height = cc.height;
				tmp.w = cc.w;			tmp.depth  = cc.depth;

				int width = (cc.width + 1) >> 1;

			  tmp.index = cc.index + X_SPLIT;
				tmp.u = width;		tmp.width = cc.width - width;
				stack2[idx2++] =  tmp;

			  tmp.index = cc.index;
				tmp.u = cc.u;			tmp.width = width;
				stack2[idx2++] =  tmp;
			}
			
			idx1 = idx2;		idx2 = 0;
			CubeI *swap = stack1;		stack1 = stack2;		stack2 = swap;
		}

		// copy the cubes to the list
		CCubeIArray &subBandLevel = subbands[level];
		for (int i = 0; i < idx1 - 1; i++)
		{
			subBandLevel[idx1 - i - 2] = stack1[i];
		}

		// subdivide only the first cube
		stack1[0] = stack1[idx1 - 1];		idx1 = 1;
	}

	// add the cube to the first subband
	subbands[0][0] = stack1[0];

  delete []stack1;		delete []stack2;
}

template <class TYPE>
int *CMGGridCodec<TYPE>::CreateSubbands(const CPointUVW &nrSlices, CPointUVW &sizeL, CPointUVW &sizeH, int &nrSubbands)
{
  const Z_SPLIT = 4;
  const Y_SPLIT = 2;
  const X_SPLIT = 1;

  sizeL = nrSlices, sizeH = 0;

	// temporary stacks of cubes that are subdivided
	int idx1 = 0, idx2 = 0;
	int * stack1 = new int[8];
	int * stack2 = new int[8];
	
	stack1[idx1++] = 0;

  int countSplit = 0;
	if (nrSlices.w >= MIN_DEC_SIZE)
	{
		stack2[idx2++] = 0;
		stack2[idx2++] = Z_SPLIT;

    sizeH.w = sizeL.w / 2;  sizeL.w -= sizeH.w;
		
    countSplit = 1;
		idx1 = idx2;		idx2 = 0;
		int *swap = stack1;		stack1 = stack2;		stack2 = swap;
	}

	if (nrSlices.v >= MIN_DEC_SIZE)
	{
		for (int i = 0; i < idx1; i++)
		{
			stack2[idx2++] = stack1[i];
			stack2[idx2++] = stack1[i] + Y_SPLIT;
		}
    sizeH.v = sizeL.v / 2;  sizeL.v -= sizeH.v;
		
    ++countSplit;
		idx1 = idx2;		idx2 = 0;
		int *swap = stack1;		stack1 = stack2;		stack2 = swap;
	}

	if (nrSlices.u >= MIN_DEC_SIZE)
	{
		for (int i = 0; i < idx1; i++)
		{
			stack2[idx2++] = stack1[i];
			stack2[idx2++] = stack1[i] + X_SPLIT;
		}
    sizeH.u = sizeL.u / 2;  sizeL.u -= sizeH.u;
		
    ++countSplit;
		idx1 = idx2;		idx2 = 0;
		int *swap = stack1;		stack1 = stack2;		stack2 = swap;
	}

  delete []stack2;

  // since the LLL subband will not be tiled we remove it from the list
  ASSERT(idx1 >= 2);
  for (int idx = 1; idx < idx1; idx++)
    stack1[idx - 1] = stack1[idx];

  nrSubbands = idx1 - 1;

  if (countSplit == 3)
  {
    // the subbands have to be arranged in the following order
    // in order to have efficient lossy encoding
    // LLH-LHL-HLL-LHH-HLH-HHL-HHH
    int tmp = stack1[2];  stack1[2] = stack1[3];  stack1[3] = tmp;
  }

  return stack1;
}

template <class TYPE>
void CMGGridCodec<TYPE>::CreateTiles(int levelG, int levelM)
{
  const GStreamLevel &gridLevel = m_pStreamDescr->GetGridLevel(levelG);
  const CPointUVW &nrTiles     = gridLevel.GetNrTiles();

  CPointUVW nrSlices = m_pStreamDescr->GetMaxNrSlices();
  CMGGrid::ComputeNrSlicesDown(nrSlices, m_nrLevelsG - 1 - levelG);

  CPointUVW szTile = nrSlices / nrTiles;
  if (szTile.u == 0 || szTile.v == 0 || szTile.w == 0)
    throw CMGDataException(CMGErrMsg::WRONG_NR_TILES);

  CreateSubbandTiles(levelG, nrSlices, szTile);
}

template <class TYPE>
void CMGGridCodec<TYPE>::CreateSubbandTiles(int levelG, const CPointUVW &nrSlices, CPointUVW szTile)
{
  GStreamLevel &gridLevel = m_pStreamDescr->GetGridLevel(levelG);

  // allocate the Tiles array
  CTileArray &tiles = m_subbandTiles[levelG];

  if (!gridLevel.CanSplit())
  {
    // if nrSlices is less or equalt to MIN_SPLIT_SIZE than
    // the cube level will not be split into subbands, it will be coded as it is
    ASSERT(gridLevel.GetTotalNrTiles() == 1);

    Cube cube(0, 0, 0, nrSlices.u, nrSlices.v, nrSlices.w);
    tiles.SetSize(1);   tiles[0].SetSize(1);
    tiles[0][0] = cube;
    return;
  }
  else
  {
    // the cube level is split into the following subbands
    int nrSubbands;
    CPointUVW sizeL, sizeH;
    int *subbands = CreateSubbands(nrSlices, sizeL, sizeH, nrSubbands);

    // each subband is split into tiles
    tiles.SetSize(gridLevel.GetTotalNrTiles());

    // update the tile size
    szTile.u = min(szTile.u, nrSlices.u);
    szTile.v = min(szTile.v, nrSlices.v);
    szTile.w = min(szTile.w, nrSlices.w);

    // take the Tiles one by one and determine the cubes
    // from every subband belonging to each Tile
    Cube roi;
    CMGMesh::InitFirstROI(roi, nrSlices, szTile);

    int index = 0;
    do
    {
      CCubeArray &cubes = tiles[index++];
      cubes.SetSize(nrSubbands);

      MapROIToTiles(roi, sizeL, sizeH, subbands, nrSubbands, cubes);
    }
    while (CMGMesh::GetNextROI(roi, nrSlices, szTile));

    delete []subbands;
  }
}

template <class TYPE>
void CMGGridCodec<TYPE>::MapROIToTiles(const Cube &roi, const CPointUVW &sizeL, const CPointUVW &sizeH, int *subbands, int nrSubbands, CCubeArray &cubes)
{
  for (int index = 0; index < nrSubbands; index++)
  {
    Cube tile;

    switch (subbands[index])
    {
    case LLL: AfxMessageBox("error");
    case LLH: // filter on X
      tile.u      =  sizeL.u       + ((roi.u % 2 == 0) ? roi.u / 2: (roi.u - 1) / 2);
      tile.v      = (sizeH.v != 0) ? ((roi.v % 2 == 0) ? roi.v / 2: (roi.v + 1) / 2) : roi.v;
      tile.w      = (sizeH.w != 0) ? ((roi.w % 2 == 0) ? roi.w / 2: (roi.w + 1) / 2) : roi.w; 

      tile.width  =                   (roi.u % 2 == 0) ?  roi.width / 2      : (roi.width + 1) / 2;
      tile.height = (sizeH.v != 0) ? ((roi.v % 2 == 0) ? (roi.height + 1) / 2:  roi.height / 2) : roi.height;
      tile.depth  = (sizeH.w != 0) ? ((roi.w % 2 == 0) ? (roi.depth  + 1) / 2:  roi.depth  / 2) : roi.depth;
      break;
    case LHL: // filter on Y
      tile.u      = (sizeH.u != 0) ? ((roi.u % 2 == 0) ? roi.u / 2: (roi.u + 1) / 2) : roi.u;
      tile.v      =  sizeL.v       + ((roi.v % 2 == 0) ? roi.v / 2: (roi.v - 1) / 2); 
      tile.w      = (sizeH.w != 0) ? ((roi.w % 2 == 0) ? roi.w / 2: (roi.w + 1) / 2) : roi.w; 

      tile.width  = (sizeH.u != 0) ? ((roi.u % 2 == 0) ? (roi.width + 1)  / 2:  roi.width  / 2) : roi.width;
      tile.height =                   (roi.v % 2 == 0) ?  roi.height / 2     : (roi.height + 1) / 2;
      tile.depth  = (sizeH.w != 0) ? ((roi.w % 2 == 0) ? (roi.depth  + 1) / 2:  roi.depth  / 2) : roi.depth;
      break;
    case HLL: // filter on Z
      tile.u      = (sizeH.u != 0) ? ((roi.u % 2 == 0) ? roi.u / 2: (roi.u + 1) / 2) : roi.u;
      tile.v      = (sizeH.v != 0) ? ((roi.v % 2 == 0) ? roi.v / 2: (roi.v + 1) / 2) : roi.v; 
      tile.w      =  sizeL.w       + ((roi.w % 2 == 0) ? roi.w / 2: (roi.w - 1) / 2); 

      tile.width  = (sizeH.u != 0) ? ((roi.u % 2 == 0) ? (roi.width + 1)  / 2:  roi.width  / 2) : roi.width;
      tile.height = (sizeH.v != 0) ? ((roi.v % 2 == 0) ? (roi.height + 1) / 2:  roi.height / 2) : roi.height;
      tile.depth  =                   (roi.w % 2 == 0) ?  roi.depth / 2      : (roi.depth  + 1) / 2;
      break;
    case LHH: // filter on X and Y
      tile.u      =  sizeL.u       + ((roi.u % 2 == 0) ? roi.u / 2: (roi.u - 1) / 2);
      tile.v      =  sizeL.v       + ((roi.v % 2 == 0) ? roi.v / 2: (roi.v - 1) / 2); 
      tile.w      = (sizeH.w != 0) ? ((roi.w % 2 == 0) ? roi.w / 2: (roi.w + 1) / 2) : roi.w; 

      tile.width  =                   (roi.u % 2 == 0) ?  roi.width  / 2     : (roi.width  + 1) / 2;
      tile.height =                   (roi.v % 2 == 0) ?  roi.height / 2     : (roi.height + 1) / 2;
      tile.depth  = (sizeH.w != 0) ? ((roi.w % 2 == 0) ? (roi.depth  + 1) / 2:  roi.depth  / 2) : roi.depth;
      break;
    case HLH: // filter on X and Z
      tile.u      =  sizeL.u       + ((roi.u % 2 == 0) ? roi.u / 2: (roi.u - 1) / 2);
      tile.v      = (sizeH.v != 0) ? ((roi.v % 2 == 0) ? roi.v / 2: (roi.v + 1) / 2) : roi.v; 
      tile.w      =  sizeL.w       + ((roi.w % 2 == 0) ? roi.w / 2: (roi.w - 1) / 2); 

      tile.width  =                   (roi.u % 2 == 0) ?  roi.width  / 2     : (roi.width  + 1) / 2;
      tile.height = (sizeH.v != 0) ? ((roi.v % 2 == 0) ? (roi.height + 1) / 2:  roi.height / 2) : roi.height;
      tile.depth  =                   (roi.w % 2 == 0) ?  roi.depth / 2      : (roi.depth  + 1) / 2;
      break;
    case HHL: // filter on Y and Z
      tile.u      = (sizeH.u != 0) ? ((roi.u % 2 == 0) ? roi.u / 2: (roi.u + 1) / 2) : roi.u;
      tile.v      =  sizeL.v       + ((roi.v % 2 == 0) ? roi.v / 2: (roi.v - 1) / 2); 
      tile.w      =  sizeL.w       + ((roi.w % 2 == 0) ? roi.w / 2: (roi.w - 1) / 2); 

      tile.width  = (sizeH.u != 0) ? ((roi.u % 2 == 0) ? (roi.width + 1)  / 2:  roi.width  / 2) : roi.width;
      tile.height =                   (roi.v % 2 == 0) ?  roi.height / 2     : (roi.height + 1) / 2;
      tile.depth  =                   (roi.w % 2 == 0) ?  roi.depth  / 2     : (roi.depth  + 1) / 2;
      break;
    case HHH: // filter on X and Y and Z
      tile.u      = sizeL.u + ((roi.u % 2 == 0) ? roi.u / 2: (roi.u - 1) / 2);
      tile.v      = sizeL.v + ((roi.v % 2 == 0) ? roi.v / 2: (roi.v - 1) / 2); 
      tile.w      = sizeL.w + ((roi.w % 2 == 0) ? roi.w / 2: (roi.w - 1) / 2); 

      tile.width  =           (roi.u % 2 == 0) ?  roi.width  / 2     : (roi.width  + 1) / 2;
      tile.height =           (roi.v % 2 == 0) ?  roi.height / 2     : (roi.height + 1) / 2;
      tile.depth  =           (roi.w % 2 == 0) ?  roi.depth  / 2     : (roi.depth  + 1) / 2;
      break;
    default: ASSERT(0);
    }

    cubes[index] = tile;
  }
}

///////////////////////////////////////////////////////////////////////////////////////////////
// class CMGGridDecoder

template <class TYPE>
class CMGGridDecoder : public CMGGridCodec<TYPE>
{
public:
	CMGGridDecoder(CMGStreamParser *pParser, CMGGrid *grid, float cr);
	virtual ~CMGGridDecoder();

	// implementation
public:
	void	RecoverLevel(int level);								            // read and decompress a resolution level
  void  UpdateGridCorners();
  bool  RecoverTile(int tileIdx, int levelGrid);
  void  GetTilesForROI(int roiIdx, int levelMesh, CIntArray *tiles);

protected:
  void  DecodeSubband(int subLevel, bool &bModified);
  void  DecodeTile(int dir, int subLevel, int tileIdx, bool &bModified);
	void	InitialiseMeshGrid();
	void	UpdateMeshGrid();

protected:
	CMatrix<TYPE>     m_wavelet[3];
	CMatrix<TYPE>     m_image[3];
};

// Construction/Destruction

template <class TYPE>
CMGGridDecoder<TYPE>::CMGGridDecoder(CMGStreamParser *pParser, CMGGrid *grid, float cr)
{
	m_pGrid = grid;
  m_pParser = pParser;
  m_subbandTiles = NULL;
  m_pStreamDescr = &pParser->m_streamDescr;

	m_cr = cr;
	m_nrSlices = grid->GetMaxNrSlices();
  m_nrLevelsM = grid->GetMaxNrLevels();

	// initialise the precoder
  switch (m_pStreamDescr->GetFilterType())
  {
  case CMGGrid::LINEAR_FILTER:
  	m_pPrecode = new CWaveletCodecShort<TYPE>(m_nrSlices, pParser->m_streamDescr.GetCyclicMode(), true);
    break;
  case CMGGrid::SMOOTH_FILTER:
  	m_pPrecode = new CWaveletCodecSmooth<TYPE>(m_nrSlices, pParser->m_streamDescr.GetCyclicMode(), true);
    break;
  }
	m_pPrecode->GetEfficientDim(m_nrSlices, m_dimWave);

	// Subtract 1 since the first and second decomposition level are coded together.
  // They are too small to be coded separately or as tiles.
	m_nrDecLevels = m_pPrecode->GetDecLevels();
	m_nrLevelsG = max(m_pPrecode->GetMaxDecLevel(m_nrDecLevels) - 1, 1);

	// initialise of the stream
	m_pCodec = new CCodec3D<TYPE>(m_nrSlices, m_nrLevelsG, true);

	// initialise the reference grid decoder
	m_pCodec->InitializeDecoder();

  for (int dir = X_GRID; dir <= Z_GRID; dir++)
  {
	  m_wavelet[dir].Init(m_dimWave.u, m_dimWave.v, m_dimWave.w);
    m_image[dir].Init(m_dimWave.u, m_dimWave.v, m_dimWave.w);
  }

  // create the tiles array
	m_subbandTiles = new CTileArray[m_nrLevelsG];
}

template <class TYPE>
CMGGridDecoder<TYPE>::~CMGGridDecoder()
{
	delete m_pCodec;	        m_pCodec = NULL;
	delete m_pPrecode;		    m_pPrecode = NULL;
  delete []m_subbandTiles;  m_subbandTiles = NULL;
}

template <class TYPE>
void CMGGridDecoder<TYPE>::DecodeSubband(int subLevel, bool &bModified)
{
  // decode the (x, y, z) grid coordinates
  for (int dir = X_GRID; dir <= Z_GRID; dir++)
  {
    const CTileArray &tiles = m_subbandTiles[subLevel];
    for (int tileIdx = 0; tileIdx < tiles.GetSize(); tileIdx++)
      DecodeTile(dir, subLevel, tileIdx, bModified);
  }
	m_pPrecode->WaveletRec3D(m_wavelet[0], m_wavelet[1], m_wavelet[2],
                           m_image[0]  , m_image[1]  , m_image[2]);
}

template <class TYPE>
void CMGGridDecoder<TYPE>::DecodeTile(int dir, int subLevel, int tileIdx, bool &bModified)
{
  CMGStreamParser::QueryType query;
  switch (dir)
  {
  case X_GRID:  query = CMGStreamParser::Q_GRIDX;   break;
  case Y_GRID:  query = CMGStreamParser::Q_GRIDY;   break;
  case Z_GRID:  query = CMGStreamParser::Q_GRIDZ;   break;
  }

  GStreamLevel &gridLevel = m_pStreamDescr->GetGridLevel(subLevel);
  int threshold = gridLevel.GetThreshold(dir);
  RGCoding &paramRG  = m_pStreamDescr->GetGridCodingParam(dir);
  const CCubeArray &cubes = m_subbandTiles[subLevel][tileIdx];

  // tell the parser that the decoding of a new grid tile starts
  // the parser should set the stream position for for each tile
  bool bTileExists = m_pParser->OnBeginGridDecodingTile(query, subLevel, tileIdx);
  if (bTileExists)
  {
    m_pCodec->ReadROIs(m_pParser->GetStream(), cubes, m_wavelet[dir],
                       threshold, paramRG.scale, true);

    // if the Tile is the entire grid, there was no splitting because the 
    // size of the grid is too low, write back the corners before applying the wavelet transform
    if (!gridLevel.CanSplit())
    {
      m_pCodec->RetrieveDC(m_wavelet[dir], paramRG);
    }

//	  m_pPrecode->WaveletRec3D(m_wavelet[dir], m_image[dir]);

		bModified = true;
  }

  gridLevel.SetStreamedTile(dir, tileIdx, true);

  // tell the parser that the decoding of a new grid tile ends
  // the parser should update the bit counter for the decoded tile
  m_pParser->OnEndGridDecodingTile(subLevel);
}

template <class TYPE>
bool CMGGridDecoder<TYPE>::RecoverTile(int tileIdx, int levelGrid)
{
  // decode the (x, y, z) grid coordinates
  ASSERT(levelGrid < m_nrLevelsG);
	bool bModified = false;

  for (int dir = X_GRID; dir <= Z_GRID; dir++)
    DecodeTile(dir, levelGrid, tileIdx, bModified);

	// update the reference grid
	if (bModified)
  {
    m_pPrecode->WaveletRec3D(m_wavelet[0], m_wavelet[1], m_wavelet[2],
                             m_image[0]  , m_image[1]  , m_image[2]);
    InitialiseMeshGrid();
  }

  return bModified;
}

template <class TYPE>
void CMGGridDecoder<TYPE>::GetTilesForROI(int roiIdx, int levelMesh, CIntArray *tiles)
{
  const MStreamLevel &meshLevel = m_pStreamDescr->GetMeshLevel(levelMesh);

  // retrieve the position of the ROI based on the index
  Cube roi;
  CPointUVW nrROIs = meshLevel.GetNrROIs();
  CMGMesh::MapIndexToPosition(roiIdx, nrROIs, roi);

  int endLevel = levelMesh + m_nrLevelsG - m_nrLevelsM;
  for (int subLevel = 0; subLevel <= endLevel; subLevel++)
  {
    GStreamLevel &gridLevel     = m_pStreamDescr->GetGridLevel(subLevel);
    const CPointUVW &nrTiles   = gridLevel.GetNrTiles();
    const CPointUVW &tileGroup = gridLevel.GetTileGrouping();
    
    CPointUVW tilePos = roiPos * nrTiles / nrROIs;
    ASSERT (nrTiles.u != 1 || tilePos.u == 0);
    ASSERT (nrTiles.v != 1 || tilePos.v == 0);
    ASSERT (nrTiles.w != 1 || tilePos.w == 0);

    int tileIdx = tilePos.w * (nrTiles.u * nrTiles.v) +
                  tilePos.v * (nrTiles.u) +
                  tilePos.u;
    if (!gridLevel.IsTileStreamed(X_GRID, tileIdx) ||
        !gridLevel.IsTileStreamed(Y_GRID, tileIdx) ||
        !gridLevel.IsTileStreamed(Z_GRID, tileIdx))
      tiles[subLevel].Add(tileIdx);
  }
}

template <class TYPE>
void CMGGridDecoder<TYPE>::RecoverLevel(int levelG)
{
  ASSERT(levelG < m_nrLevelsG);
	bool modified = false;

  DecodeSubband(levelG, modified);

	// update the reference grid
//	if (modified)   UpdateMeshGrid();
	if (modified)   InitialiseMeshGrid();
}

template <class TYPE>
void CMGGridDecoder<TYPE>::UpdateGridCorners()
{
  RGCoding *paramRG  = m_pStreamDescr->GetGridCodingParam();
  for (int dir = X_GRID; dir <= Z_GRID; dir++)
  {
    m_wavelet[dir] = 0;
    m_pCodec->RetrieveDC(m_wavelet[dir], paramRG[dir]);
  }
	m_pPrecode->WaveletRec3D(m_wavelet[0], m_wavelet[1], m_wavelet[2],
                           m_image[0]  , m_image[1]  , m_image[2]);

  // update the entire volume
  InitialiseMeshGrid();
}

template <class TYPE>
void CMGGridDecoder<TYPE>::UpdateMeshGrid()
{
	for (int w = 0; w < m_nrSlices.w; w++)
	{
		GridUV &gridUV = (*m_pGrid)[w];
		TYPE **sliceX = m_image[X_GRID][w];
		TYPE **sliceY = m_image[Y_GRID][w];
		TYPE **sliceZ = m_image[Z_GRID][w];

		for (int v = 0; v < m_nrSlices.v; v++)
		{
			GridU &gridU = gridUV[v];
			TYPE *elemX = sliceX[v];
			TYPE *elemY = sliceY[v];
			TYPE *elemZ = sliceZ[v];

			for (int u = 0; u < m_nrSlices.u; u++)
			{
				CFPoint3DT &p = gridU[u];
				p.x += (float)*elemX++;
				p.y += (float)*elemY++;
				p.z += (float)*elemZ++;
			}
		}
	}
}

template <class TYPE>
void CMGGridDecoder<TYPE>::InitialiseMeshGrid()
{
	for (int w = 0; w < m_nrSlices.w; w++)
	{
		GridUV &gridUV = (*m_pGrid)[w];
		TYPE **sliceX = m_image[X_GRID][w];
		TYPE **sliceY = m_image[Y_GRID][w];
		TYPE **sliceZ = m_image[Z_GRID][w];

		for (int v = 0; v < m_nrSlices.v; v++)
		{
			GridU &gridU = gridUV[v];
			TYPE *elemX = sliceX[v];
			TYPE *elemY = sliceY[v];
			TYPE *elemZ = sliceZ[v];

			for (int u = 0; u < m_nrSlices.u; u++)
			{
				CFPoint3DT &p = gridU[u];
				p.x = (float)*elemX++;
				p.y = (float)*elemY++;
				p.z = (float)*elemZ++;
			}
		}
	}	
}


///////////////////////////////////////////////////////////////////////////////////////////////
// class CMGGridCoder

template <class TYPE>
class CMGGridCoder : public CMGGridCodec<TYPE>
{
public:
	CMGGridCoder(CMGStreamParser *pParser, CMGGrid *grid, float cr, int nrQuantBits,
               const bool *bLODGrid, const CPointUVW *tileGroup, int filterType);
	virtual ~CMGGridCoder();

	// implementation
public:
	void	StreamHeader();							                        // preprocess the input data and call StreamRGHeader
	void	StreamLevel(int level);			                        // compress and write a resolution level
  void  InitExchangeObjectGrid(const bool *bLODGrid, const CPointUVW *tileGroup);
  void  InitExchangeObjectLevels(int levelG, CPointUVW &nrSlices, CPointUVW &szTile, const CPointUVW &szROI, const bool *bLODGrid, const CPointUVW *tileGroup);
  void  InitGridLevel(GStreamLevel &gridLevel, int levelG, bool bCanSplit,
                      const CPointUVW &nrTiles, double dim);

protected:
  void  EncodeSubband(int subLevel);
  void  EncodeTile(int dir, int subLevel, int tileIdx);

  void	ExtractCoordinates(CMatrix<TYPE> &coordX, CMatrix<TYPE> &coordY, CMatrix<TYPE> &coordZ);
	void	ComputeScale();

protected:
	int						m_nrBits;                       // number of quantisation bits for TYPE to int conversion
	CMatrix<int>  m_exchange[3];                  // buffers holding the quantised grid coefficients (x, y, z)
};

// Construction/Destruction

template <class TYPE>
CMGGridCoder<TYPE>::CMGGridCoder(CMGStreamParser *pParser, CMGGrid *grid, float cr, int nrQuantBits,
                                 const bool *bLODGrid, const CPointUVW *tileGroup, int filterType)
{
	m_pGrid = grid;
	m_pParser = pParser;
  m_subbandTiles = NULL;
  m_pStreamDescr = &pParser->m_streamDescr;

	m_cr = cr;
	m_nrBits = nrQuantBits;
	m_nrSlices = grid->GetMaxNrSlices();
  m_nrLevelsM = grid->GetMaxNrLevels();

  m_pStreamDescr->SetFilterType(filterType);

	// initialise the precoder
  switch (m_pStreamDescr->GetFilterType())
  {
  case CMGGrid::LINEAR_FILTER:
  	m_pPrecode = new CWaveletCodecShort<TYPE>(m_nrSlices, pParser->m_streamDescr.GetCyclicMode(), true);
    break;
  case CMGGrid::SMOOTH_FILTER:
  	m_pPrecode = new CWaveletCodecSmooth<TYPE>(m_nrSlices, pParser->m_streamDescr.GetCyclicMode(), true);
    break;
  }
	m_pPrecode->GetEfficientDim(m_nrSlices, m_dimWave);

	// Subtract 1 since the first and second decomposition level are coded together.
  // They are too small to be coded separately or as tiles.
	m_nrDecLevels = m_pPrecode->GetDecLevels();
	m_nrLevelsG = max(m_pPrecode->GetMaxDecLevel(m_nrDecLevels) - 1, 1);

	// initialise of the stream
	m_pCodec = new CCodec3D<TYPE>(m_nrSlices, m_nrLevelsG, true);

	// initialise the reference grid encoder
	m_pCodec->InitializeCoder();

  for (int dir = X_GRID; dir <= Z_GRID; dir++)
	  m_exchange[dir].Init(m_dimWave.u, m_dimWave.v, m_dimWave.w);

  // create the tiles array
  m_subbandTiles = new CTileArray[m_nrLevelsG];

  // initialise the exchange object with the grid coding details
  InitExchangeObjectGrid(bLODGrid, tileGroup);
}

template <class TYPE>
CMGGridCoder<TYPE>::~CMGGridCoder()
{
	delete m_pCodec;			    m_pCodec = NULL;
	delete m_pPrecode;		    m_pPrecode = NULL;
  delete []m_subbandTiles;  m_subbandTiles = NULL;
}

template <class TYPE>
void CMGGridCoder<TYPE>::InitExchangeObjectGrid(const bool *bLODGrid, const CPointUVW *tileGroup)
{
	TRY
	{
    m_pStreamDescr->InitGridLevels(m_nrLevelsG);

    m_pStreamDescr->SetNrQuantBitsGrid(m_nrBits);

    // intialise the levels information
    CPointUVW nrSlices, szTile;
    int endLevel = m_nrLevelsG - m_nrLevelsM;

    for (int level = 0; level < m_nrLevelsM; level++)
    {
      MStreamLevel &meshLevel = m_pStreamDescr->GetMeshLevel(level);
      const CPointUVW &szROI = meshLevel.GetROISize();

      if (level == 0)
      {
        for (int subLevel = 0; subLevel <= endLevel; subLevel++)
        {
          // compute the number of slices corresponding to sublevel
          nrSlices = m_pStreamDescr->GetMaxNrSlices();
          CMGGrid::ComputeNrSlicesDown(nrSlices, m_nrLevelsG - 1 - subLevel);

          // initialize the level information
          InitExchangeObjectLevels(subLevel, nrSlices, szTile, szROI, bLODGrid, tileGroup);

          // initalize the tiles for mesh level
          CreateSubbandTiles(subLevel, nrSlices, szTile);
        }
      }
      else
      {
        int subLevel = level + endLevel;
        if (subLevel < 0) continue;

        // compute number of slices corresponding to level
        nrSlices = m_pStreamDescr->GetMaxNrSlices();
        CMGGrid::ComputeNrSlicesDown(nrSlices, m_nrLevelsG - 1 - subLevel);

        InitExchangeObjectLevels(subLevel, nrSlices, szTile, szROI, bLODGrid, tileGroup);

        // initalize the tiles for mesh level
        CreateSubbandTiles(subLevel, nrSlices, szTile);
      }
    }
	}
	CATCH_ALL(e)
	{
		throw CMGDataException(e);
	}
	END_CATCH_ALL
}

template <class TYPE>
void CMGGridCoder<TYPE>::InitExchangeObjectLevels(int levelG, CPointUVW &nrSlices, CPointUVW &szTile, const CPointUVW &szROI, const bool *bLODGrid, const CPointUVW *tileGroup)
{
  GStreamLevel &gridLevel = m_pStreamDescr->GetGridLevel(levelG);

  if (bLODGrid[levelG])
  {
    // compute the number of tiles with grouping
    CPointUVW nrTiles;
    szTile = szROI * tileGroup[levelG];

    bool bCanSplit = GetNrTiles(nrSlices, szTile, nrTiles);

    double dim = szTile.u * szTile.v * szTile.w;
    InitGridLevel(gridLevel, levelG, bCanSplit, nrTiles, dim);
  }
  else
  {
    double dim = szROI.u * szROI.v * szROI.w;
    InitGridLevel(gridLevel, levelG, false, 1, dim);
  }
}

template <class TYPE>
void CMGGridCoder<TYPE>::InitGridLevel(GStreamLevel &gridLevel, int levelG, bool bCanSplit, const CPointUVW &nrTiles, double dim)
{
  if (!bCanSplit)
  {
    gridLevel.SetSplit(false);
    ASSERT(nrTiles == 1);
  }
  else
  {
    gridLevel.SetSplit(true);
  }

  // estimate the compression ratio for a level depending on the
  // max CR specified for the last level
  if (levelG > 0)
  {
	  double factor = sqrt(m_nrLevelsG - levelG);
	  double cr = pow(m_cr, 1.0 / factor);
	  int len = (int)(dim / cr + 0.5);

    // initialise the tiles arrays
    gridLevel.InitTiles(nrTiles, len);
  }
  else
  {
    // initialise the tiles arrays
    gridLevel.InitTiles(nrTiles, -1);
  }
}

template <class TYPE>
void CMGGridCoder<TYPE>::StreamHeader()
{
	// preproces the input buffers and compute the thresholds and mean
	// extract coordinates from the mesh
	CMatrix<TYPE>coord[3];
  for (int dir = X_GRID; dir <= Z_GRID; dir++)
	  coord[dir].Init(m_dimWave.u, m_dimWave.v, m_dimWave.w);

	ExtractCoordinates(coord[X_GRID], coord[Y_GRID], coord[Z_GRID]);

	// do the precoding, apply the scaling and convert from TYPE to int
  CIntArray thresholds;  thresholds.SetSize(m_nrLevelsG);
  RGCoding *paramRG  = m_pStreamDescr->GetGridCodingParam();

	m_pPrecode->WaveletDec3D(coord[0], coord[1], coord[2]);

  for (dir = X_GRID; dir <= Z_GRID; dir++)
  {
	  m_pCodec->PreProcess(m_subbandTiles, coord[dir], m_exchange[dir], paramRG[dir], thresholds, m_nrBits);
    m_pStreamDescr->SetThresholds(dir, thresholds);
  }
}

template <class TYPE>
void CMGGridCoder<TYPE>::EncodeSubband(int subLevel)
{
  // tell the parser that the coding of a new grid level starts
  // the parser should initialise a stream for each tile
  m_pParser->OnBeginGridCodingLevel(subLevel);

  const CTileArray &tiles = m_subbandTiles[subLevel];

  // encode the (x, y, z) grid coordinates
  for (int dir = X_GRID;  dir <= Z_GRID; dir++)
  {
    // tell the parser that the coding of a new grid coordinate starts
    m_pParser->OnBeginGridCodingDir(dir, subLevel);

    for (int tileIdx = 0; tileIdx < tiles.GetSize(); tileIdx++)
      EncodeTile(dir, subLevel, tileIdx);

    // tell the parser that the coding of a new grid coordinate ends
    m_pParser->OnEndGridCodingDir(dir);
  }

  // tell the parser that the coding of a new grid level ends
  // the parser should write the counters for each stream, and finally
  // the streams to the output stream
  m_pParser->OnEndGridCodingLevel(subLevel);
}

template <class TYPE>
void CMGGridCoder<TYPE>::EncodeTile(int dir, int subLevel, int tileIdx)
{
  GStreamLevel &gridLevel = m_pStreamDescr->GetGridLevel(subLevel);
  int threshold = gridLevel.GetThreshold(dir);
  RGCoding &paramRG  = m_pStreamDescr->GetGridCodingParam(dir);
  const CCubeArray &cubes = m_subbandTiles[subLevel][tileIdx];

  // tell the parser that the coding of a new grid tile starts
  // the parser should supply a new stream for each tile
  m_pParser->OnBeginGridCodingTile(dir);

  // perform the coding of the Tile
	m_pCodec->WriteROIs(m_pParser->GetStream(), cubes, m_exchange[dir], threshold);

  // tell the parser that the coding of a new grid tile ends
  // the parser should update the bit counter for the coded tile
  m_pParser->OnEndGridCodingTile(dir, subLevel, tileIdx);
}

template <class TYPE>
void CMGGridCoder<TYPE>::StreamLevel(int level)
{
  // Each mesh resolution has its corresponding grid resolution, with the exception
  // that the first resultion level mesh (depending on the number of slices for that level)
  // may have several grid resolutions. For efficiency reasons first and second resolution level
  // of the grid are coded as one tile. For this reason we particular cases as follows:
  
  // TO DO the parser should reset the file pointers
  ASSERT(level < m_nrLevelsM);

  int endLevel = m_nrLevelsG - m_nrLevelsM;
  if (level == 0)
  {
    for (int subLevel = 0; subLevel <= endLevel; subLevel++)
      EncodeSubband(subLevel);
  }
  else
  {  
    int subLevel = level + endLevel;
    if (subLevel >= 0)
      EncodeSubband(subLevel);
  }
}

template <class TYPE>
void CMGGridCoder<TYPE>::ExtractCoordinates(CMatrix<TYPE> &coordX, CMatrix<TYPE> &coordY, CMatrix<TYPE> &coordZ)
{
	for (int w = 0; w < m_nrSlices.w; w++)
	{
		const GridUV &gridUV = (*m_pGrid)[w];
		TYPE **sliceX = coordX[w];
		TYPE **sliceY = coordY[w];
		TYPE **sliceZ = coordZ[w];

		for (int v = 0; v < m_nrSlices.v; v++)
		{
			const GridU &gridU = gridUV[v];
			TYPE *elemX = sliceX[v];
			TYPE *elemY = sliceY[v];
			TYPE *elemZ = sliceZ[v];

			for (int u = 0; u < m_nrSlices.u; u++)
			{
				const CFPoint3DT &p = gridU[u];
				*elemX++ = p.x;
				*elemY++ = p.y;
				*elemZ++ = p.z;
			}
		}
	}
}


#endif // !defined(AFX_MGGRIDCODEC_H__85E0A823_97E8_4B6D_8F77_A9974CF76CD7__INCLUDED_)
