/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

// MGMeshCoderBinary.cpp: implementation of the CMGMesh class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MGMesh.h"
#include "MGGridCodec.h"
#include "BitStream.h"
#include "MGStreamParse.h"
#include "MgErrMsg.h"


// Save as binary file
bool CMGMesh::Save(CFile *file, const bool *bLODMesh, CPointUVW *szROI, const bool *bLODGrid, const CPointUVW *tileGroup,
                   bool hasConnectivityInfo, bool hasRefineInfo, bool hasRepositionInfo, bool hasGridInfo, 
                   int nrRefinementBits, bool interleaveRefine, int nrQuabtizationBits, float cr, int filterType, CString &error)
{
  bool res = true;
  try
  {
    // check the ROI size for each resolution level
//    CheckROISize(szROI);

    // create the parser object
//    CreateParser(NULL, CBitStream::OP_WRITE);
    CreateParser(file, CBitStream::OP_WRITE);

    // initialise the exchange object with the mesh coding details
    InitExchangeObjectMeshBinary(bLODMesh, szROI, hasConnectivityInfo, hasRefineInfo, hasRepositionInfo, hasGridInfo,
                                 nrRefinementBits, interleaveRefine);

    // create and start the encoder object
    CreateAndStartEncoder(bLODGrid, tileGroup, nrQuabtizationBits, cr, filterType);

	  // initialise the parser and store the header of the stream
    bool keepHeader = false;
    StoreMeshGridHeader(keepHeader);

	  // stream all the levels
	  for (int level = 0; level < GetMaxNrLevels(); level++)
	  {
      // store the mesh coding for a level
      StoreMGMesh(level);

      // store the grid coding for a level
      StoreMGGrid(level);
    }
  }
  catch (CMGDataException &e)
  {
    res = false;    error = e.GetErrorMessage();
  }

  // copy the size and content of the memory stream into file
  WriteOutput(file);

	// stop the encoder and flush the buffers
  try
  {
	  StopEncoder();
  }
  catch (CMGDataException &e)
  {
    res = false;    error = e.GetErrorMessage();
  }

	return res;
}

/**********************************************************************************/

void CMGMesh::StoreMeshGridHeader(bool keepHeader)
{
  m_pParser->StoreMeshGridHeader(keepHeader);
}

void CMGMesh::WriteOutput(CFile *file)
{
  m_pParser->WriteOutput(file);
}

void CMGMesh::DoConnectivityCoding(int level)
{
  if (!m_pStreamDescr->IsConnectivityInfoAvailable(level))
    return;

  // begin of connectivity coding
  OnBeginMeshCodingConnectivityROI();

  // a ROI may consist of several sub-patches
  bool bFinish = false;
  while (!bFinish)
  {
    // find an appropriate starting point
    GetNextStartingPoint(level);

    // first save the vertices' connectivities
	  SaveMeshConnectivityROI(bFinish, level);
  }

  // end of mesh connectivity coding
  OnEndMeshCodingConnectivityROI();
}

void CMGMesh::DoVertsReposCoding(int level)
{
  if (!m_pStreamDescr->IsRepositionInfoAvailable(level))
    return;

  // begin of vertices' reposition bits coding
  OnBeginMeshCodingVertsReposROI();
  
  // second save the reposition bits
	SaveMeshReposROI(level);

  // end of vertices' reposition bits coding
  OnEndMeshCodingVertsReposROI();
}

void CMGMesh::DoVertsRefineCoding(int level)
{
  if (!m_pStreamDescr->IsRefinementInfoAvailable(level))
    return;

  // begin of vertices' reposition bits coding
  OnBeginMeshCodingVertsRefineROI();
  
  // second save the refine bits
	SaveMeshRefineROI(level);

  // end of vertices' reposition bits coding
  OnEndMeshCodingVertsRefineROI();
}

void CMGMesh::StoreMGMesh(int level)
{
	// start a new level description
  OnBeginMeshCodingLevel(level);

  m_boxIndex = 0;
  InitFirstROI(level);

  do
  {
	  // start a new ROI description
    OnBeginMeshCodingROI(m_boxIndex);

    bool bEmpty = IsROIEmpty(level);
    if (!bEmpty)
    {
      // do the connectivity, vertices' reposition/refine coding
      if (m_pStreamDescr->IsConnectivityInfoAvailable(level))
        DoConnectivityCoding(level);

      if (m_pStreamDescr->IsRefinementInfoAvailable(level))
        DoVertsRefineCoding(level);

      if (m_pStreamDescr->IsRepositionInfoAvailable(level))
        DoVertsReposCoding(level);

      CleanUpVerticesList(level, m_boxIndex);
    }

    // end of ROI description
    OnEndMeshCodingROI(bEmpty);

    // index of the ROI
    ++m_boxIndex;
  }
  while (GetNextROI(level));

  // end of level description
  OnEndMeshCodingLevel(level);
}

void CMGMesh::StoreMGGrid(int level)
{
  if (m_pStreamDescr->IsGridInfoAvailable(level))
    m_pMGOUT->StreamLevel(level);
}

void CMGMesh::CheckROISize(CPointUVW *szROI)
{
  const CPointUVW &sz0 = szROI[0];
  for (int level = 1; level < GetMaxNrLevels(); level++)
  {
    const CPointUVW &sz = szROI[level];
    if ((sz0.u != sz.u && sz0.u % sz.u != 0) ||
        (sz0.v != sz.v && sz0.v % sz.v != 0) ||
        (sz0.w != sz.w && sz0.w % sz.w != 0))
      throw CMGDataException(CMGErrMsg::WRONG_ROI_SIZE);
  }
}

/**********************************************************************************/

void CMGMesh::CreateParser(CFile *file, int op)
{
	TRY
	{
    m_pParser = new CMGStreamParser(file, op);
	}
	CATCH_ALL(e)
	{
		throw CMGDataException(e);
	}
	END_CATCH_ALL
}

void CMGMesh::CreateAndStartEncoder(const bool *bLODGrid, const CPointUVW *tileGroup, int quantizationBits,
                                    float cr, int filterType)
{
	TRY
	{
		m_pMGOUT = new CMGGridCoder<float>(m_pParser, m_grid, cr, quantizationBits, bLODGrid, tileGroup, filterType);
    m_pMGOUT->StreamHeader();
	}
	CATCH_ALL(e)
	{
		throw CMGDataException(e);
	}
	END_CATCH_ALL
}

void CMGMesh::StopEncoder()
{
  if (m_startVertsROI != NULL)
  {
    for (int level = 0; level < GetMaxNrLevels(); level++)
      if (m_startVertsROI[level] != NULL) delete []m_startVertsROI[level];

    delete []m_startVertsROI;   m_startVertsROI = NULL;
  }

  delete m_pParser;   m_pParser = NULL;
	delete m_pMGOUT;		m_pMGOUT = NULL;
}

/**********************************************************************************/
  
void CMGMesh::ComputeROISize(const CPointUVW &nrSlices, CPointUVW &szROI, CPointUVW &maxSzROI, CPointUVW &nrROIs)
{
  ASSERT(nrSlices >= CMGStreamParser::MIN_SLICES);
  ASSERT(szROI >= CMGStreamParser::MIN_ROI_SIZE);

  if (szROI.u < CMGStreamParser::MIN_ROI_SIZE) szROI.u = CMGStreamParser::MIN_ROI_SIZE;
  if (szROI.v < CMGStreamParser::MIN_ROI_SIZE) szROI.v = CMGStreamParser::MIN_ROI_SIZE;
  if (szROI.w < CMGStreamParser::MIN_ROI_SIZE) szROI.w = CMGStreamParser::MIN_ROI_SIZE;

  nrROIs = nrSlices / szROI;
  if (nrROIs.u < 1) nrROIs.u = 1;
  if (nrROIs.v < 1) nrROIs.v = 1;
  if (nrROIs.w < 1) nrROIs.w = 1;

  szROI = nrSlices / nrROIs;
  maxSzROI = nrSlices - (nrROIs - 1) * szROI;
}

void CMGMesh::InitExchangeObjectMeshBinary(const bool *bLODMesh, CPointUVW *szROI, bool hasConnectivityInfo,
                                           bool hasRefineInfo, bool hasRepositionInfo, bool hasGridInfo,
                                           int refinmentBits, bool interleaveRefine)
{
	TRY
	{
    int nrLevels = GetMaxNrLevels();

    ASSERT(m_startVertsROI == NULL);
    m_startVertsROI = new CPtrList*[nrLevels];
    memset(m_startVertsROI, 0, nrLevels * sizeof(void *));

    // initialise the exchange objects
    m_pStreamDescr = &m_pParser->m_streamDescr;
    m_pStreamDescr->InitMeshLevels(m_grid->GetNrLevelsUVW());

    // store the total number of slices
    m_pStreamDescr->SetMaxNrSlices(m_nrSlices);

    // mesh type
    m_pStreamDescr->SetMeshType(GetMeshType());
    if (GetMeshType() == QUADRI_MESH)
    {
      if (hasConnectivityInfo)
      {
        m_pStreamDescr->SetSameConnectivity(IsSameConnectivity());
        m_pStreamDescr->SetAutoResolution(IsAutoResolution());
      }
      else
      {
        m_pStreamDescr->SetSameConnectivity(false);
        m_pStreamDescr->SetAutoResolution(false);
      }

      const CPointUVW & nrSlices = GetNrSlices();
      if (m_pStreamDescr->IsSingleLayered())
      {
        m_pStreamDescr->SetOffsetAmplitude(GetOffsetAmplitude());
        if (GetOffsetAmplitude() == 0)
        {
          hasRefineInfo = 0; // no offsets
        }
      }

      if (IsAutoResolution())
      {
        // check for consistency
        ASSERT(IsSameConnectivity());
        if (hasRefineInfo)  
          interleaveRefine = true;       // need to set it because only the first
      }                                  // resolution level is available
    }

    // store the flags
    m_pStreamDescr->HasConnectivityInfo(hasConnectivityInfo);
    m_pStreamDescr->HasRefineInfo(hasRefineInfo);
    m_pStreamDescr->HasRepositionInfo(hasRepositionInfo);
    m_pStreamDescr->HasGridInfo(hasGridInfo);

    // refinement info
    if (hasRefineInfo)
    {
      ASSERT(refinmentBits > 0);

      m_bInterleaveRefine = interleaveRefine;
      m_nrRefBitsVert = refinmentBits;
    }
    else
    {
      ASSERT(interleaveRefine == false);
      ASSERT(refinmentBits == 0);

      m_bInterleaveRefine = false;
      m_nrRefBitsVert = 0;
    }

    m_pStreamDescr->SetInterleaveRefine(m_bInterleaveRefine);
    m_pStreamDescr->SetNrRefineBitsVertex(m_nrRefBitsVert);
    m_pStreamDescr->SetCyclicMode(GetCyclicMode());

    // intialise the level information
    MStreamLevels &streamLevel = m_pStreamDescr->GetMeshLevel();
    for (int level = 0; level < nrLevels; level++)
    {
      MStreamLevel &meshLevel = streamLevel[level];
      meshLevel.SetNrSlices(GetNrSlices(level));

      // determine the ROI size
      CPointUVW maxROISz, maxROI = GetNrSlices(level);
      CPointUVW nrROIs, roiSZ = szROI[level];

      ComputeROISize(maxROI, roiSZ, maxROISz, nrROIs);
      
      m_pParser->SetROISize(roiSZ, maxROISz, level);
      meshLevel.Init(nrROIs);
    }
	}
	CATCH_ALL(e)
	{
		throw CMGDataException(e);
	}
	END_CATCH_ALL
}

void CMGMesh::InitExchangeObjectMesh(const bool *bLODMesh, CPointUVW *szROI, bool hasConnectivityInfo,
                                     bool hasRefineInfo, bool hasRepositionInfo, bool interleaveRefine)
{
	TRY
	{
    int nrLevels = GetMaxNrLevels();

    ASSERT(m_startVertsROI == NULL);
    m_startVertsROI = new CPtrList*[nrLevels];
    memset(m_startVertsROI, 0, nrLevels * sizeof(void *));

    // initialise the exchange objects
    m_pStreamDescr = new MGStreamDescr;
    m_pStreamDescr->InitMeshLevels(m_grid->GetNrLevelsUVW());

    // store the total number of slices
    m_pStreamDescr->SetMaxNrSlices(m_nrSlices);

    // mesh type
    m_pStreamDescr->SetMeshType(GetMeshType());
    if (GetMeshType() == QUADRI_MESH)
    {
      if (hasConnectivityInfo)
      {
        m_pStreamDescr->SetSameConnectivity(IsSameConnectivity());
        m_pStreamDescr->SetAutoResolution(IsAutoResolution());
      }
      else
      {
        m_pStreamDescr->SetSameConnectivity(false);
        m_pStreamDescr->SetAutoResolution(false);
      }

      const CPointUVW & nrSlices = GetNrSlices();
      if (m_pStreamDescr->IsSingleLayered())
      {
        m_pStreamDescr->SetOffsetAmplitude(GetOffsetAmplitude());
        if (GetOffsetAmplitude() == 0)
        {
          hasRefineInfo = 0; // no offsets
        }
      }

      if (IsAutoResolution())
      {
        // check for consistency
        ASSERT(IsSameConnectivity());
        if (hasRefineInfo)
          interleaveRefine = true;       // need to set it because only the first
      }                                  // resolution level is available
    }

    // store the flags
    m_pStreamDescr->HasConnectivityInfo(hasConnectivityInfo);
    m_pStreamDescr->HasRefineInfo(hasRefineInfo);
    m_pStreamDescr->HasRepositionInfo(hasRepositionInfo);
    m_pStreamDescr->HasGridInfo(true);

    // refinement info
    if (hasRefineInfo)
    {
      m_bInterleaveRefine = interleaveRefine;
      m_nrRefBitsVert = 1;
    }
    else
    {
      ASSERT(interleaveRefine == false);

      m_bInterleaveRefine = false;
      m_nrRefBitsVert = 0;
    }

    m_pStreamDescr->SetInterleaveRefine(m_bInterleaveRefine);
    m_pStreamDescr->SetNrRefineBitsVertex(m_nrRefBitsVert);
    m_pStreamDescr->SetCyclicMode(GetCyclicMode());

    // intialise the level information
    MStreamLevels &streamLevel = m_pStreamDescr->GetMeshLevel();
    for (int level = 0; level < nrLevels; level++)
    {
      MStreamLevel &meshLevel = streamLevel[level];
      meshLevel.SetNrSlices(GetNrSlices(level));

      // determine the ROI size
      CPointUVW maxROISz, maxROI = GetNrSlices(level);
      CPointUVW nrROIs, roiSZ = szROI[level];

      ComputeROISize(maxROI, roiSZ, maxROISz, nrROIs);
      
      meshLevel.Init(nrROIs);
      meshLevel.SetROISize(roiSZ, 0);
    }
	}
	CATCH_ALL(e)
	{
		throw CMGDataException(e);
	}
	END_CATCH_ALL
}

/******************************************************************************/

int CMGMesh::GetNrVerticesROI(int level)
{
  int count = 0;
	for (int u = m_posVertMin.u; u <= m_posVertMax.u; u++)
	{
		CSurfNodeList *pList = m_nodeStack[u];
		if (pList == NULL) continue;

		for (int v = m_posVertMin.v; v <= m_posVertMax.v; v++)
		{
			CSurfNodeT *pSN = pList[v].GetFirstNode();
			while (pSN != NULL)
			{
				if (IsVertexInROI(pSN, level))
        {
          if (pSN->IsVisited() != 0)
            throw CMGDataException("Overlaping vertices");
          ++count;
        }
        else
        {
          if (pSN->GetLevel() <= level)
            ASSERT(InROI(pSN->GetPos(), pSN->GetDir(), level) == false);
        }

        pSN = pSN->GetDownList();
			}
		}
  }

  return count;
}
	
int	CMGMesh::ComputeNrVertices(int level)
{
	ASSERT(level >= 0 && level < GetMaxNrLevels());
	
  int count = 0;
  for (int u = 0; u < m_nrSlices.u; u++)
  {
		// check row
		CSurfNodeList *pList = m_nodeStack[u];
		if (pList == NULL) continue;

		// check colums
    for (int v = 0; v < m_nrSlices.v; v++)
    {
      CSurfNodeT *pSN = pList[v].GetFirstNode();
			while(pSN != NULL)
			{
        if (pSN->GetLevel() <= level)
        {
          ++count;
        }
				pSN = pSN->GetDownList();
			}
    }
  }
  return count;
}

void CMGMesh::CheckValidVerticesNumber(int nrROIs, int level)
{
  // reset the streamed flag for the vertices
  ResetVisitedVerticesGrid(level);

  // reset the position of the ROI
  InitFirstROI(level);
  int nrVerts = 0, count = 0;
  do
  {
    ++count;
    nrVerts += GetNrVerticesROI(level);
  }
  while (GetNextROI(level));
  ASSERT(count == nrROIs);

  int nrVerLevel = ComputeNrVertices(level);
  ASSERT(nrVerts == nrVerLevel);
}

void CMGMesh::OnBeginMeshCodingLevel(int level)
{
  // initialise the buffer containing the start vertices positions for each roi
  CPointUVW maxROI = GetNrSlices(level);

  const MStreamLevel &meshLevel = m_pStreamDescr->GetMeshLevel(level);
  int nrROIs = meshLevel.GetTotalNrROIs();
  m_startVertsROI[level] = new CPtrList[nrROIs];

#ifdef _DEBUG
  CheckValidVerticesNumber(nrROIs, level);
#endif

  // reset the streamed flag for the vertices
  ResetVisitedVerticesGrid(level);

  // mark the beginning of a mesh level
  m_pParser->OnBeginMeshCodingLevel(level);
}

void CMGMesh::OnBeginMeshExportLevel(int level)
{
  const MStreamLevel &meshLevel = m_pStreamDescr->GetMeshLevel(level);
  int nrROIs = meshLevel.GetTotalNrROIs();
  m_startVertsROI[level] = new CPtrList[nrROIs];

#ifdef _DEBUG
  CheckValidVerticesNumber(nrROIs, level);
#endif

  // reset the streamed flag for the vertices
  ResetVisitedVerticesGrid(level);
}

void CMGMesh::OnEndMeshCodingLevel(int level)
{
  // mark the end of a mesh level
  m_pParser->OnEndMeshCodingLevel();

  // delete the ROI list
  delete []m_startVertsROI[level];    m_startVertsROI[level] = NULL;
}

void CMGMesh::OnBeginMeshCodingROI(int roiIdx)
{
  m_pParser->OnBeginMeshCodingROI(roiIdx);
}

void CMGMesh::OnEndMeshCodingROI(bool bEmpty)
{
  m_pParser->OnEndMeshCodingROI(bEmpty);
}

void CMGMesh::OnBeginMeshCodingConnectivityROI()
{
  m_pParser->OnBeginMeshCodingConnectivityROI();
}

void CMGMesh::OnEndMeshCodingConnectivityROI()
{
  m_pParser->OnEndMeshCodingConnectivityROI();
}

void CMGMesh::OnBeginMeshCodingVertsReposROI()
{
  m_pParser->OnBeginMeshCodingVertsReposROI();
}

void CMGMesh::OnEndMeshCodingVertsReposROI()
{
  m_pParser->OnEndMeshCodingVertsReposROI();
}

void CMGMesh::OnBeginMeshCodingVertsRefineROI()
{
  m_pParser->OnBeginMeshCodingVertsRefineROI();
}

void CMGMesh::OnEndMeshCodingVertsRefineROI()
{
  m_pParser->OnEndMeshCodingVertsRefineROI();
}

/**********************************************************************************/

void CMGMesh::SaveMeshConnectivityROI(bool &bFinish, int level)
{
  CSurfNodeT *pSNStart = CheckStartingPoint(level);
  if (!pSNStart)
  {
    bFinish = false;  return;
  }

  // we need the starting point when we refine the vertices
  CacheStartPoint(pSNStart, level);

  // stream the position of the first node
  StreamStartNode(level, pSNStart, true);
  
  // write the node
  StreamNodeBox(level, pSNStart); 

	// check whether the entire box is streamed
  bFinish = IsROIStreamed(level);
}

void CMGMesh::SaveMeshRefineROI(int level)
{
  CPtrList &startNodeList = m_startVertsROI[level][m_boxIndex];
  if (startNodeList.IsEmpty())  return;

  // write the vertices' refine bits
  int nrBits = GetNrRefBitsVert();
	int t = 1 << (nrBits - 1);
  int maxVal = (1 << nrBits) - 1;

  for (int k = nrBits; k >= 1; k--, t >>= 1)
  {
    // reset the visited vertices
    ResetVisitedVerticesBox(level);

    // loop for all start vertices
    POSITION pos = startNodeList.GetHeadPosition();
    while (pos)
    {
      CSurfNodeT *pSNStart = (CSurfNodeT *)startNodeList.GetNext(pos);

      StreamVerticesRefineBits(level, t, maxVal, pSNStart);
    }
  }
}

void CMGMesh::SaveMeshReposROI(int level)
{
  int maxLevel = GetMaxNrLevels() - 1;
  ASSERT(level < maxLevel);

  CPtrList &startNodeList = m_startVertsROI[level][m_boxIndex];
  if (startNodeList.IsEmpty())  return;

  // reset the visited vertices
  ResetVisitedVerticesBox(level);

  // loop for all start vertices
  POSITION pos = startNodeList.GetHeadPosition();
  while (pos)
  {
    CSurfNodeT *pSNStart = (CSurfNodeT *)startNodeList.GetNext(pos);

    // write the reposition bit
    StreamVerticesRepositionBit(level, pSNStart);
  }
}

void CMGMesh::CleanUpVerticesList()
{
  // clear the list
  if (m_pStreamDescr->HasConnectivityInfo())
  {
    int maxLevel = IsAutoResolution() ? 1 : GetMaxNrLevels();
    for (int level = 0; level < maxLevel; level++)
      CleanUpVerticesList(level);
  }
  delete []m_startVertsROI;   m_startVertsROI = NULL;
}

void CMGMesh::CleanUpVerticesList(int level)
{
  // clear the list
  const MStreamLevel &meshLevel = m_pStreamDescr->GetMeshLevel(level);
  int maxNrROIs = meshLevel.GetTotalNrROIs();

  for (int boxIndex = 0; boxIndex < maxNrROIs; boxIndex++)
    CleanUpVerticesList(level, boxIndex);

  delete []m_startVertsROI[level];  m_startVertsROI[level] = NULL;
}

void CMGMesh::CleanUpVerticesList(int level, int boxIndex)
{
  // clear the list
  CPtrList &startNodeList = m_startVertsROI[level][boxIndex];
  startNodeList.RemoveAll();
}

/**********************************************************************************/

void CMGMesh::StreamStartNode(int level, CSurfNodeT *pSNStart, bool bLOD)
{
	CPointUVW pos3D = pSNStart->GetPos();
	FindPosNode(pos3D, pSNStart->GetDir(), level);

  if (bLOD)
  {
    pos3D.u -= m_localROI.u;
    pos3D.v -= m_localROI.v;
    pos3D.w -= m_localROI.w;
  }

  BYTE dir = pSNStart->GetDir() >> OFFSET_DIR;
	StreamAbsoluteNode(level, pos3D, dir);
}

void CMGMesh::StreamAbsoluteNode(int level, const CPointUVW &pos3D, BYTE dir)
{
  m_pParser->SetAbsoluteVertex(pos3D, dir, level);
}

void CMGMesh::StreamNodeBox(int level, CSurfNodeT *pSNStart)
{
	// add the starting vertex to the list
	AddStartNodeBox(pSNStart, true);

	while (m_nodeList.GetCount() != 0)
	{
		CSurfNodeT *pSN = (CSurfNodeT *)m_nodeList.RemoveHead();
		ASSERT(pSN->GetLevel() <= level);

		BYTE symbol;
		CSurfNodeT *pSNNext1 = pSN->GetNextNode1(level);
		CSurfNodeT *pSNNext2 = pSN->GetNextNode2(level);

		// write the Link1 connectivity vector
		if (pSNNext1 != NULL)
			GetDirectionDifference(pSN->GetDir(), pSNNext1->GetDir(), symbol);
		else
			symbol = NO_LINK;
		WriteConnectivitySymbol(symbol);

		// write the Link2 connectivity vector
		if (pSNNext2 != NULL)
			GetDirectionDifference(pSN->GetDir(), pSNNext2->GetDir(), symbol);
		else
			symbol = NO_LINK;
		WriteConnectivitySymbol(symbol);

		AddNodesToListBox(pSNNext1, pSNNext2, level, true);
	}
}

/**********************************************************************************/

void CMGMesh::StreamVerticesRepositionBit(int level, CSurfNodeT *pSNStart)
{
	ASSERT(level < GetMaxNrLevels() - 1);

	// add the starting vertex to the list
	AddStartNodeBox(pSNStart);

	while (m_nodeList.GetCount() != 0)
	{
		CSurfNodeT *pSN = (CSurfNodeT *)m_nodeList.RemoveHead();
		ASSERT(pSN->GetLevel() <= level);

    bool bSpecialCase;
		BYTE val = VertexRepositionBit(level + 1, pSN, bSpecialCase);
		WriteRepositionSymbol(val, bSpecialCase);

		AddNodesToListBox(pSN->GetNextNode1(level),
									    pSN->GetNextNode2(level), level, false);
	}
}

void CMGMesh::StreamVerticesRefineBits(int level, int mask, int maxVal, CSurfNodeT *pSNStart)
{
	// add the starting vertex to the list
	AddStartNodeBox(pSNStart);

	while (m_nodeList.GetCount() != 0)
	{
		CSurfNodeT *pSN = (CSurfNodeT *)m_nodeList.RemoveHead();
		ASSERT(pSN->GetLevel() <= level);

		float offset;
		if (level < GetMaxNrLevels() - 1)
			offset = GetLevelOffset(pSN->GetDir(), pSN->GetPos(), pSN->GetOffset(), level);
		else
			offset = pSN->GetOffset();
		ASSERT(offset >= 0 && offset < 1);

		int bitVal = (int) floor(offset * maxVal + 0.5f);
		if ((bitVal & mask) != 0)
			WriteRefineSymbol(1);
		else
			WriteRefineSymbol(0);

		AddNodesToListBox(pSN->GetNextNode1(level),
									    pSN->GetNextNode2(level), level, false);
	}
}

/**********************************************************************************/

void CMGMesh::WriteConnectivitySymbol(BYTE symbol)
{
  m_pParser->SetVertexConnectivity(symbol);
}

void CMGMesh::WriteRepositionSymbol(BYTE symbol, bool bSpecialCase)
{
  m_pParser->SetVertexReposition(symbol, bSpecialCase);
}

void CMGMesh::WriteRefineSymbol(BYTE symbol)
{
  m_pParser->SetVertexRefineBit(symbol);
}

/**********************************************************************************/