/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

// MGMeshDecoderBinary.cpp: implementation of the CMGMesh class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MGMesh.h"
#include "MGGridCodec.h"
#include "BitStream.h"
#include "MGStreamParse.h"
#include "MgErrMsg.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


/**********************************************************************************/

bool CMGMesh::LoadCompressed(CString & error, int streamSize, bool dump)
{
  bool res = true;
  m_totalNrVertices = 0;

  try
  {
    // initialise the decoder and decode the grid corners
    StartDecoder();

    // update the corners of the grid
    UpdateGridCorners();

    // if no connectivity generate a default
    if (!m_pStreamDescr->HasConnectivityInfo())
      DefaultMeshDescriptor();

    while (streamSize > 0)
    {
      // parse the part type and resolution level
      int size = ParseMeshGridPart();
      streamSize -= size;
    }
    ASSERT(streamSize == 0);
  
    // if no connectivity generate a default
    if (!m_pStreamDescr->HasConnectivityInfo())
      BuildDefaultCW();
  }
  catch (CMGDataException &e)
  {
    res = false;    error = e.GetErrorMessage();
  }
  
  CleanUpVerticesList();

  // Output the MeshGrid description
  if (dump)
  {
	  CStdioFile output;
	  CFileException e;
	  if (output.Open("dump.xml", CFile::modeCreate | CFile::modeWrite, &e))
	  {
      Dump(output);     output.Close();
	  }
    else
    {
		  TRACE0("Error creating the output file.\n");
    }
  }

  // stop the decoder and flush the buffers
  try
  {
	  StopDecoder();
  }
  catch (CMGDataException &e)
  {
    res = false;    error = e.GetErrorMessage();
  }
	
	return res;
}

/**********************************************************************************/

void CMGMesh::StartDecoder()
{
	TRY
	{
		m_pMGIN = new CMGGridDecoder<float>(m_pParser, m_grid, 1.0f);
	}
	CATCH_ALL(e)
	{
		throw CMGDataException(e);
	}
	END_CATCH_ALL
}

void CMGMesh::StopDecoder()
{
  if (m_startVertsROI != NULL)
  {
    for (int level = 0; level < GetMaxNrLevels(); level++)
      if (m_startVertsROI[level] != NULL) delete []m_startVertsROI[level];

    delete []m_startVertsROI;   m_startVertsROI = NULL;
  }

	delete m_pMGIN;			m_pMGIN = NULL;
}

void CMGMesh::UpdateGridCorners()
{
  m_pMGIN->UpdateGridCorners();
}

/**********************************************************************************/

void CMGMesh::ParseStream(int streamSize)
{
  m_pParser->ParseMeshGridStream(streamSize);
}

int CMGMesh::ParseMeshGridPart()
{
  // parse the part tag and size of part
  BYTE tag;   int partSize;
  int nrBytes = ParseTagAndSize(tag, partSize);

#ifdef _DEBUG
  int startPos = m_pParser->GetStreamPositon();
#endif

  int level;
  switch (tag)
  {
	case MGMeshInfoTag:                   ParseLevel(level);  MeshDescriptor(level);        break;
  case MGGridInfoTag:                   ParseLevel(level);  GridDescriptor(level);        break;
  case MGMeshConnectivityROIInfoTag:    ParseLevel(level);  MeshConnectivityROI(level);   break;
  case MGMeshConnectivityInfoTag:       ParseLevel(level);  MeshConnectivity(level);      break;
  case MGVerticesRepositionROIInfoTag:  ParseLevel(level);  VertexRepositionROI(level);   break;
  case MGVerticesRepositionInfoTag:     ParseLevel(level);  VertexReposition(level);      break;
  case MGVerticesRefinementROIInfoTag:  ParseLevel(level);  VertexRefineROI(level);       break;
  case MGVerticesRefinementInfoTag:     ParseLevel(level);  VertexRefine(level);          break;
  case MGGridCoefficientsROIInfoTag:    ParseLevel(level);  GridCoefficientsROI(level);   break;
  case MGGridCoefficientsInfoTag:       ParseLevel(level);  GridCoefficients(level);      break;
  default:
    // skip the part
    TRACE0(CMGErrMsg::GetMessage(CMGErrMsg::WRONG_PART_IDENTIFIER));
    m_pParser->MoveStreamPosition(partSize);
  }

#ifdef _DEBUG
  int finalPos = m_pParser->GetStreamPositon();
  ASSERT(finalPos == startPos + partSize);
#endif

  m_pParser->EndOfPart();

  return nrBytes + partSize;
}

int CMGMesh::ParseStreamSize(int &counterLength)
{
  return m_pParser->ParseStreamSize(counterLength);
}

int CMGMesh::ParseTagAndSize(BYTE &tag, int &size)
{
  return m_pParser->ParseTagAndSize(tag, size);
}

void CMGMesh::ParseLevel(int &level)
{
  m_pParser->ParseLevel(level);
}

void CMGMesh::DefaultMeshDescriptor()
{
  CPointUVW nrROIs = 1;
  MStreamLevel &meshLevel0 = m_pStreamDescr->GetMeshLevel(0);
  meshLevel0.Init(nrROIs);

  // store the number of slices
  CPointUVW nrSlices = m_pStreamDescr->GetMaxNrSlices();
  const CPointUVW &nrLevels = m_pStreamDescr->GetNrLevelsMesh();

  CMGGrid::ComputeNrSlicesLevel(0, nrLevels, nrSlices);
  meshLevel0.SetNrSlices(nrSlices);

  // compute the ROI size
  // compute the max ROI size
  CPointUVW szROI = nrSlices;
  CPointUVW maxSzROI = nrSlices;
  
  // initilaise the mesh level
  m_pParser->SetROISize(szROI, maxSzROI, 0);

  for (int levelM = 1; levelM < GetMaxNrLevels(); levelM++)
    m_pParser->DefaultMeshDescriptor(levelM);
}

void CMGMesh::MeshDescriptor(int level)
{
  // parse the description & byte allign data
  m_pParser->MeshDescriptor(level);

  if (m_pStreamDescr->IsAutoResolution())
  {
    ASSERT(level == 0);
    for (int levelM = 1; levelM < GetMaxNrLevels(); levelM++)
      m_pParser->DefaultMeshDescriptor(levelM);
  }

  // do some initialisation
  RetrieveMeshLevelHeaders(level);
}

void CMGMesh::GridDescriptor(int levelG)
{
  // parse the description & byte allign data
  int levelM = m_pParser->GridDescriptor(levelG);

  m_pMGIN->CreateTiles(levelG, levelM);
}

void CMGMesh::BuildDefaultCWLevels()
{
  for (int levelM = 0; levelM < GetMaxNrLevels() - 1; levelM++)
  {
    DefaultMeshConnectivity(levelM);
    BuildQuadrilateralSurface(levelM + 1);
  }
}

void CMGMesh::BuildDefaultCW()
{
  // build first level CW
  BYTE cyclicMode = m_pStreamDescr->GetCyclicMode();
  DefineConectivityUV((cyclicMode & CYCLIC_U) != 0,
                      (cyclicMode & CYCLIC_V) != 0);
  BuildQuadrilateralSurface(0);

  // build the higher resolution CW
  BuildDefaultCWLevels();
}

void CMGMesh::BuildMeshUV(int u1, int u2, int v1, int v2)
{
	CPointUVW	pos3D;
	CSurfNodeList *pNodeList = NULL;

  // the w position is always 0
  pos3D.w = 0;

  // top left corner
  pos3D.u = u1;
  pos3D.v = v1;
	pNodeList = GetColumn(pos3D.u);
	CSurfNodeT *pSN1 = GetAddNode(pNodeList, pos3D, TOP_FACE, 0, 0.5f);

  // top right corner
  pos3D.u = u1;
  pos3D.v = v2;
	pNodeList = GetColumn(pos3D.u);
	CSurfNodeT *pSN2 = GetAddNode(pNodeList, pos3D, TOP_FACE, 0, 0.5f);

  // bottom left corner
  pos3D.u = u2;
  pos3D.v = v1;
	pNodeList = GetColumn(pos3D.u);
	CSurfNodeT *pSN3 = GetAddNode(pNodeList, pos3D, TOP_FACE, 0, 0.5f);

  // bottom right corner
  pos3D.u = u2;
  pos3D.v = v2;
	pNodeList = GetColumn(pos3D.u);
	CSurfNodeT *pSN4 = GetAddNode(pNodeList, pos3D, TOP_FACE, 0, 0.5f);

	// set the connectivity in the UV plane
	pSN1->SetPrevNodeWU(pSN3, TOP_FACE);
	pSN3->SetNextNodeWU(pSN1, TOP_FACE);
	pSN2->SetPrevNodeWU(pSN4, TOP_FACE);
	pSN4->SetNextNodeWU(pSN2, TOP_FACE);

	pSN1->SetNextNodeVW(pSN2, TOP_FACE);
	pSN2->SetPrevNodeVW(pSN1, TOP_FACE);
	pSN3->SetNextNodeVW(pSN4, TOP_FACE);
	pSN4->SetPrevNodeVW(pSN3, TOP_FACE);
}

bool CMGMesh::DefineConectivityUV(bool cyclicU, bool cyclicV)
{
  const CPointUVW &nrSlices = GetNrSlices(0);
  try
  {
    // (u,v) = position of the top left corner of a quad
    int v0 = GetNextSliceV(-1, 0);
    for (int v = 0; v < nrSlices.v - 1; v++)
    {
      int v1 = GetNextSliceV(v0, 0);
      int u0 = GetNextSliceU(-1, 0);

      for (int u = 0; u < nrSlices.u - 1; u++)
      {
        int u1 = GetNextSliceU(u0, 0);
        BuildMeshUV(u0, u1, v0, v1);
        u0 = u1;
      }

      v0 = v1;
    }
  }
  catch (CMemoryException *e)
  {
    e->ReportError();   return false;
  }

	return true;
}

void CMGMesh::MeshConnectivity(int level)
{
  // retrieve the mesh connectivity header for the level & byte allign the data
  m_pParser->MeshConnectivityHeader(level);

  // retrieve and decode the mesh connectivity bits
  RetrieveMeshConnectivityPart(level);

  // build the triangles
  BuildSmoothSurface(level);

  if (m_pStreamDescr->IsAutoResolution())
  {
    ASSERT(level == 0);
    for (int levelM = 0; levelM < GetMaxNrLevels() - 1; levelM++)
    {
      DefaultMeshConnectivity(levelM);
      BuildQuadrilateralSurface(levelM + 1);

      // update the number vertices
      m_nrVertices[levelM + 1] = m_totalNrVertices;
    }
  }
}

void CMGMesh::VertexReposition(int level)
{
  // retrieve the vertices reposition header for the level & byte allign the data
  m_pParser->VertexRepositionHeader(level);

  // retrieve and decode the vertices reposition bits
  RetrieveVertexRepositionPart(level);
}

void CMGMesh::VertexRefine(int level)
{
  // retrieve the vertices refinement header for the level & byte allign the data
  m_pParser->VertexRefineHeader(level);

  // retrieve and decode the vertices refinement bits
  RetrieveVertexRefinePart(level);
}

void CMGMesh::GridCoefficients(int levelG)
{
  // retrieve the mesh connectivity header for the level & byte allign the data
  m_pParser->GridCoefficientsHeader(levelG);

  // retrieve and decode the tiles bits
  m_pMGIN->RecoverLevel(levelG);

  // compute the level corresponding to the mesh
  int nrLevelsM = m_pStreamDescr->GetMaxNrLevelsMesh();
  int nrLevelsG = m_pStreamDescr->GetMaxNrLevelsGrid();
  int endLevel  = nrLevelsG - nrLevelsM;

  int levelM = levelG - endLevel;
  if (levelM < 0) levelM = 0;

  UpdateMGfromRG(levelM);
}

int *CMGMesh::MakeIndices(int number)
{
	int i = 0;
	int *index = new int[number];

	for (i = 0; i < number; i++)
  {
	  index[i] = i;	
	}
  return index;
}

/**********************************************************************************/

void CMGMesh::RetrieveMeshLevelHeaders(int level)
{
  try
  {
    MStreamLevel &meshLevel = m_pStreamDescr->GetMeshLevel(level);

    const CPointUVW &nrROIs = meshLevel.GetNrROIs();
    int totalNumROIs = nrROIs.u * nrROIs.v * nrROIs.w;
    if (totalNumROIs > 0)
    {
      m_startVertsROI[level] = new CPtrList[totalNumROIs];
    }
  }
  catch (CMemoryException *e)
  {
    throw CMGDataException(e);
  }
}

/**********************************************************************************/

bool CMGMesh::DecodeMeshConnectivity(int roiIdx, int level)
{
  ASSERT(level >= 0 && level < GetMaxNrLevels());
  MStreamLevel &meshLevel = m_pStreamDescr->GetMeshLevel(level);

  // check the limits
  int count = meshLevel.GetTotalNrROIs();
  if (roiIdx < 0 || roiIdx >= count)
    return false;

  // check the cache whether the ROI connectivity has already been recovered  
  if (meshLevel.IsDoneROIConnectivity(roiIdx))
    return false;

  // initialise the ROI position and size
  InitROI(roiIdx, level);

  // do the decoding of the connectivity if the ROI is not empty
  ResetVisitedVerticesBox(level);
  return RetrieveMeshConnectivityROI(roiIdx, level);
}

bool CMGMesh::DecodeMeshRepositionBit(int roiIdx, int level)
{
  ASSERT(level >= 0 && level < GetMaxNrLevels());
  MStreamLevel &meshLevel = m_pStreamDescr->GetMeshLevel(level);

  // check the limits
  int count = meshLevel.GetTotalNrROIs();
  if (roiIdx < 0 || roiIdx >= count)
    return false;

  // check the cache, the ROI connectivity should have been already recovered  
  if (!meshLevel.IsDoneROIConnectivity(roiIdx))
    return false;

  // initialise the ROI position and size
  InitROI(roiIdx, level);

  // do the decoding of the connectivity if the ROI is not empty
  ResetVisitedVerticesBox(level);
  return RetrieveMeshRepositionBitROI(roiIdx, level);
}

bool CMGMesh::DecodeMeshRefineBits(int roiIdx, int level)
{
  ASSERT(level >= 0 && level < GetMaxNrLevels());
  MStreamLevel &meshLevel = m_pStreamDescr->GetMeshLevel(level);

  // check the limits
  int count = meshLevel.GetTotalNrROIs();
  if (roiIdx < 0 || roiIdx >= count)
    return false;

  // check the cache, the ROI connectivity should have been already recovered
  // but the ROI should not have been repositioned for the next level
  if (!meshLevel.IsDoneROIConnectivity(roiIdx) || meshLevel.IsDoneROIReposition(roiIdx))
    return false;

  // initialise the ROI position and size
  InitROI(roiIdx, level);

  // do the decoding of the connectivity if the ROI is not empty
  ResetVisitedVerticesBox(level);
  return RetrieveMeshRefineBitsROI(roiIdx, level);
}

/**********************************************************************************/

void CMGMesh::DefaultMeshConnectivity(int level)
{
  ASSERT(m_pStreamDescr->GetMeshType() == QUADRI_MESH);

  // reset the flags
  for (int u = 0; u < m_nrSlices.u; u++)
  {
		CSurfNodeList *pList = m_nodeStack[u];
		if (pList == NULL) continue;

    for (int v = 0; v < m_nrSlices.v; v++)
    {
      register CSurfNodeT *pSN = pList[v].GetFirstNode();
      for (; pSN != NULL; pSN = pSN->GetDownList())
      {
				if (pSN->GetLevel() > level)	continue;

				pSN->ResetCount();
        pSN->ResetTriangleType1();  pSN->ResetTriangleType2();
        pSN->ResetHexagonType1();
      }
    }
  }

  // track the patches and insert new vertices
  for (u = 0; u < m_nrSlices.u; u++)
  {
		CSurfNodeList *pList = m_nodeStack[u];
		if (pList == NULL) continue;

    for (int v = 0; v < m_nrSlices.v; v++)
    {
      register CSurfNodeT *pSN = pList[v].GetFirstNode();
      for (; pSN != NULL; pSN = pSN->GetDownList())
      {
				if (pSN->GetLevel() > level)	continue;

        // all polygons have been defined
        if (pSN->GetCount() == 4) continue;

	      CSurfNodeT *pN1 = pSN->GetNextNode1(level);
	      CSurfNodeT *pN2 = pSN->GetNextNode2(level);

	      if (pN1 != NULL && pN2 != NULL)
	      {
		      CSurfNodeT *pN1N2 = pN1->GetNextNode2(level);
		      CSurfNodeT *pN2N1 = pN2->GetNextNode1(level);

		      if (pN1N2 != NULL && pN2N1 != NULL && pN1N2 == pN2N1)
		      {
			      pSN->IncCount();	pN1->IncCount();	pN1N2->IncCount();	pN2->IncCount();
			      
            // add 5 vertices and restore the links
            SubdividePatch(pSN, pN1, pN2, pN1N2, level + 1);
		      }
        }
      }
    }
  }
}

void CMGMesh::SubdividePatch(CSurfNodeT *pSN, CSurfNodeT *pN1, CSurfNodeT *pN2, CSurfNodeT *pN12, int level)
{
  ASSERT(pSN != NULL && pN1 != NULL && pN2 != NULL && pN12 != NULL);

  int dir = pSN->GetDir();
  ASSERT(dir == pN1->GetDir() && dir == pN2->GetDir() && dir == pN12->GetDir());

  int dirSplit1, dirSplit2;
  int splitMode1 = IsSpecialSplit(level, dir, pSN->GetPos(), pN1->GetPos(), dirSplit1);
  int splitMode2 = IsSpecialSplit(level, dir, pSN->GetPos(), pN2->GetPos(), dirSplit2);

  // split at least one direction
  ASSERT(splitMode1 != NO_SPLIT || splitMode2 != NO_SPLIT);

  if (splitMode1 == UNIFORM && splitMode2 == UNIFORM)   // uniform split both directions
  {
    CSurfNodeT *pNS1, *pNS2, *pNS3, *pNS4, *pNS5;

    SplitLink1(pSN, pN1, pNS1, dirSplit1, level);
    SplitLink1(pN2, pN12, pNS4, dirSplit1, level);
    SplitLink2(pSN, pN2, pNS2, dirSplit2, level);
    SplitLink2(pN1, pN12, pNS3, dirSplit2, level);
    SplitLink1(pNS2, pNS3, pNS5, dirSplit1, level);
    SplitLink2(pNS1, pNS4, pNS5, dirSplit2, level);
  }
  else if (splitMode1 == NON_UNIFORM && splitMode2 == UNIFORM)  // special case in direction 1
  {
    CSurfNodeT *pNS11, *pNS12, *pNS2, *pNS3, *pNS41, *pNS42, *pNS51, *pNS52;

    SplitLink1(pSN, pN1, pNS11, pNS12, dirSplit1, level);
    SplitLink1(pN2, pN12, pNS41, pNS42, dirSplit1, level);
    SplitLink2(pSN, pN2, pNS2, dirSplit2, level);
    SplitLink2(pN1, pN12, pNS3, dirSplit2, level);
    SplitLink1(pNS2, pNS3, pNS51, pNS52, dirSplit1, level);
    SplitLink2(pNS11, pNS41, pNS51, dirSplit2, level);
    SplitLink2(pNS12, pNS42, pNS52, dirSplit2, level);
  }
  else if (splitMode1 == UNIFORM && splitMode2 == NON_UNIFORM)  // special case in direction 2
  {
    CSurfNodeT *pNS1, *pNS21, *pNS22, *pNS31, *pNS32, *pNS4, *pNS51, *pNS52;

    SplitLink1(pSN, pN1, pNS1, dirSplit1, level);
    SplitLink1(pN2, pN12, pNS4, dirSplit1, level);
    SplitLink2(pSN, pN2, pNS21, pNS22, dirSplit2, level);
    SplitLink2(pN1, pN12, pNS31, pNS32, dirSplit2, level);
    SplitLink1(pNS21, pNS31, pNS51, dirSplit1, level);
    SplitLink1(pNS22, pNS32, pNS52, dirSplit1, level);
    SplitLink2(pNS1, pNS4, pNS51, pNS52, dirSplit2, level);
  }
  else if (splitMode1 == NO_SPLIT)
  {
    if (splitMode2 == UNIFORM) // uniform in direction 2
    {
      CSurfNodeT *pNS2, *pNS3;

      SplitLink2(pSN, pN2, pNS2, dirSplit2, level);
      SplitLink2(pN1, pN12, pNS3, dirSplit2, level);
      UpdateLink1(pSN, pN1, level);
      UpdateLink1(pNS2, pNS3, level);
      UpdateLink1(pN2, pN12, level);
    }
    else
    {
      ASSERT(splitMode2 == NON_UNIFORM);

      CSurfNodeT *pNS21, *pNS22, *pNS31, *pNS32;

      SplitLink2(pSN, pN2, pNS21, pNS22, dirSplit2, level);
      SplitLink2(pN1, pN12, pNS31, pNS32, dirSplit2, level);
      UpdateLink1(pSN, pN1, level);
      UpdateLink1(pNS21, pNS31, level);
      UpdateLink1(pNS22, pNS32, level);
      UpdateLink1(pN2, pN12, level);
    }
  }
  else if (splitMode2 == NO_SPLIT)
  {
    if (splitMode1 == UNIFORM) // uniform in direction 1
    {
      CSurfNodeT *pNS1, *pNS2;

      SplitLink1(pSN, pN1, pNS1, dirSplit1, level);
      SplitLink1(pN2, pN12, pNS2, dirSplit1, level);
      UpdateLink2(pSN, pN2, level);
      UpdateLink2(pNS1, pNS2, level);
      UpdateLink2(pN1, pN12, level);
    }
    else
    {
      ASSERT(splitMode1 == NON_UNIFORM);

      CSurfNodeT *pNS11, *pNS12, *pNS21, *pNS22;

      SplitLink1(pSN, pN1, pNS11, pNS12, dirSplit1, level);
      SplitLink1(pN2, pN12, pNS21, pNS22, dirSplit1, level);
      UpdateLink2(pSN, pN2, level);
      UpdateLink2(pNS11, pNS21, level);
      UpdateLink2(pNS12, pNS22, level);
      UpdateLink2(pN1, pN12, level);
    }
  }
  else  // special case in both directions
  {
    ASSERT(splitMode1 == NON_UNIFORM && splitMode2 == NON_UNIFORM);

    CSurfNodeT *pNS11, *pNS12, *pNS21, *pNS22, *pNS31, *pNS32;
    CSurfNodeT *pNS41, *pNS42, *pNS51, *pNS52, *pNS61, *pNS62;

    SplitLink1(pSN, pN1, pNS11, pNS12, dirSplit1, level);
    SplitLink1(pN2, pN12, pNS41, pNS42, dirSplit1, level);
    SplitLink2(pSN, pN2, pNS21, pNS22, dirSplit2, level);
    SplitLink2(pN1, pN12, pNS31, pNS32, dirSplit2, level);
    SplitLink1(pNS21, pNS31, pNS51, pNS52, dirSplit1, level);
    SplitLink1(pNS22, pNS32, pNS61, pNS62, dirSplit1, level);
    SplitLink2(pNS11, pNS41, pNS51, pNS61, dirSplit2, level);
    SplitLink2(pNS12, pNS42, pNS52, pNS62, dirSplit2, level);
  }
}

void CMGMesh::GetSplitPos(const CPointUVW &pos1, const CPointUVW &pos2, CPointUVW &pos, int dirSplit, int nextLevel)
{
  ASSERT(nextLevel > 0);

  int value, level = nextLevel - 1;

  pos = pos1;

  switch (dirSplit)
  {
  case DIR_U:
    ASSERT(pos1.v == pos2.v && pos1.w == pos2.w);

    if (pos1.u < pos2.u)
    {
      value = pos1.u;
      if (GetCyclicMode() & CYCLIC_U &&
         (pos1.u == 0 && pos2.u == GetPrevSliceU(GetNrSlicesU() - 1, level)))
      {
        value = pos2.u;
      }
    }
    else
    {
      value = pos2.u;
      if (GetCyclicMode() & CYCLIC_U &&
          (pos2.u == 0 && pos1.u == GetPrevSliceU(GetNrSlicesU() - 1, level)))
      {
        value = pos1.u;
      }
    }

    pos.u = GetNextSliceU(value, nextLevel);
    break;
  case DIR_V:
    ASSERT(pos1.u == pos2.u && pos1.w == pos2.w);

    if (pos1.v < pos2.v)
    {
      value = pos1.v;
      if (GetCyclicMode() & CYCLIC_V &&
         (pos1.v == 0 && pos2.v == GetPrevSliceV(GetNrSlicesV() - 1, level)))
      {
        value = pos2.v;
      }
    }
    else
    {
      value = pos2.v;
      if (GetCyclicMode() & CYCLIC_V &&
          (pos2.v == 0 && pos1.v == GetPrevSliceV(GetNrSlicesV() - 1, level)))
      {
        value = pos1.v;
      }
    }

    pos.v = GetNextSliceV(value, nextLevel);
    break;
  case DIR_W:
    ASSERT(pos1.u == pos2.u && pos1.v == pos2.v);

    if (pos1.w < pos2.w)
    {
      value = pos1.w;
      if (GetCyclicMode() & CYCLIC_W &&
         (pos1.w == 0 && pos2.w == GetPrevSliceW(GetNrSlicesW() - 1, level)))
      {
        value = pos2.w;
      }
    }
    else
    {
      value = pos2.w;
      if (GetCyclicMode() & CYCLIC_W &&
          (pos2.w == 0 && pos1.w == GetPrevSliceW(GetNrSlicesW() - 1, level)))
      {
        value = pos1.w;
      }
    }

    pos.w = GetNextSliceW(value, nextLevel);
    break;
  }
}

void CMGMesh::GetSplitPos(const CPointUVW &pos1, const CPointUVW &pos2, CPointUVW &pos11, CPointUVW &pos22, int dirSplit, int nextLevel)
{
  ASSERT(nextLevel > 0);

  bool flag;
  int value, level = nextLevel - 1;

  pos11 = pos22 = pos1;

  switch (dirSplit)
  {
  case DIR_U:
    ASSERT(pos1.v == pos2.v && pos1.w == pos2.w);

    if (pos1.u < pos2.u)
    {
      value = pos1.u;   flag = true;
      if (GetCyclicMode() & CYCLIC_U &&
         (pos1.u == 0 && pos2.u == GetPrevSliceU(GetNrSlicesU() - 1, level)))
      {
        value = pos2.u;   flag = false;
      }
    }
    else
    {
      value = pos2.u;   flag = false;
      if (GetCyclicMode() & CYCLIC_U &&
          (pos2.u == 0 && pos1.u == GetPrevSliceU(GetNrSlicesU() - 1, level)))
      {
        value = pos1.u;   flag = true;
      }
    }

    if (flag)
    {
      pos11.u = GetNextSliceU(value, nextLevel);
      pos22.u = GetNextSliceU(pos11.u, nextLevel);
    }
    else
    {
      pos22.u = GetNextSliceU(value, nextLevel);
      pos11.u = GetNextSliceU(pos22.u, nextLevel);
    }
    break;
  case DIR_V:
    ASSERT(pos1.u == pos2.u && pos1.w == pos2.w);

    if (pos1.v < pos2.v)
    {
      value = pos1.v;   flag = true;
      if (GetCyclicMode() & CYCLIC_V &&
         (pos1.v == 0 && pos2.v == GetPrevSliceV(GetNrSlicesV() - 1, level)))
      {
        value = pos2.v;   flag = false;
      }
    }
    else
    {
      value = pos2.v;   flag = false;
      if (GetCyclicMode() & CYCLIC_V &&
          (pos2.v == 0 && pos1.v == GetPrevSliceV(GetNrSlicesV() - 1, level)))
      {
        value = pos1.v;   flag = true;
      }
    }

    if (flag)
    {
      pos11.v = GetNextSliceV(value, nextLevel);
      pos22.v = GetNextSliceV(pos11.v, nextLevel);
    }
    else
    {
      pos22.v = GetNextSliceV(value, nextLevel);
      pos11.v = GetNextSliceV(pos22.v, nextLevel);
    }
    break;
  case DIR_W:
    ASSERT(pos1.u == pos2.u && pos1.v == pos2.v);

    if (pos1.w < pos2.w)
    {
      value = pos1.w;   flag = true;
      if (GetCyclicMode() & CYCLIC_W &&
         (pos1.w == 0 && pos2.w == GetPrevSliceW(GetNrSlicesW() - 1, level)))
      {
        value = pos2.w;   flag = false;
      }
    }
    else
    {
      value = pos2.w;   flag = false;
      if (GetCyclicMode() & CYCLIC_W &&
          (pos2.w == 0 && pos1.w == GetPrevSliceW(GetNrSlicesW() - 1, level)))
      {
        value = pos1.w;   flag = true;
      }
    }

    if (flag)
    {
      pos11.w = GetNextSliceW(value, nextLevel);
      pos22.w = GetNextSliceW(pos11.w, nextLevel);
    }
    else
    {
      pos22.w = GetNextSliceW(value, nextLevel);
      pos11.w = GetNextSliceW(pos22.w, nextLevel);
    }
  }
}

void CMGMesh::UpdateLink1(CSurfNodeT *pSN1, CSurfNodeT *pSN2, int level)
{
	pSN1->SetNextNode1(pSN2, level);
	pSN2->SetPrevNode1(pSN1, level);
}

void CMGMesh::UpdateLink2(CSurfNodeT *pSN1, CSurfNodeT *pSN2, int level)
{
	pSN1->SetNextNode2(pSN2, level);
	pSN2->SetPrevNode2(pSN1, level);
}

void CMGMesh::SplitLink1(CSurfNodeT *pSN1, CSurfNodeT *pSN2, CSurfNodeT *&pNS, int dirSplit, int level)
{
  CPointUVW insPos;
  GetSplitPos(pSN1->GetPos(), pSN2->GetPos(), insPos, dirSplit, level);

	pNS = StoreVertex(insPos, level, pSN1->GetDir(), 0.5f);
	pSN1->SetNextNode1(pNS, level);
	pSN2->SetPrevNode1(pNS, level);
	pNS->SetPrevNode1(pSN1, level);
	pNS->SetNextNode1(pSN2, level);
}

void CMGMesh::SplitLink1(CSurfNodeT *pSN1, CSurfNodeT *pSN2, CSurfNodeT *&pNS1, CSurfNodeT *&pNS2, int dirSplit, int level)
{
  CPointUVW insPos1, insPos2;
  GetSplitPos(pSN1->GetPos(), pSN2->GetPos(), insPos1, insPos2, dirSplit, level);

	pNS1 = StoreVertex(insPos1, level, pSN1->GetDir(), 0.5f);
	pNS2 = StoreVertex(insPos2, level, pSN1->GetDir(), 0.5f);
	pSN1->SetNextNode1(pNS1, level);
	pNS1->SetPrevNode1(pSN1, level);
	pNS1->SetNextNode1(pNS2, level);
	pNS2->SetPrevNode1(pNS1, level);
	pNS2->SetNextNode1(pSN2, level);
	pSN2->SetPrevNode1(pNS2, level);
}

void CMGMesh::SplitLink2(CSurfNodeT *pSN1, CSurfNodeT *pSN2, CSurfNodeT *&pNS, int dirSplit, int level)
{
  CPointUVW insPos;
  GetSplitPos(pSN1->GetPos(), pSN2->GetPos(), insPos, dirSplit, level);

	pNS = StoreVertex(insPos, level, pSN1->GetDir(), 0.5f);
	pSN1->SetNextNode2(pNS, level);
	pSN2->SetPrevNode2(pNS, level);
	pNS->SetPrevNode2(pSN1, level);
	pNS->SetNextNode2(pSN2, level);
}

void CMGMesh::SplitLink2(CSurfNodeT *pSN1, CSurfNodeT *pSN2, CSurfNodeT *&pNS1, CSurfNodeT *&pNS2, int dirSplit, int level)
{
  CPointUVW insPos1, insPos2;
  GetSplitPos(pSN1->GetPos(), pSN2->GetPos(), insPos1, insPos2, dirSplit, level);

	pNS1 = StoreVertex(insPos1, level, pSN1->GetDir(), 0.5f);
	pNS2 = StoreVertex(insPos2, level, pSN1->GetDir(), 0.5f);
	pSN1->SetNextNode2(pNS1, level);
	pNS1->SetPrevNode2(pSN1, level);
	pNS1->SetNextNode2(pNS2, level);
	pNS2->SetPrevNode2(pNS1, level);
	pNS2->SetNextNode2(pSN2, level);
	pSN2->SetPrevNode2(pNS2, level);
}

void CMGMesh::RetrieveMeshConnectivityPart(int level)
{
  // reset the streamed flag for the vertices
  if (level > 0) ResetVisitedVerticesGrid(level - 1);

  // retrieve the connectivity of the mesh
  m_boxIndex = 0;
  InitFirstROI(level);

  do
  {
    // retrieve the connectivity of the mesh
    bool bNotEmpty = RetrieveMeshConnectivityROI(m_boxIndex, level);

    // index of the ROI
    ++m_boxIndex;
  }
  while (GetNextROI(level));
}

void CMGMesh::RetrieveVertexRefinePart(int level)
{
  // refine the vertices
  if (m_pStreamDescr->IsRefinementInfoAvailable(level))
  {
    m_boxIndex = 0;
    InitFirstROI(level);

    do
    {
      // retrieve the vertex refine bits
      bool bNotEmpty = RetrieveMeshRefineBitsROI(m_boxIndex, level);

      // index of the ROI
      ++m_boxIndex;
    }
    while (GetNextROI(level));
  }
}

void CMGMesh::RetrieveVertexRepositionPart(int level)
{
  ASSERT (level < GetMaxNrLevels() - 1);

  m_boxIndex = 0;
  InitFirstROI(level);

  do
  {
    // retrieve the vertex reposition bit
    bool bNotEmpty = RetrieveMeshRepositionBitROI(m_boxIndex, level);

    // index of the ROI
    ++m_boxIndex;
  }
  while (GetNextROI(level));
}

bool CMGMesh::RetrieveMeshRepositionBitROI(int roiIdx, int level)
{
  ASSERT(level < GetMaxNrLevels() - 1);

  CPtrList &startNodeList = m_startVertsROI[level][roiIdx];
  if (startNodeList.GetCount() == 0)
    return false;

  MStreamLevel &meshLevel = m_pStreamDescr->GetMeshLevel(level);
  if (meshLevel.GetRepositionBitsMode() == REPOSITION_ANY)
    m_pParser->OnMeshDecodingROIReposition(roiIdx, level);

  // reset the visited vertices
  ResetVisitedVerticesBox(level);

  // loop for all start vertices
  POSITION pos = startNodeList.GetHeadPosition();
  while (pos)
  {
    CSurfNodeT *pSNStart = (CSurfNodeT *)startNodeList.GetNext(pos);

    // write the refine bit
    RecoverVerticesRepositionBit(level, pSNStart, meshLevel.GetRepositionBitsMode());
  }

  // mark as recovered
  meshLevel.DoneROIReposition(roiIdx, true);

  return true;
}

bool CMGMesh::RetrieveMeshRefineBitsROI(int roiIdx, int level)
{
  CPtrList &startNodeList = m_startVertsROI[level][roiIdx];
  if (startNodeList.GetCount() == 0)
    return false;

  ASSERT(m_pStreamDescr->IsRefinementInfoAvailable(level));

  m_pParser->OnMeshDecodingROIRefine(roiIdx, level);

  // read the vertices' refine bits
  int nrBits = GetNrRefBitsVert();
	int t = 1 << (nrBits - 1);
	float maxVal = (float)(1 << nrBits);

  for (int k = nrBits; k >= 1; k--, t >>= 1)
  {
    // reset the visited vertices
    ResetVisitedVerticesBox(level);

    // loop for all start vertices
    POSITION pos = startNodeList.GetHeadPosition();
    while (pos)
    {
      CSurfNodeT *pSNStart = (CSurfNodeT *)startNodeList.GetNext(pos);

      RecoverVerticesRefineBits(level, t, maxVal, pSNStart);
    }
  }

  // mark as recovered
  MStreamLevel &meshLevel = m_pStreamDescr->GetMeshLevel(level);
  meshLevel.DoneROIRefine(roiIdx, true);

  return true;
}

bool CMGMesh::RetrieveMeshConnectivityROI(int roiIdx, int level)
{
  // if the ROI is empty return
  if (!m_pParser->OnMeshDecodingROI(roiIdx, level))
    return false;

  // reset the counter
  m_nrVerticesROI = 0;

  // the mesh description inside a ROI may consists of several subpatches
  bool bFinish = false;
  while (!bFinish)
  {
    // recover the position of the first node and store it
    CSurfNodeT *pSNStart;
    RecoverStartNodeBox(level, pSNStart);

    // we need the starting point when we refine the vertices
    CacheStartPoint(pSNStart, level);

    // read the nodes
    RecoverNodeBox(level, pSNStart);

    m_pParser->EndOfPatch(m_nrVerticesROI, bFinish, roiIdx, level);
  }

  m_nrVertices[level] += m_nrVerticesROI;

  MStreamLevel &meshLevel = m_pStreamDescr->GetMeshLevel(level);
  meshLevel.DoneROIConnectivity(roiIdx, true);

  return true;
}

/**********************************************************************************/

void CMGMesh::RecoverNodeBox(int level, CSurfNodeT *pSNStart)
{
	BYTE dirNext;
	bool dirSwitch;
	CPointUVW pos3D;

	// add the starting vertex to the list
	AddStartNodeBox(pSNStart, true);

	while (m_nodeList.GetCount() != 0)
	{
		CSurfNodeT *pSN = (CSurfNodeT *)m_nodeList.RemoveHead();

		// read the Link1 connectivity vector
		CSurfNodeT *pSNNext1 = NULL;
		ReadConnectivitySymbol(dirNext);			ASSERT(dirNext >= 0 && dirNext <= 3);
		if (dirNext != NO_LINK)
		{
			pos3D = pSN->GetPos();
			UpdateCoordinateFromConnectivity(pSN->GetDir(), dirNext, 1, dirSwitch, pos3D, level);
		
			// store the vertex
			pSNNext1 = StoreVertex(pos3D, level, dirNext, 0.5f);
			pSN->SetNextNode1(pSNNext1, level);

			if (dirSwitch == 1)
				pSNNext1->SetPrevNode2(pSN, level);
			else
				pSNNext1->SetPrevNode1(pSN, level);
		}

		// read the Link2 connectivity vector
		CSurfNodeT *pSNNext2 = NULL;
		ReadConnectivitySymbol(dirNext);			ASSERT(dirNext >= 0 && dirNext <= 3);
		if (dirNext != NO_LINK)
		{
			pos3D = pSN->GetPos();
			UpdateCoordinateFromConnectivity(pSN->GetDir(), dirNext, 2, dirSwitch, pos3D, level);
			
			// store the vertex
			pSNNext2 = StoreVertex(pos3D, level, dirNext, 0.5f);
			pSN->SetNextNode2(pSNNext2, level);

			if (dirSwitch == 1)
				pSNNext2->SetPrevNode1(pSN, level);
			else
				pSNNext2->SetPrevNode2(pSN, level);
		}

		// add to the list
		AddNodesToListBox(pSNNext1, pSNNext2, level, true);
	}
}

void CMGMesh::RecoverStartNodeBox(int level, CSurfNodeT *&pSNStart)
{
	BYTE dir;					CPointUVW pos3D;
	RecoverAbsoluteNodeBox(level, pos3D, dir);     dir <<= OFFSET_DIR;

  pos3D.u += m_localROI.u;
  pos3D.v += m_localROI.v;
  pos3D.w += m_localROI.w;

	// bring the position from level to the max level
	RestorePosNode(pos3D, level);

	pSNStart = StoreVertex(pos3D, level, dir, 0.5f);
}

void CMGMesh::RecoverVerticesRepositionBit(int level, CSurfNodeT *pSNStart, int reposMode)
{
	ASSERT(level < GetMaxNrLevels() - 1);

	// add the starting vertex to the list
	AddStartNodeBox(pSNStart);

	while (m_nodeList.GetCount() != 0)
	{
    // reset the offset of the vertices for the next level
		CSurfNodeT *pSN = (CSurfNodeT *)m_nodeList.RemoveHead();
    pSN->SetOffset(0.5f);

		BYTE symbol;
    if (reposMode == REPOSITION_ANY)
    {
      bool bSpecialCase = IsSpecialCase(level + 1, pSN);
		  ReadRepositionSymbol(symbol, bSpecialCase);
    }
    else
    {
      symbol = reposMode;
    }

		// update the reference planes for each vertex
    switch (symbol)
    {
    case 0: UpdateVertexCoord(level + 1, pSN);  break;  // no reposition
    case 1: RepositionVertex(level + 1, pSN);   break;  // reposition on next -> no special case
    case 2: RepositionVertex(level + 1, pSN);   break;  // reposition on first following -> special case
    case 3: RepositionVertex(level + 1, pSN);
            RepositionVertex(level + 1, pSN);   break;  // reposition on second following -> special case
    }

		AddNodesToListBox(pSN->GetNextNode1(level), pSN->GetNextNode2(level), level);
	}
}

void CMGMesh::RecoverVerticesRefineBits(int level, int t, float maxVal, CSurfNodeT *pSNStart)
{
	// add the starting vertex to the list
	AddStartNodeBox(pSNStart);

	while (m_nodeList.GetCount() != 0)
	{
		CSurfNodeT *pSN = (CSurfNodeT *)m_nodeList.RemoveHead();

		BYTE symbol;
		ReadRefineSymbol(symbol);

		float inc;
		if (symbol == 1)
			inc = (float)(t >> 1) / maxVal;
		else
			inc = -(float)((t + 1) >> 1) / maxVal;

		AddPosOffsetBit(level, pSN, inc);

		AddNodesToListBox(pSN->GetNextNode1(level), pSN->GetNextNode2(level), level);
	}
}

void CMGMesh::AddPosOffsetBit(int level, CSurfNodeT *pSN, float val)
{
	float offset = pSN->GetOffset() + val;
	ASSERT(offset >= 0 && offset <= 1); 
	pSN->SetOffset(offset);

	UpdateVertexCoord(level, pSN);
}

/**********************************************************************************/

void CMGMesh::RecoverAbsoluteNodeBox(int level, CPointUVW &pos3D, BYTE &dir)
{
  m_pParser->GetAbsoluteVertex(pos3D, dir, level);
}

/**********************************************************************************/
	
void CMGMesh::ReadConnectivitySymbol(BYTE &symbol)
{
  m_pParser->GetVertexConnectivity(symbol);
}

void CMGMesh::ReadRefineSymbol(BYTE &symbol)
{
  m_pParser->GetVertexRefineBit(symbol);
}

void CMGMesh::ReadRepositionSymbol(BYTE &symbol, bool bSpecialCase)
{
  m_pParser->GetVertexReposition(symbol, bSpecialCase);
}
