/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

// MGStreamDecoder.cpp: implementation of the CMGStreamParser class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MGStreamParse.h"
#include "MGGridCodec.h"
#include "BitStream.h"
#include "MGErrMsg.h"
#include "MGMesh.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


// parsing the main header of the stream
int CMGStreamParser::ParseMeshGridHeader()
{
  // parse the tag and size
  BYTE tag;   int partSize;
  int nrBytes = ParseTagAndSize(tag, partSize);

  if (tag != MGInfoTag)
    throw CMGDataException(CMGErrMsg::EXPECT_HEADER);

#ifdef _DEBUG
  int startPos = m_streamMG.GetPosition();
#endif

  // parse the mesh description
  ParseNrLevels();
  ParseNrSlices();
  ParseFlags();
  ParseReservedBits();
  ParseMeshType();
  ParseCyclicMode();
  ParseNrRefineBitsVertex();

  // parse the grid description
  ParseFilterType();
  ParseGridCodingParam();
  ParseGridCorners();

  // byte allign the data
  FlushBitStream();

  // store the header size
#ifdef _DEBUG
  int endPos = m_streamMG.GetPosition();
  ASSERT(endPos - startPos == partSize);
#endif
  m_streamDescr.SetHeaderSize(partSize);

  return partSize + nrBytes;
}

// parse the size of the stream
int CMGStreamParser::ParseStreamSize(int &counterLength)
{
  int streamSize;
  counterLength = ParseValue(streamSize);

  // byte allign the data
  FlushBitStream();

  m_streamDescr.SetStreamSize(streamSize);
  return streamSize;
}

// storing the parts identifier and resolution level
int CMGStreamParser::ParseTagAndSize(BYTE &tag, int &size)
{
  int nrBits = TAG_BITS;
  m_streamMG.ReadBits(&tag, TAG_BITS);
  nrBits += ParseValue(size);

  // byte allign the data
  FlushBitStream();

  return AlignInBytes(nrBits);
}

// parse the resolution level
void CMGStreamParser::ParseLevel(int &level)
{
  // store the number of bits allocated for the level
  BYTE nrBits;
  m_streamMG.ReadBits(&nrBits, LEVEL_BITS);

  // write the level
  m_streamMG.ReadBits((DWORD *)&level, nrBits);
}

int CMGStreamParser::ParseValue(int &number)
{
  BYTE bNextByte, value;
	int length = 0;
  number = 0;

	do
	{
    m_streamMG.ReadBit(&bNextByte);
		number <<= 7;
    m_streamMG.ReadBits(&value, 7);
    number |= value;
		length += 8;
	}
	while (bNextByte);

	return length;
}

// parsing mesh related information of the main header of the stream
void CMGStreamParser::ParseNrLevels()
{
  CPointUVW nrLevels;
  Parse3Fields(nrLevels, LEVEL_BITS, CMGErrMsg::WRONG_NR_LEVELS);

	if (nrLevels.u < MIN_LEVEL || 
      nrLevels.v < MIN_LEVEL ||
      nrLevels.w < MIN_LEVEL)
	{
    throw CMGDataException(CMGErrMsg::WRONG_NR_LEVELS);
	}

  // allocate internal arrays
  m_streamDescr.InitMeshLevels(nrLevels);
}

void CMGStreamParser::ParseNrSlices()
{
	// read the number of slices
  CPointUVW  nrSlices; 
  Parse3Fields(nrSlices, SLICES_BITS, CMGErrMsg::WRONG_NR_SLICES);

  CheckNumberSlices(nrSlices);

  m_streamDescr.SetMaxNrSlices(nrSlices);
}

void CMGStreamParser::Parse3Fields(CPointUVW &var, int lenBits, int errNr)
{
  // read the number of bits allocated for the variables
  BYTE nrBits;
  m_streamMG.ReadBits(&nrBits, lenBits);   m_streamDescr.SetNrBitsSlicesField(nrBits);

  if (nrBits == 0)
  {
    throw CMGDataException(errNr);
  }

	// read the fields
  DWORD value;
	m_streamMG.ReadBits(&value, nrBits);    var.u = value;
	m_streamMG.ReadBits(&value, nrBits);    var.v = value;
	m_streamMG.ReadBits(&value, nrBits);    var.w = value;
}

void CMGStreamParser::ParseFlags()
{
  BYTE flag;
  m_streamMG.ReadBit(&flag);    m_streamDescr.HasConnectivityInfo(flag == 1 ? true: false);
  m_streamMG.ReadBit(&flag);    m_streamDescr.HasRefineInfo(flag == 1 ? true: false);
  m_streamMG.ReadBit(&flag);    m_streamDescr.HasRepositionInfo(flag == 1 ? true: false);
  m_streamMG.ReadBit(&flag);    m_streamDescr.HasGridInfo(flag == 1 ? true: false);
  m_streamDescr.HasNormalsInfo(false);
}

void CMGStreamParser::ParseMeshType()
{
  BYTE meshType;
	m_streamMG.ReadBits(&meshType, MESH_TYPE_BITS);    m_streamDescr.SetMeshType(meshType);

  if (meshType == QUADRI_MESH)
  {
    if (m_streamDescr.HasConnectivityInfo())
    {
      BYTE flag;
      m_streamMG.ReadBit(&flag);     m_streamDescr.SetSameConnectivity(flag == 1 ? true: false);
      m_streamMG.ReadBit(&flag);     m_streamDescr.SetAutoResolution(flag == 1 ? true : false);
    }
    else
    {
      m_streamDescr.SetSameConnectivity(true);
      m_streamDescr.SetAutoResolution(true);
    }

    if (m_streamDescr.IsSingleLayered())
    {
      BYTE nrBits;            m_streamMG.ReadBits(&nrBits, 5);
      DWORD offsetAmplitude;  m_streamMG.ReadBits(&offsetAmplitude, nrBits);
      
      m_streamDescr.SetOffsetAmplitude((int)offsetAmplitude);
    }
  }
}

void CMGStreamParser::ParseNrRefineBitsVertex()
{
  // parse whether the refine bits are interleaved in the stream or not
  if (m_streamDescr.HasRefineInfo())
  {
    BYTE flag;
    m_streamMG.ReadBit(&flag);  m_streamDescr.SetInterleaveRefine(flag == TRUE);

    // parse the number of refine bits
    BYTE nrBits;
	  m_streamMG.ReadBits(&nrBits, REFINE_BITS);   m_streamDescr.SetNrRefineBitsVertex(nrBits);

	  if (nrBits == 0 || nrBits > MAX_REFINE_BITS)
	  {
      throw CMGDataException(CMGErrMsg::WRONG_VERTEX_REFINE_BITS);
	  }
  }
  else
  {
    m_streamDescr.SetInterleaveRefine(false);
    m_streamDescr.SetNrRefineBitsVertex(0);
  }
}

void CMGStreamParser::ParseCyclicMode()
{
  BYTE flag;
  m_streamMG.ReadBits(&flag, CYCLIC_MODE_BITS);   m_streamDescr.SetCyclicMode(flag);
}

void CMGStreamParser::ParseReservedBits()
{
  WORD bits;
  m_streamMG.ReadBits(&bits, RESERVED_BITS);
}

// parsing grid related information of the main header of the stream
void CMGStreamParser::ParseFilterType()
{
  BYTE bValue;
  m_streamMG.ReadBits(&bValue, FILTER_BITS);   m_streamDescr.SetFilterType(bValue);
}

void CMGStreamParser::ParseGridCodingParam()
{
  // read the number of bits allocated for the variables
  BYTE nrBits;
  m_streamMG.ReadBits(&nrBits, FIELD_BITS);   m_streamDescr.SetNrBitsGridParam(nrBits);

	if (nrBits == 0 || nrBits > MAX_FIELD_BITS)
	{
    throw CMGDataException(CMGErrMsg::WRONG_NR_SCALE_BITS);
	}

  RGCoding * codingRG = m_streamDescr.GetGridCodingParam();

  // read the scale
  for (int dir = X_GRID; dir <= Z_GRID; dir++)
  {
    DWORD wValue;
    m_streamMG.ReadBits(&wValue, nrBits);   codingRG[dir].scale = wValue;
  }

  if (codingRG[X_GRID].scale < MIN_SCALE ||
      codingRG[Y_GRID].scale < MIN_SCALE ||
      codingRG[Z_GRID].scale < MIN_SCALE)
	{
    throw CMGDataException(CMGErrMsg::WRONG_SCALE_VALUE);
	}

  // prepare the header description of the levels
  InitGridCodingHeader();
}

void CMGStreamParser::ParseGridCorners()
{
  // retrieve the number of quantization bits of the grid values
  BYTE nrQuantBits;
  m_streamMG.ReadBits(&nrQuantBits, QUANT_BITS);

	if (nrQuantBits == 0 || nrQuantBits > MAX_QUANT_BITS)
	{
    throw CMGDataException(CMGErrMsg::WRONG_NR_QUANT_BITS);
	}

  m_streamDescr.SetNrQuantBitsGrid(nrQuantBits);

  // which grid corners are stored in the stream
  int uMax = 1, vMax = 1, wMax = 1;
  const CPointUVW &nrSlices = m_streamDescr.GetMaxNrSlices();
  
  // check for single layered
  if (nrSlices.u == 1) uMax = 0;  // only layer 0
  if (nrSlices.v == 1) vMax = 0;  // only layer 0
  if (nrSlices.w == 1) wMax = 0;  // only layer 0

  // check for cyclic
  int cyclicMode = m_streamDescr.GetCyclicMode();
  if (cyclicMode & 1) { ASSERT(uMax == 1); uMax = 0; }  // only one corner in the u direction
  if (cyclicMode & 2) { ASSERT(vMax == 1); vMax = 0; }  // only one corner in the v direction
  if (cyclicMode & 4) { ASSERT(wMax == 1); wMax = 0; }  // only one corner in the w direction

  // retrieve the grid corner values
  for (int w = 0; w <= wMax; w++)
  {
    for (int v = 0; v <= vMax; v++)
    {
      for (int u = 0; u <= uMax; u++)
      {
        for (int dir = X_GRID; dir <= Z_GRID; dir++)
        {
          RGCoding &codingRG = m_streamDescr.GetGridCodingParam(dir);

          BYTE sgn;     m_streamMG.ReadBit(&sgn);
          DWORD value;  m_streamMG.ReadBits(&value, nrQuantBits);
    
          codingRG.gridCorners[4*w + 2*v + u] = value;
          if (sgn == 1)
            codingRG.gridCorners[4*w + 2*v + u] *= -1;
        }
      }
    }
  }

  // initialise the remaining grid corners
  if (wMax == 0)
  {
    // copy layer 0 to layer 1 such that both are similar
    for (int k = 0; k < 8; k += 8)
    {
      for (int dir = X_GRID; dir <= Z_GRID; dir++)
      {
        RGCoding &codingRG = m_streamDescr.GetGridCodingParam(dir);
        codingRG.gridCorners[k + 4] = codingRG.gridCorners[k + 0];
        codingRG.gridCorners[k + 5] = codingRG.gridCorners[k + 1];
        codingRG.gridCorners[k + 6] = codingRG.gridCorners[k + 2];
        codingRG.gridCorners[k + 7] = codingRG.gridCorners[k + 3];
      }
    }
  }

  if (vMax == 0)
  {
    // copy line 0 to line 1 in both layers
    for (int k = 0; k < 8; k += 4)
    {
      for (int dir = X_GRID; dir <= Z_GRID; dir++)
      {
        RGCoding &codingRG = m_streamDescr.GetGridCodingParam(dir);
        codingRG.gridCorners[k + 2] = codingRG.gridCorners[k + 0];
        codingRG.gridCorners[k + 3] = codingRG.gridCorners[k + 1];
      }
    }
  }

  if (uMax == 0)
  {
    // copy point 0 to point 1 in both lines from the two layers
    for (int k = 0; k < 8; k += 2)
    {
      for (int dir = X_GRID; dir <= Z_GRID; dir++)
      {
        RGCoding &codingRG = m_streamDescr.GetGridCodingParam(dir);
        codingRG.gridCorners[k + 1] = codingRG.gridCorners[k + 0];
      }
    }
  }
}

// parse the stream headers for each level
void CMGStreamParser::ParseMeshGridStream(int streamSize)
{
  int pos;
  while (streamSize > 0)
  {
    // parse the part tag and size of part
    BYTE tag;   int partSize;
    int nrBytes = ParseTagAndSize(tag, partSize);
    streamSize -= (nrBytes + partSize);

#ifdef _DEBUG
  int startPos = GetStreamPositon();
#endif

    // parse the resolution level
    int level;
    ParseLevel(level);

    switch (tag)
    {
	  case MGMeshInfoTag:
      MeshDescriptor(level);
      break;
    case MGGridInfoTag:
      GridDescriptor(level);
      break;
    case MGMeshConnectivityROIInfoTag:
      pos = MeshConnectivityROIHeader(level);
      m_streamMG.SetPosition(pos);
      break;
    case MGMeshConnectivityInfoTag:
      pos = MeshConnectivityHeader(level);
      m_streamMG.SetPosition(pos);
      break;
    case MGVerticesRepositionROIInfoTag:
      pos = VertexRepositionROIHeader(level);
      m_streamMG.SetPosition(pos);
      break;
    case MGVerticesRepositionInfoTag:
      pos = VertexRepositionHeader(level);
      m_streamMG.SetPosition(pos);
      break;
    case MGVerticesRefinementInfoTag:
      pos = VertexRefineHeader(level);
      m_streamMG.SetPosition(pos);
      break;
    case MGVerticesRefinementROIInfoTag:
      pos = VertexRefineROIHeader(level);
      m_streamMG.SetPosition(pos);
      break;
    case MGGridCoefficientsROIInfoTag:
      pos = GridCoefficientsROIHeader(level);
      m_streamMG.SetPosition(pos);
      break;
    case MGGridCoefficientsInfoTag:
      pos = GridCoefficientsHeader(level);
      m_streamMG.SetPosition(pos);
      break;
    default:
      // skip the part
      TRACE0(CMGErrMsg::GetMessage(CMGErrMsg::WRONG_PART_IDENTIFIER));
      m_streamMG.MovePosition(partSize);
    }

#ifdef _DEBUG
  ASSERT(pos == startPos + partSize);
#endif

    EndOfPart();
  }
}

// parsing mesh related information of the level header of the stream
void CMGStreamParser::MeshDescriptor(int level)
{
  // get the current position, needed to update the counters
  int pos = m_streamMG.GetPosition();

  // store the position of the header
  MStreamLevel &meshLevel = m_streamDescr.GetMeshLevel(level);
  meshLevel.SetHeaderPosition(pos);

  // store the number of slices
  CPointUVW nrSlices = m_streamDescr.GetMaxNrSlices();
  const CPointUVW &nrLevels = m_streamDescr.GetNrLevelsMesh();

  CMGGrid::ComputeNrSlicesLevel(level, nrLevels, nrSlices);
  meshLevel.SetNrSlices(nrSlices);

  // retrieve the number of ROIs
  int totalNumROIs = ParseNrROIs(level);
  if (totalNumROIs > 0)
  {
    // number of bits for the ROI indices fields
    int nBitsIndex = GetNrBits(totalNumROIs);
    meshLevel.SetNrBitsROIIndex(nBitsIndex);

    // retrieve the mesh header for the level & byte allign the data
    ParseMeshCodingHeader(level);
  }
  // byte allign the data
  FlushBitStream();

  // store the size of the header
  meshLevel.SetHeaderSize(m_streamMG.GetPosition() - pos);
}

void CMGStreamParser::DefaultMeshDescriptor(int level)
{
  ASSERT(level > 0 && m_streamDescr.IsAutoResolution());
  MStreamLevel &meshLevel0 = m_streamDescr.GetMeshLevel(0);
  const CPointUVW &nrROIs = meshLevel0.GetNrROIs();

  MStreamLevel &meshLevel = m_streamDescr.GetMeshLevel(level);
  meshLevel.Init(nrROIs);

  // store the number of slices
  CPointUVW nrSlices = m_streamDescr.GetMaxNrSlices();
  const CPointUVW &nrLevels = m_streamDescr.GetNrLevelsMesh();

  CMGGrid::ComputeNrSlicesLevel(level, nrLevels, nrSlices);
  meshLevel.SetNrSlices(nrSlices);

  // compute the ROI size
  CPointUVW szROI = nrSlices / nrROIs;

  if ((nrSlices.u >= 2 && szROI.u < MIN_ROI_SIZE) ||
      (nrSlices.v >= 2 && szROI.v < MIN_ROI_SIZE) ||
      (nrSlices.w >= 2 && szROI.w < MIN_ROI_SIZE))
    throw CMGDataException(CMGErrMsg::WRONG_ROI_SIZE);

  // compute the max ROI size
  CPointUVW maxSzROI = nrSlices - szROI * (nrROIs - 1);
  
  // initialise the mesh level
  SetROISize(szROI, maxSzROI, level);
}

int CMGStreamParser::GridDescriptor(int levelG)
{
  // get the current position, needed to update the counters
  int pos = m_streamMG.GetPosition();

  // store the position of the header
  GStreamLevel &gridLevel = m_streamDescr.GetGridLevel(levelG);
  gridLevel.SetHeaderPosition(pos);

  // compute the level corresponding to the mesh
  int nrLevelsM = m_streamDescr.GetMaxNrLevelsMesh();
  int nrLevelsG = m_streamDescr.GetMaxNrLevelsGrid();
  int endLevel  = nrLevelsG - nrLevelsM;

  int levelM = levelG - endLevel;
  if (levelM < 0) levelM = 0;

  // retrieve the number of tiles and the threshold
  int totalNumTiles = ParseNrTiles(levelG, levelM);
  if (totalNumTiles > 0)
  {
    ParseThreshold(levelG);

    // retrieve the grid coding header for the level
    ParseGridCodingHeader(levelG);
  }
  // byte allign the data
  FlushBitStream();

  // store the size of the header
  gridLevel.SetHeaderSize(m_streamMG.GetPosition() - pos);

  return levelM;
}

int CMGStreamParser::MeshConnectivityHeader(int level)
{
  ParseROICodingLookup(level);               FlushBitStream();
  ParseConnectivityCodingHeader(level);      FlushBitStream();

  return UpdateConnectivityFilePointers(level);
}

int CMGStreamParser::VertexRepositionHeader(int level)
{
  ParseRepositionCodingHeader(level);     FlushBitStream();

  return UpdateRepositionFilePointers(level);
}

int CMGStreamParser::VertexRefineHeader(int level)
{
  ParseRefinementCodingHeader(level);     FlushBitStream();

  return UpdateRefineFilePointers(level);
}

int CMGStreamParser::GridCoefficientsHeader(int level)
{
  ParseTilesCodingHeader(level);          FlushBitStream();

  return UpdateTilesFilePointers(level);
}

int CMGStreamParser::ParseNrROIs(int levelM)
{
  ASSERT(levelM >= 0 && levelM < m_streamDescr.GetMaxNrLevelsMesh());

  // read the number of bits allocated for the variables
  MStreamLevel &meshLevel = m_streamDescr.GetMeshLevel(levelM);

  BYTE nrBits;
  m_streamMG.ReadBits(&nrBits, ROI_BITS);    meshLevel.SetNrBitsROIField(nrBits);

	if (nrBits == 0 || nrBits > MAX_ROI_BITS)
	{
    throw CMGDataException(CMGErrMsg::WRONG_NRROI_BITS);
	}

  CPointUVW nrROIs;
  if (nrBits == 0)
  {
    nrROIs.u = nrROIs.v = nrROIs.w = 0;
  }
  else
  {
    // read the number of ROIs
    DWORD size;
    m_streamMG.ReadBits(&size, nrBits);     nrROIs.u = size;
    m_streamMG.ReadBits(&size, nrBits);     nrROIs.v = size;
    m_streamMG.ReadBits(&size, nrBits);     nrROIs.w = size;
  }
  int totalNumROIs = meshLevel.Init(nrROIs);
  if (totalNumROIs == 0)  return 0;

  // compute the ROI size
  const CPointUVW &nrSlices = meshLevel.GetNrSlices();
  CPointUVW szROI = nrSlices / nrROIs;

  if ((nrSlices.u >= 2 && szROI.u < MIN_ROI_SIZE) ||
      (nrSlices.v >= 2 && szROI.v < MIN_ROI_SIZE) ||
      (nrSlices.w >= 2 && szROI.w < MIN_ROI_SIZE))
    throw CMGDataException(CMGErrMsg::WRONG_ROI_SIZE);

  // compute the max ROI size
  CPointUVW maxSzROI = nrSlices - szROI * (nrROIs - 1);
  
  // initialise the mesh level
  SetROISize(szROI, maxSzROI, levelM);

  return totalNumROIs;
}

void CMGStreamParser::ParseMeshCodingHeader(int level)
{
  ASSERT(level >= 0 && level < m_streamDescr.GetMaxNrLevelsMesh());

  MStreamLevel &meshLevel  = m_streamDescr.GetMeshLevel(level);

  if (m_streamDescr.IsRepositionInfoAvailable(level))
  {
    BYTE repositionFlag;
	  m_streamMG.ReadBits(&repositionFlag, REPOS_MODE_BITS);

    meshLevel.SetRepositionBitsMode(repositionFlag);
  }

  // read the number of bits for storing the counters: can be 0 bits
  BYTE nrBits;
  CBoolArray &codingLookup = meshLevel.GetCodingLookup();

  // for the connectivity counters
  if (m_streamDescr.IsConnectivityInfoAvailable(level))
  {
    m_streamMG.ReadBits(&nrBits, VERTS_BITS);
    meshLevel.SetNrBitsConnectivityCounterField(nrBits);

	  if (nrBits > MAX_VERTS_BITS)
    {
      throw CMGDataException(CMGErrMsg::WRONG_NR_VERTEX_BITS);
    }
  }

  if (m_streamDescr.IsRepositionInfoAvailable(level))
  {
    // for the reposition counters
    m_streamMG.ReadBits(&nrBits, VERTS_BITS);
    meshLevel.SetNrBitsVertsReposCounterField(nrBits);

	  if (nrBits > MAX_VERTS_BITS)
    {
      throw CMGDataException(CMGErrMsg::WRONG_NR_VERTEX_BITS);
    }
  }

  if (m_streamDescr.IsRefinementInfoAvailable(level))
  {
    // for the refinement counters
    m_streamMG.ReadBits(&nrBits, VERTS_BITS);
    meshLevel.SetNrBitsVertsRefineCounterField(nrBits);

	  if (nrBits > MAX_VERTS_BITS)
    {
      throw CMGDataException(CMGErrMsg::WRONG_NR_VERTEX_BITS);
    }
  }
}

void CMGStreamParser::ParseROICodingLookup(int level)
{
  ASSERT(level >= 0 && level < m_streamDescr.GetMaxNrLevelsMesh());

  MStreamLevel &meshLevel  = m_streamDescr.GetMeshLevel(level);
  CBoolArray &codingLookup = meshLevel.GetCodingLookup();

  // read the ROI coding lookup
  int nrROIs = meshLevel.GetTotalNrROIs();
  for (int k = 0; k < nrROIs; k++)
  {
    BYTE bValue;  m_streamMG.ReadBit(&bValue);
    codingLookup[k] = bValue != 0 ? true: false;
  }
}

void CMGStreamParser::ParseConnectivityCodingHeader(int level)
{
  ASSERT(level >= 0 && level < m_streamDescr.GetMaxNrLevelsMesh());

  MStreamLevel &meshLevel  = m_streamDescr.GetMeshLevel(level);
  CBoolArray &codingLookup = meshLevel.GetCodingLookup();

  // read the number of bits allocated for the variables
  BYTE nrBits = meshLevel.GetNrBitsConnectivityCounterField();

  // read the counters
  int nrROIs = meshLevel.GetTotalNrROIs();
  for (int k = 0; k < nrROIs; k++)
  {
    if (nrBits > 0 && codingLookup[k])
    {
      DWORD value;      m_streamMG.ReadBits(&value, nrBits);
      meshLevel.SetNrBytesConnectivityCoding(k, value);
    }
    else
    {
      meshLevel.SetNrBytesConnectivityCoding(k, 0);
    }
  }
}

void CMGStreamParser::ParseRepositionCodingHeader(int level)
{
  ASSERT(level >= 0 && level < m_streamDescr.GetMaxNrLevelsMesh());

  if (!m_streamDescr.IsRepositionInfoAvailable(level)) return;

  MStreamLevel &meshLevel  = m_streamDescr.GetMeshLevel(level);
  CBoolArray &codingLookup = meshLevel.GetCodingLookup();

  // read the number of bits allocated for the variables
  BYTE nrBits = meshLevel.GetNrBitsVertsReposCounterField();

  // read the counters
  int nrROIs = meshLevel.GetTotalNrROIs();
  for (int k = 0; k < nrROIs; k++)
  {
    if (nrBits > 0 && codingLookup[k])
    {
      DWORD value;      m_streamMG.ReadBits(&value, nrBits);
      meshLevel.SetNrBytesVertsReposCoding(k, value);
    }
    else
    {
      meshLevel.SetNrBytesVertsReposCoding(k, 0);
    }
  }
}

void CMGStreamParser::ParseRefinementCodingHeader(int level)
{
  ASSERT(level >= 0 && level < m_streamDescr.GetMaxNrLevelsMesh());

  if (!m_streamDescr.IsRefinementInfoAvailable(level)) return;

  MStreamLevel &meshLevel  = m_streamDescr.GetMeshLevel(level);
  CBoolArray &codingLookup = meshLevel.GetCodingLookup();

  // read the number of bits allocated for the variables
  BYTE nrBits = meshLevel.GetNrBitsVertsRefineCounterField();

  // read the counters
  int nrROIs = meshLevel.GetTotalNrROIs();
  for (int k = 0; k < nrROIs; k++)
  {
    if (nrBits > 0 && codingLookup[k])
    {
      DWORD value;      m_streamMG.ReadBits(&value, nrBits);
      meshLevel.SetNrBytesVertsRefineCoding(k, value);
    }
    else
    {
      meshLevel.SetNrBytesVertsRefineCoding(k, 0);
    }
  }
}

// parsing grid related information of the level header of the stream
int CMGStreamParser::ParseNrTiles(int levelG, int levelM)
{
  ASSERT(levelG >= 0 && levelG < m_streamDescr.GetMaxNrLevelsGrid());
  ASSERT(levelM >= 0 && levelM < m_streamDescr.GetMaxNrLevelsMesh());

  // read the number of bits allocated for the variables
  GStreamLevel &gridLevel = m_streamDescr.GetGridLevel(levelG);
  MStreamLevel &meshLevel = m_streamDescr.GetMeshLevel(levelM);

  BYTE nrBits;
  m_streamMG.ReadBits(&nrBits, TILE_BITS);    gridLevel.SetNrBitsTileField(nrBits);

	if (nrBits == 0 || nrBits > MAX_TILE_BITS)
	{
    throw CMGDataException(CMGErrMsg::WRONG_NRTILES_BITS);
	}

  CPointUVW nrTiles;
  if (nrBits == 0)
  {
    nrTiles.u = nrTiles.v = nrTiles.w = 0;
  }
  else
  {
    // read the number of tiles
    DWORD size;
    m_streamMG.ReadBits(&size, nrBits);     nrTiles.u = size;
    m_streamMG.ReadBits(&size, nrBits);     nrTiles.v = size;
    m_streamMG.ReadBits(&size, nrBits);     nrTiles.w = size;
  }
  int totalNumTiles = gridLevel.InitTiles(nrTiles, 0);
  if (totalNumTiles == 0) return 0;

  if (!m_streamDescr.IsAutoResolution())
  {
    const CPointUVW &nrROIs = meshLevel.GetNrROIs();
    if (nrROIs.u < nrTiles.u ||
        nrROIs.v < nrTiles.v ||
        nrROIs.w < nrTiles.w)
      throw CMGDataException(CMGErrMsg::WRONG_NR_TILES);
  }

  CPointUVW nrSlices = meshLevel.GetNrSlices();
  if (levelM == 0)
  {
    int maxNrLevelsGrid = m_streamDescr.GetMaxNrLevelsGrid();
    int maxNrLevelsMesh = m_streamDescr.GetMaxNrLevelsMesh();
    if (maxNrLevelsGrid > maxNrLevelsMesh)
      CMGGrid::ComputeNrSlicesDown(nrSlices, maxNrLevelsGrid - maxNrLevelsMesh - levelG);
  }

  bool bSplit = true;
  if (nrSlices.u < MIN_SPLIT_SIZE &&
      nrSlices.v < MIN_SPLIT_SIZE &&
      nrSlices.w < MIN_SPLIT_SIZE)
  {
    bSplit = false;
    int totalNrTiles = nrTiles.u * nrTiles.v * nrTiles.w;
    if (totalNrTiles > 1)
      throw CMGDataException(CMGErrMsg::WRONG_NR_TILES);
  }

  // initialise the grid level
  gridLevel.SetSplit(bSplit);

  return totalNumTiles;
}

void CMGStreamParser::ParseThreshold(int level)
{
  GStreamLevel &gridLevel = m_streamDescr.GetGridLevel(level);
  for (int dir = X_GRID; dir <= Z_GRID; dir++)
  {
    BYTE bValue;
    m_streamMG.ReadBits(&bValue, FIELD_BITS);
    gridLevel.SetThreshold(dir, bValue != 0 ? (1 << (bValue - 1)) : 0);
  }
}

void CMGStreamParser::ParseGridCodingHeader(int levelG)
{
  ASSERT(levelG >= 0 && levelG < m_streamDescr.GetMaxNrLevelsGrid());

  GStreamLevel &gridLevel = m_streamDescr.GetGridLevel(levelG);

  // read the number of bits allocated for the variables
  BYTE nrBits;
  m_streamMG.ReadBits(&nrBits, GRID_BITS);   gridLevel.SetNrBitsGridCounterField(nrBits);

	if (nrBits > MAX_GRID_BITS)
	{
    throw CMGDataException(CMGErrMsg::WRONG_NR_GRID_BITS);
	}
}

void CMGStreamParser::ParseTilesCodingHeader(int levelG)
{
  ASSERT(levelG >= 0 && levelG < m_streamDescr.GetMaxNrLevelsGrid());

  GStreamLevel &gridLevel = m_streamDescr.GetGridLevel(levelG);

  // the number of bits allocated for the counters
  BYTE nrBits = gridLevel.GetNrBitsGridCounterField();

  // read the counters for the tiles
  for (int k = 0; k < gridLevel.GetTotalNrTiles(); k++)
  {
    for (int dir = X_GRID; dir <= Z_GRID; dir++)
    {
      if (nrBits > 0)
      {
        BYTE bValue;        m_streamMG.ReadBit(&bValue);
        if (bValue == 0)
        {
          gridLevel.SetNrBytesTileCoding(dir, k, 0);
        }
        else
        {
          DWORD value;
          m_streamMG.ReadBits(&value, nrBits);     gridLevel.SetNrBytesTileCoding(dir, k, value);
        }
      }
      else
      {
        gridLevel.SetNrBytesTileCoding(dir, k, 0);
      }
    }
  }
}

/*******************************************************************************/
// Mesh decoding stream access

void CMGStreamParser::GetAbsoluteVertex(CPointUVW &pos3D, BYTE &dir, int level)
{
  MStreamLevel &meshLevel = m_streamDescr.GetMeshLevel(level);
  const CPointUVW &nrBits = meshLevel.GetNrBitsVertPosField();

  DWORD wValue;
  m_streamMG.ReadBits(&wValue, nrBits.u);     pos3D.u = wValue;
  m_streamMG.ReadBits(&wValue, nrBits.v);     pos3D.v = wValue;
  m_streamMG.ReadBits(&wValue, nrBits.w);     pos3D.w = wValue;

  m_streamMG.ReadBits(&dir, DIR_BITS);
}

void CMGStreamParser::GetVertexConnectivity(BYTE &symbol)
{
  if (m_streamDescr.IsSameConnectivity())
  {
	  m_streamMG.ReadBits(&symbol, 1);
    if (symbol == 1) symbol = NO_LINK;
  }
  else
  {
	  m_streamMG.ReadBits(&symbol, BITS_PER_CONNECT);
  }
}

void CMGStreamParser::GetVertexRefineBit(BYTE &symbol)
{
	m_streamMG.ReadBit(&symbol);
}

void CMGStreamParser::GetVertexReposition(BYTE &symbol, bool bSpecialCase)
{
	m_streamMG.ReadBit(&symbol);

  if (symbol == 1 && bSpecialCase)
  {
    BYTE bit;
	  m_streamMG.ReadBit(&bit);
    symbol = 2 | bit;
  }
}
