/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

// MGStreamDescription.cpp: implementation of the CMGStreamDescription class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MGStreamDescription.h"
#include "MGErrMsg.h"
#include "MGStreamParse.h"
#include "MGMesh.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

MStreamLevel::MStreamLevel()
{
  m_headerSz = -1;
  m_headerPos = -1;
  m_nrROIs = -1;
  m_nrSlices = -1;
  m_reposMode = REPOSITION_ANY;

  m_bRequested = false;
  m_bDecoded = false;

  m_nrBitsROIField = -1;
  m_nrBitsROIIndex = -1;
  m_nrBitsVertPos = -1;
  m_nrBitsConnetivityCounter = -1;
  m_nrBitsVertsReposCounter = -1;
  m_nrBitsVertsRefineCounter = -1;

  m_msgs = NULL;
}

int MStreamLevel::Init(const CPointUVW &nrROIs)
{
  int totalNrROIs = 0;
  try
  {
    m_nrROIs = nrROIs;
    totalNrROIs = nrROIs.u * nrROIs.v * nrROIs.w;
    if (totalNrROIs == 0) return 0;

    m_codingLookup.SetSize(totalNrROIs);
    m_connectivityCount.SetSize(totalNrROIs);
    m_vertsReposCount.SetSize(totalNrROIs);
    m_vertsRefineCount.SetSize(totalNrROIs);
    m_posStreamConnectivity.SetSize(totalNrROIs);
    m_posStreamRefine.SetSize(totalNrROIs);
    m_posStreamReposition.SetSize(totalNrROIs);
    m_cacheROI.SetSize(totalNrROIs);
    m_cacheRefine.SetSize(totalNrROIs);
    m_cacheReposition.SetSize(totalNrROIs);

    for (int roiIdx = 0; roiIdx < totalNrROIs; roiIdx++)
    {
      m_codingLookup[roiIdx] = false;
      m_connectivityCount[roiIdx] = 0;
      m_vertsReposCount[roiIdx] = -1;
      m_vertsRefineCount[roiIdx] = -1;
      m_posStreamConnectivity[roiIdx] = -1;
      m_posStreamRefine[roiIdx] = -1;
      m_posStreamReposition[roiIdx] = -1;
      m_cacheROI[roiIdx] = false;
      m_cacheRefine[roiIdx] = false;
      m_cacheReposition[roiIdx] = false;
    }
  }
  catch (CMemoryException *e)
  {
    throw CMGDataException(e);
  }

  return totalNrROIs;
}
  
void MStreamLevel::SetROISize(const CPointUVW &szROI, const CPointUVW &nrBits)
{
  m_szROI = szROI;

  SetNrBitsVertPosField(nrBits);
}

int MStreamLevel::AddMsg(LPCTSTR lpszFormat, ...)
{
  ASSERT(m_msgs != NULL);
  if (!m_msgs)  return -1;

	CString str;

	va_list argList;
	va_start(argList, lpszFormat);
	str.FormatV(lpszFormat, argList);
	va_end(argList);

	return AddMsg(str);
}

void MStreamLevel::DumpStream(CStringArray &msgs)
{
  m_msgs = &msgs;

  // mesh header information
	AddMsg("Nr of Slices = (%d, %d, %d)", m_nrSlices.u, m_nrSlices.v, m_nrSlices.w);

	AddMsg("Nr of bits per <nr ROIs> field = %d", m_nrBitsROIField);
	AddMsg("Nr of bits per <ROI index> field = %d", m_nrBitsROIIndex);
	AddMsg("Nr of ROIs = (%d, %d, %d)", m_nrROIs.u, m_nrROIs.v, m_nrROIs.w);
	AddMsg("Size of ROI = (%d, %d, %d)", m_szROI.u, m_szROI.v, m_szROI.w);

  int repositionFlag = GetRepositionBitsMode();
  switch (repositionFlag)
  {
  case REPOSITION_NO:   AddMsg("No reposition bits are all '0'.");   break;
  case REPOSITION_YES:  AddMsg("No reposition bits are all '1'.");   break;
  case REPOSITION_ANY:  AddMsg("Reposition bits are available.");    break;
  }

	AddMsg("Nr of bits per <Connectivity Coding Size> field = %d", m_nrBitsConnetivityCounter);
	AddMsg("Nr of bits per <Vertices' Reposition Coding Size> field = %d", m_nrBitsVertsReposCounter);
	AddMsg("Nr of bits per <Vertices' Refinement Coding Size> field = %d", m_nrBitsVertsRefineCounter);

  // dump the description for each ROI
  Cube roi;
  CMGMesh::InitFirstROI(roi, m_nrSlices, m_szROI);

  int nrROI = 0;
  do
  {
	  AddMsg("ROI %d at Pos = (%d, %d, %d), Size = (%d, %d, %d), \
            Connectivity (Stream Pos = %d, nr bytes = %d), \
            Reposition (Stream Pos = %d, nr bytes = %d), \
            Refinement (Stream Pos = %d, nr bytes = %d)",
            nrROI, roi.u, roi.v, roi.w, roi.width, roi.height, roi.depth,
            m_posStreamConnectivity[nrROI], m_connectivityCount[nrROI],
            m_posStreamReposition[nrROI], m_vertsReposCount[nrROI],
            m_posStreamRefine[nrROI], m_vertsRefineCount[nrROI]);
    ++nrROI;
  }
  while (CMGMesh::GetNextROI(roi, m_nrSlices, m_szROI));

	AddMsg("Nr of bits per <Start Vertex position> field = (%d, %d, %d)",
          m_nrBitsVertPos.u, m_nrBitsVertPos.v, m_nrBitsVertPos.w);
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

GStreamLevel::GStreamLevel()
{
  m_headerSz = -1;
  m_headerPos = -1;
  m_bSplit = false;
  m_nrTiles = -1;
  m_threshold[0] = m_threshold[1] = m_threshold[2] = -1;

  m_bRequested = false;
  m_bDecoded = false;

  m_nrBitsFieldTile = -1;
  m_nrBitsGridCounter = -1;

  m_msgs = NULL;
}

int GStreamLevel::InitTiles(const CPointUVW &nrTiles, int initValue)
{
  int totalNrTiles = 0;

  try
  {
    m_nrTiles = nrTiles;
    totalNrTiles = GetTotalNrTiles();
    if (totalNrTiles == 0) return 0;

    for (int dir = 0; dir <= 2; dir++)
    {
      CIntArray  &tileCount = m_tilesCount[dir];
      CIntArray  &posStream = m_posStreamTile[dir];
      CBoolArray &cacheTile = m_cacheTile[dir];

      tileCount.SetSize(totalNrTiles);
      posStream.SetSize(totalNrTiles);
      cacheTile.SetSize(totalNrTiles);
      for (int tileIdx = 0; tileIdx < totalNrTiles; tileIdx++)
      {
        cacheTile[tileIdx] = false;
        tileCount[tileIdx] = initValue;   posStream[tileIdx] = -1;
      }
    }
  }
  catch (CMemoryException *e)
  {
    throw CMGDataException(e);
  }

  return totalNrTiles;
}

int GStreamLevel::AddMsg(LPCTSTR lpszFormat, ...)
{
  ASSERT(m_msgs != NULL);
  if (!m_msgs)  return -1;

	CString str;

	va_list argList;
	va_start(argList, lpszFormat);
	str.FormatV(lpszFormat, argList);
	va_end(argList);

	return AddMsg(str);
}

void GStreamLevel::DumpStream(CStringArray &msgs)
{
  m_msgs = &msgs;

  // grid header information
  AddMsg("Threshold X = %d, Y = %d, Z = %d", m_threshold[0], m_threshold[1], m_threshold[2]);

  const CPointUVW &nrTiles = GetNrTiles();
	AddMsg("Nr of Tiles = (%d, %d, %d)", nrTiles.u, nrTiles.v, nrTiles.w);

	AddMsg("Nr of bits per <Grid Size> field = %d", m_nrBitsGridCounter);

  int totalNumTiles = nrTiles.u*nrTiles.v*nrTiles.w;
  for (int tileIdx = 0; tileIdx < totalNumTiles; tileIdx++)
  {
	  AddMsg("Coded Grid size X = %d, Y = %d, Z = %d",
			  		m_tilesCount[0][tileIdx], m_tilesCount[1][tileIdx], m_tilesCount[2][tileIdx]);
	  AddMsg("Coded Grid stream position X = %d, Y = %d, Z = %d",
			  		m_posStreamTile[0][tileIdx], m_posStreamTile[1][tileIdx], m_posStreamTile[2][tileIdx]);
  }
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

MGStreamDescr::MGStreamDescr()
{
  m_headerSz = -1;
  m_maxNrLevelsM = -1;
  m_maxNrLevelsG = -1;
  m_nrLevelsM = -1;
  m_maxNrSlices = -1;
  m_streamSize = -1;
  m_bSingleLayered = false;
  m_bHasConnectivityInfo = false;
  m_bHasRefineInfo = false;  
  m_bHasRepositionInfo = false;  
  m_bHasGridInfo = false;
  m_bHasNormalsInfo = false;
  m_bInterleaveRefine = false;
  m_nrRefBitsVert = -1;
  m_nrQuantBits = -1;
  m_meshType = GENERIC_MESH;
  m_bSameConnect = false;
  m_bAutoResolution = false;
  m_offsetAmp = 0;
  m_cyclicMode = CYCLIC_NONE;

  m_nrBitsFieldSlices = -1;
  m_nrBitsFieldROI = -1;
  m_nrBitsFieldTile = -1;
  m_nrBitsFieldGrid = -1;

  m_msgs = NULL;
}

bool MGStreamDescr::IsConnectivityInfoAvailable(int level) const
{
  return (HasConnectivityInfo() &&
         (level == 0 ||
         (level > 0 && !IsAutoResolution())));
}

bool MGStreamDescr::IsRefinementInfoAvailable(int level) const
{
  int maxLevel = GetMaxNrLevelsMesh() - 1;
  if (HasRefineInfo())
  {
    ASSERT(GetNrRefineBitsVertex() > 0);
    
    return ((level == 0 && IsAutoResolution()) ||
            ((level == maxLevel || InterleaveRefine()) &&
             !IsAutoResolution())); 
  }
  else
  {
    ASSERT(GetNrRefineBitsVertex() == 0);
    ASSERT(InterleaveRefine() == false);
    return false;
  }
}

bool MGStreamDescr::IsRepositionInfoAvailable(int level) const
{
  if (HasRepositionInfo())
  {
    int maxLevel = GetMaxNrLevelsMesh() - 1;
    const MStreamLevel &meshLevel  = m_streamLevelM[level];
    BYTE repositionFlag = meshLevel.GetRepositionBitsMode();

    return (level < maxLevel &&
            repositionFlag == REPOSITION_ANY &&
            !IsAutoResolution());
  }
  else
  {
    return false;
  }
}

bool MGStreamDescr::IsGridInfoAvailable(int level) const
{
  return HasGridInfo();
}

void MGStreamDescr::SetMaxNrSlices(const CPointUVW &nrSlices)
{
  m_maxNrSlices = nrSlices;
  bool singleLayered = (nrSlices.u == 1 || nrSlices.v == 1 || nrSlices.w == 1);
  SetSingleLayered(singleLayered);
}

void MGStreamDescr::InitMeshLevels(const CPointUVW &nrLevels)
{
  m_nrLevelsM = nrLevels;
  m_maxNrLevelsM = max(nrLevels.u, max(nrLevels.v, nrLevels.w));

  m_streamLevelM.SetSize(m_maxNrLevelsM);
}

void MGStreamDescr::InitGridLevels(BYTE nrLevels)
{
  m_maxNrLevelsG = nrLevels;
  m_streamLevelG.SetSize(nrLevels);
}

void MGStreamDescr::SetThresholds(int dir, const CIntArray &thresholds)
{
  for (int level = 0; level < m_maxNrLevelsG; level++)
  {
    GStreamLevel &gridLevel = GetGridLevel(level);
    gridLevel.SetThreshold(dir, thresholds[level]);
  }
}

int MGStreamDescr::AddMsg(LPCTSTR lpszFormat, ...)
{
  ASSERT(m_msgs != NULL);
  if (!m_msgs)  return -1;

	CString str;

	va_list argList;
	va_start(argList, lpszFormat);
	str.FormatV(lpszFormat, argList);
	va_end(argList);

	return AddMsg(str);
}

void MGStreamDescr::DumpStream(CStringArray &msgs)
{
  m_msgs = &msgs;

  // mesh header
  AddMsg(" ");  AddMsg("GLOBAL MESH DESCRIPTION");    AddMsg(" ");

	AddMsg("Nr of Resolution Levels Mesh = (%d, %d, %d)",
          m_nrLevelsM.u, m_nrLevelsM.v, m_nrLevelsM.w);
	AddMsg("Nr of Resolution Levels Grid = %d", m_maxNrLevelsG);

	AddMsg("Nr of bits per <Slice> field = %d", m_nrBitsFieldSlices);
	AddMsg("Total Nr of Slices = (%d, %d, %d)",
				  m_maxNrSlices.u, m_maxNrSlices.v, m_maxNrSlices.w);

  int meshType = GetMeshType();
  switch (meshType)
  {
  case GENERIC_MESH:  AddMsg("Mesh type is generic.");      break;
  case TRI_MESH:      AddMsg("MeshType is triangular.");    break;
  case QUADRI_MESH:   AddMsg("MeshType is quadrilateral."); break;
  case HEXA_MESH:     AddMsg("MeshType is hexagonal.");     break;
  default: ASSERT(0);
  }

  AddMsg("Interlevel Full Refine = %s", m_bInterleaveRefine ? "true" : "false");
  AddMsg("Nr of Refine Bits Vertices = %d", m_nrRefBitsVert);

  // grid header
  AddMsg(" ");  AddMsg("GLOBAL GRID DESCRIPTION");    AddMsg(" ");
  switch (m_filterType)
  {
  case CMGGrid::LINEAR_FILTER:  AddMsg("Interpolation Filter type: LINEAR");  break;
  case CMGGrid::SMOOTH_FILTER:  AddMsg("Interpolation Filter type: SMOOTH");  break;
  default:                      AddMsg("Interpolation Filter type: UNKNOWN");  break;
  }
	AddMsg("Nr of bits per <Grid Scale> field = %d", m_nrBitsFieldGrid);
	AddMsg("Reference Grid X: Scale = %d", m_gridCoding[0].scale);
	AddMsg("Reference Grid Y: Scale = %d", m_gridCoding[1].scale);
	AddMsg("Reference Grid Z: Scale = %d", m_gridCoding[2].scale);

	AddMsg("Nr of bits per <Grid Corner> field = %d", m_nrQuantBits);
  for (int nrCorner = 0; nrCorner < 8; nrCorner++)
  {
	  AddMsg("Grid Corner %d = (%d, %d, %d)", nrCorner,
            m_gridCoding[0].gridCorners[nrCorner],
            m_gridCoding[1].gridCorners[nrCorner], 
            m_gridCoding[2].gridCorners[nrCorner]);
  }

  // mesh level information
  int endLevel = m_maxNrLevelsG - m_maxNrLevelsM;
  for (int level = 0; level < m_maxNrLevelsM; level++)
  {
    AddMsg(" ");  AddMsg("MESH DESCRIPTION LEVEL %d", level);   AddMsg(" ");

    m_streamLevelM[level].DumpStream(msgs);

    // grid level information
    if (level == 0)
    {
      for (int subLevel = 0; subLevel <= endLevel; subLevel++)
      {
        AddMsg(" ");  AddMsg("GRID DESCRIPTION LEVEL %d", subLevel);   AddMsg(" ");
        m_streamLevelG[subLevel].DumpStream(msgs);
      }
    }
    else
    {
      int subLevel = level + endLevel;
      if (subLevel < 0) continue;

      AddMsg(" ");  AddMsg("GRID DESCRIPTION LEVEL %d", subLevel);   AddMsg(" ");
      m_streamLevelG[subLevel].DumpStream(msgs);
    }
  }
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

RGCoding::RGCoding()
{
  scale = -1;

  for (int k = 0; k < 8; k++)
    gridCorners[k] = 0;
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

MGStream::MGStream()
{
  m_pFile = NULL;
  m_pStream = NULL;
}

MGStream::~MGStream()
{
  ASSERT(m_pStream == NULL);
}

void MGStream::Initialise(CFile *file, int op)
{
  ASSERT(m_pStream == NULL);

  m_pFile = file;
	m_pStream = new CBitStream(file, op);
}

CBitStream * MGStream::InitStream(int limit)
{
  ASSERT(m_pStream == 0);

  if (m_pFile != NULL)
    m_pFile->SeekToBegin();
  else
    m_pFile = new CMemFile;

  m_pStream = new CBitStream(m_pFile, CBitStream::OP_WRITE, limit);
  return m_pStream;
}

void MGStream::SetPosition(int pos)
{
  m_pFile->Seek(pos, CFile::begin);
  m_pStream->Flush();
}

void MGStream::MovePosition(int pos)
{
  m_pFile->Seek(pos, CFile::current);
  m_pStream->Flush();
}

void MGStream::SetPositionAndLimit(int pos, int limit, int op)
{
  m_pFile->Seek(pos, CFile::begin);
  delete m_pStream;   m_pStream = NULL;
  m_pStream = new CBitStream(m_pFile, op, limit);
}

bool MGStream::RawRead(BYTE *buf, int pos, int count)
{
  if (!m_pFile) return false;

  try
  {
    m_pFile->Seek(pos, CFile::begin);
    m_pFile->Read(buf, count);
  }
  catch (CFileException *)
  {
    return false;
  }
  return true;
}