/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

// MGStreamParse.cpp: implementation of the CMGStreamParser class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MGStreamParse.h"
#include "MGGridCodec.h"
#include "BitStream.h"
#include "MGErrMsg.h"
#include "ODP.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CMGStreamParser::CMGStreamParser(CFile *file, UINT op)
{
  ASSERT(op == CBitStream::OP_WRITE || op == CBitStream::OP_READ);

  m_pMemStream = NULL;
  m_nrVertRoot = 0;
  m_level = m_roiIdx = -1;
  m_decSpecInfoFile = NULL;

  // create the stream object
  CreateStream(file, op);
}

CMGStreamParser::CMGStreamParser()
{
  m_pMemStream = NULL;
  m_nrVertRoot = 0;
  m_level = m_roiIdx = -1;
  m_decSpecInfoFile = NULL;
}

CMGStreamParser::~CMGStreamParser()
{
  CloseStream();
}

void CMGStreamParser::AttachToStream(CFile *file, UINT op)
{
  ASSERT(op == CBitStream::OP_WRITE || op == CBitStream::OP_READ);

  // create the stream object
  CreateStream(file, op);
}

void CMGStreamParser::Initialise(MeshGridDecoderSpecificInfo *pinfo)
{
  // init the mesh description
  InitNrLevels(pinfo);
  InitNrSlices(pinfo);
  InitFlags(pinfo);
  InitReservedBits(pinfo);
  InitMeshType(pinfo);
  InitCyclicMode(pinfo);
  InitNrRefineBitsVertex(pinfo);

  // init the grid description
  InitFilterType(pinfo);
  InitGridCodingParam(pinfo);
  InitGridCorners(pinfo);
}

// parsing mesh related information from the decoder specific info
void CMGStreamParser::InitNrLevels(MeshGridDecoderSpecificInfo *pinfo)
{
  CPointUVW nrLevels;
  nrLevels.u = pinfo->nLevels[0];
  nrLevels.v = pinfo->nLevels[1];
  nrLevels.w = pinfo->nLevels[2];

	if (nrLevels.u < MIN_LEVEL || 
      nrLevels.v < MIN_LEVEL ||
      nrLevels.w < MIN_LEVEL)
	{
    throw CMGDataException(CMGErrMsg::WRONG_NR_LEVELS);
	}

  // allocate internal arrays
  m_streamDescr.InitMeshLevels(nrLevels);
}

void CMGStreamParser::InitNrSlices(MeshGridDecoderSpecificInfo *pinfo)
{
	// read the number of slices
  CPointUVW  nrSlices;
  nrSlices.u = pinfo->nSlices[0];
  nrSlices.v = pinfo->nSlices[1];
  nrSlices.w = pinfo->nSlices[2];

  CheckNumberSlices(nrSlices);

  m_streamDescr.SetMaxNrSlices(nrSlices);
}

void CMGStreamParser::InitFlags(MeshGridDecoderSpecificInfo *pinfo)
{
  m_streamDescr.HasConnectivityInfo(pinfo->hasConnectivityInfo);
  m_streamDescr.HasRefineInfo(pinfo->hasRefineInfo);
  m_streamDescr.HasRepositionInfo(pinfo->hasRepositionInfo);
  m_streamDescr.HasGridInfo(pinfo->hasGridInfo);
  m_streamDescr.HasNormalsInfo(false);
}

void CMGStreamParser::InitMeshType(MeshGridDecoderSpecificInfo *pinfo)
{
  BYTE meshType = pinfo->meshType;
	m_streamDescr.SetMeshType(meshType);

  if (meshType == QUADRI_MESH)
  {
    if (m_streamDescr.HasConnectivityInfo())
    {
      BYTE flag;
      flag = pinfo->sameBorderOrientation;   m_streamDescr.SetSameConnectivity(flag == 1 ? true: false);
      flag = pinfo->uniformSplit;            m_streamDescr.SetAutoResolution(flag == 1 ? true : false);
    }
    else
    {
      m_streamDescr.SetSameConnectivity(true);
      m_streamDescr.SetAutoResolution(true);
    }

    const CPointUVW & nrSlices = m_streamDescr.GetMaxNrSlices();
    if (m_streamDescr.IsSingleLayered())
    {
      int offsetAmplitude = pinfo->offsetAmplitude;
      m_streamDescr.SetOffsetAmplitude(offsetAmplitude);
    }
  }
}

void CMGStreamParser::InitNrRefineBitsVertex(MeshGridDecoderSpecificInfo *pinfo)
{
  // check whether the refine bits are interleaved in the stream or not
  // init the number of refine bits
  if (m_streamDescr.HasRefineInfo())
  {
    BYTE flag = pinfo->fullRefine;        m_streamDescr.SetInterleaveRefine(flag == TRUE);
    BYTE nrBits = pinfo->nRefineBits;     m_streamDescr.SetNrRefineBitsVertex(nrBits);

	  if (nrBits > MAX_REFINE_BITS)
	  {
      throw CMGDataException(CMGErrMsg::WRONG_VERTEX_REFINE_BITS);
	  }
  }
  else
  {
    m_streamDescr.SetInterleaveRefine(false);
    m_streamDescr.SetNrRefineBitsVertex(0);
  }
}

void CMGStreamParser::InitCyclicMode(MeshGridDecoderSpecificInfo *pinfo)
{
  BYTE flag = pinfo->cyclicMode;   m_streamDescr.SetCyclicMode(flag);
}

void CMGStreamParser::InitReservedBits(MeshGridDecoderSpecificInfo *pinfo)
{
  ASSERT(pinfo->attributes == 0);
}

// init grid related information from the decoder specific info
void CMGStreamParser::InitFilterType(MeshGridDecoderSpecificInfo *pinfo)
{
  BYTE bValue = pinfo->filterType;    m_streamDescr.SetFilterType(bValue);
}

void CMGStreamParser::InitGridCodingParam(MeshGridDecoderSpecificInfo *pinfo)
{
  // read the scale
  RGCoding * codingRG = m_streamDescr.GetGridCodingParam();
  for (int dir = X_GRID; dir <= Z_GRID; dir++)
  {
    DWORD wValue = pinfo->gridScale[dir];   codingRG[dir].scale = wValue;
  }

  if (codingRG[X_GRID].scale < MIN_SCALE ||
      codingRG[Y_GRID].scale < MIN_SCALE ||
      codingRG[Z_GRID].scale < MIN_SCALE)
	{
    throw CMGDataException(CMGErrMsg::WRONG_SCALE_VALUE);
	}

  // prepare the header description of the levels
  InitGridCodingHeader();
}

void CMGStreamParser::InitGridCorners(MeshGridDecoderSpecificInfo *pinfo)
{
  // retrieve the number of quantization bits of the grid values
  BYTE nrQuantBits = pinfo->nQuantBits;
	if (nrQuantBits == 0 || nrQuantBits > MAX_QUANT_BITS)
	{
    throw CMGDataException(CMGErrMsg::WRONG_NR_QUANT_BITS);
	}

  m_streamDescr.SetNrQuantBitsGrid(nrQuantBits);

  // which grid corners are stored in the stream
  int uMax = 1, vMax = 1, wMax = 1;
  const CPointUVW &nrSlices = m_streamDescr.GetMaxNrSlices();
  
  // check for single layered
  if (nrSlices.u == 1) uMax = 0;  // only layer 0
  if (nrSlices.v == 1) vMax = 0;  // only layer 0
  if (nrSlices.w == 1) wMax = 0;  // only layer 0

  // check for cyclic
  int cyclicMode = m_streamDescr.GetCyclicMode();
  if (cyclicMode & 1) { ASSERT(uMax == 1); uMax = 0; }  // only one corner in the u direction
  if (cyclicMode & 2) { ASSERT(vMax == 1); vMax = 0; }  // only one corner in the v direction
  if (cyclicMode & 4) { ASSERT(wMax == 1); wMax = 0; }  // only one corner in the w direction

  // retrieve the grid corner values
  int idx = 0;
  for (int w = 0; w <= wMax; w++)
  {
    for (int v = 0; v <= vMax; v++)
    {
      for (int u = 0; u <= uMax; u++)
      {
        for (int dir = X_GRID; dir <= Z_GRID; dir++)
        {
          RGCoding &codingRG = m_streamDescr.GetGridCodingParam(dir);
          codingRG.gridCorners[4*w + 2*v + u] = pinfo->gridCorners[idx++];
        }
      }
    }
  }

  // initialise the remaining grid corners
  if (wMax == 0)
  {
    // copy layer 0 to layer 1 such that both are similar
    for (int k = 0; k < 8; k += 8)
    {
      for (int dir = X_GRID; dir <= Z_GRID; dir++)
      {
        RGCoding &codingRG = m_streamDescr.GetGridCodingParam(dir);
        codingRG.gridCorners[k + 4] = codingRG.gridCorners[k + 0];
        codingRG.gridCorners[k + 5] = codingRG.gridCorners[k + 1];
        codingRG.gridCorners[k + 6] = codingRG.gridCorners[k + 2];
        codingRG.gridCorners[k + 7] = codingRG.gridCorners[k + 3];
      }
    }
  }

  if (vMax == 0)
  {
    // copy line 0 to line 1 in both layers
    for (int k = 0; k < 8; k += 4)
    {
      for (int dir = X_GRID; dir <= Z_GRID; dir++)
      {
        RGCoding &codingRG = m_streamDescr.GetGridCodingParam(dir);
        codingRG.gridCorners[k + 2] = codingRG.gridCorners[k + 0];
        codingRG.gridCorners[k + 3] = codingRG.gridCorners[k + 1];
      }
    }
  }

  if (uMax == 0)
  {
    // copy point 0 to point 1 in both lines from the two layers
    for (int k = 0; k < 8; k += 2)
    {
      for (int dir = X_GRID; dir <= Z_GRID; dir++)
      {
        RGCoding &codingRG = m_streamDescr.GetGridCodingParam(dir);
        codingRG.gridCorners[k + 1] = codingRG.gridCorners[k + 0];
      }
    }
  }
}

// other functions

void CMGStreamParser::CheckNumberSlices(const CPointUVW  &nrSlices)
{
  int nrLevels = m_streamDescr.GetMaxNrLevelsMesh();
  int maxSlices = INT_MAX >> (nrLevels - 1);
	if (nrSlices.u < MIN_SLICES || nrSlices.u > maxSlices ||
			nrSlices.v < MIN_SLICES || nrSlices.v > maxSlices ||
			nrSlices.w < MIN_SLICES || nrSlices.w > maxSlices)
	{
    throw CMGDataException(CMGErrMsg::WRONG_NR_SLICES);
	}
}

void CMGStreamParser::EndOfPatch(int nrVertices, bool &bFinish, int roiIdx, int level)
{
  ASSERT(level >= 0 && level < m_streamDescr.GetMaxNrLevelsMesh());

  ++m_nrVertRoot;

  MStreamLevel &meshLevel = m_streamDescr.GetMeshLevel(level);
  const CPointUVW &nrBits = meshLevel.GetNrBitsVertPosField();
  int nrBitsConnectivity = m_streamDescr.IsSameConnectivity() ? 1 : BITS_PER_CONNECT;
  int nrBytesRead = (m_nrVertRoot * (nrBits.u + nrBits.v + nrBits.w + DIR_BITS) +
                    nrVertices * 2 * nrBitsConnectivity + 7) / 8;

  int nrBytesROICoding = meshLevel.GetNrBytesConnectivityCoding(roiIdx);

  ASSERT(nrBytesRead <= nrBytesROICoding);
  bFinish = nrBytesRead == nrBytesROICoding;
}

bool CMGStreamParser::IsRemainingEmpty()
{
  BYTE val = m_streamMG.GetRemainingValue();
  ASSERT(val < 128);
  return val == 0;
}

const CPointUVW &CMGStreamParser::GetMaxNrSlices() const
{
  return m_streamDescr.GetMaxNrSlices();
}

int CMGStreamParser::GetMaxNrLevelsMesh() const
{
  return m_streamDescr.GetMaxNrLevelsMesh();
}

const CPointUVW &CMGStreamParser::GetNrLevelsMesh() const
{
  return m_streamDescr.GetNrLevelsMesh();
}

int CMGStreamParser::GetMaxNrLevelsGrid() const
{
  return m_streamDescr.GetMaxNrLevelsGrid();
}

void CMGStreamParser::InitGridCodingHeader()
{
  // compute the number of levels of the grid
  const CPointUVW &nrSlices = m_streamDescr.GetMaxNrSlices();
  int nrLevelsG = CMGGridCodec<float>::GetNrLevelsG(nrSlices);
	
  // initialise the grid codec levels
  m_streamDescr.InitGridLevels(nrLevelsG);
}

void CMGStreamParser::SetROISize(const CPointUVW &szROI, const CPointUVW &maxSzROI, int level)
{
  MStreamLevel &meshLevel = m_streamDescr.GetMeshLevel(level);

  CPointUVW nrBits;
  nrBits.u = GetPower2(maxSzROI.u);
  nrBits.v = GetPower2(maxSzROI.v);
  nrBits.w = GetPower2(maxSzROI.w);

  meshLevel.SetROISize(szROI, nrBits);
}

int CMGStreamParser::UpdateConnectivityFilePointers(int level)
{
  MStreamLevel &meshLevel = m_streamDescr.GetMeshLevel(level);

  int pos = m_streamMG.GetPosition();
  int nrROIs = meshLevel.GetTotalNrROIs();

  // stream position of the connectivity bits
  for (int k = 0; k < nrROIs; k++)
  {
    meshLevel.SetPosStreamConnectivity(k, pos);
    pos += meshLevel.GetNrBytesConnectivityCoding(k);
  }

  return pos;
}

int CMGStreamParser::UpdateRepositionFilePointers(int level)
{
  MStreamLevel &meshLevel = m_streamDescr.GetMeshLevel(level);

  int pos = m_streamMG.GetPosition();
  int nrROIs = meshLevel.GetTotalNrROIs();

  if (m_streamDescr.IsRepositionInfoAvailable(level))
  {
    // stream position of the reposition bits
    for (int k = 0; k < nrROIs; k++)
    {
      meshLevel.SetPosStreamReposition(k, pos);
      pos += meshLevel.GetNrBytesVertsReposCoding(k);
    }
  }

  return pos;
}

int CMGStreamParser::UpdateRefineFilePointers(int level)
{
  MStreamLevel &meshLevel = m_streamDescr.GetMeshLevel(level);

  int pos = m_streamMG.GetPosition();
  int nrROIs = meshLevel.GetTotalNrROIs();

  if (m_streamDescr.IsRefinementInfoAvailable(level))
  {
    // stream position of the refinement bits
    for (int k = 0; k < nrROIs; k++)
    {
      meshLevel.SetPosStreamRefine(k, pos);
      pos += meshLevel.GetNrBytesVertsRefineCoding(k);
    }
  }

  return pos;
}

int CMGStreamParser::UpdateTilesFilePointers(int level)
{
  GStreamLevel &gridLevel = m_streamDescr.GetGridLevel(level);

  int filePos = m_streamMG.GetPosition();
  int nrTiles = gridLevel.GetTotalNrTiles();

  for (int tileIdx = 0; tileIdx < nrTiles; tileIdx++)
  {
    for (int dir = X_GRID; dir <= Z_GRID; dir++)
    {
      gridLevel.SetPosStreamTile(dir, tileIdx, filePos);

      filePos += gridLevel.GetNrBytesTileCoding(dir, tileIdx);
    }
  }

  return filePos;
}

/******************************************************************************/
// Mesh Coding

void CMGStreamParser::OnBeginMeshCodingLevel(int level)
{
  ASSERT(level >= 0 && level < m_streamDescr.GetMaxNrLevelsMesh());
  m_level = level;

  if (m_streamDescr.IsConnectivityInfoAvailable(m_level))
    m_streamConnectivity.InitStream(-1);

  if (m_streamDescr.IsRefinementInfoAvailable(m_level))
    m_streamVertsRefine.InitStream(-1);

  if (m_streamDescr.IsRepositionInfoAvailable(m_level))
    m_streamVertsRepos.InitStream(-1);
}

void CMGStreamParser::OnBeginMeshCodingROI(int roiIdx)
{
  m_roiIdx = roiIdx;
  MStreamLevel &meshLevel = m_streamDescr.GetMeshLevel(m_level);

  meshLevel.SetCodingLookup(roiIdx, false);
  meshLevel.SetNrBytesConnectivityCoding(roiIdx, 0);
  meshLevel.SetNrBytesVertsReposCoding(roiIdx, 0);
  meshLevel.SetNrBytesVertsRefineCoding(roiIdx, 0);
}

void CMGStreamParser::OnBeginMeshCodingConnectivityROI()
{
  m_streamConnectivity.ResetStream();

  ASSERT(m_pMemStream == NULL);
  m_pMemStream = m_streamConnectivity.GetStream();
}

void CMGStreamParser::OnBeginMeshCodingVertsReposROI()
{
  m_streamVertsRepos.ResetStream();

  ASSERT(m_pMemStream == NULL);
  m_pMemStream = m_streamVertsRepos.GetStream();
}

void CMGStreamParser::OnBeginMeshCodingVertsRefineROI()
{
  m_streamVertsRefine.ResetStream();

  ASSERT(m_pMemStream == NULL);
  m_pMemStream = m_streamVertsRefine.GetStream();
}

void CMGStreamParser::OnEndMeshCodingConnectivityROI()
{
  // flush the bitstream buffers and byte align the stream data
  // after calling flush, the bit count will be rounded upwards
  // such that the stream size in bits is a multiple of bytes
  ASSERT(m_pMemStream == m_streamConnectivity.GetStream());
  m_pMemStream->Flush();

  int nrBytes = m_pMemStream->GetBitCount() / 8;
  MStreamLevel &meshLevel = m_streamDescr.GetMeshLevel(m_level);
  meshLevel.SetNrBytesConnectivityCoding(m_roiIdx, nrBytes);

  m_pMemStream = NULL;
}

void CMGStreamParser::OnEndMeshCodingVertsReposROI()
{
  // flush the bitstream buffers and byte align the stream data
  // after calling flush, the bit count will be rounded upwards
  // such that the stream size in bits is a multiple of bytes
  ASSERT(m_pMemStream == m_streamVertsRepos.GetStream());
  m_pMemStream->Flush();

  int nrBytes = m_pMemStream->GetBitCount() / 8;
  MStreamLevel &meshLevel = m_streamDescr.GetMeshLevel(m_level);
  meshLevel.SetNrBytesVertsReposCoding(m_roiIdx, nrBytes);

  m_pMemStream = NULL;
}

void CMGStreamParser::OnEndMeshCodingVertsRefineROI()
{
  // flush the bitstream buffers and byte align the stream data
  // after calling flush, the bit count will be rounded upwards
  // such that the stream size in bits is a multiple of bytes
  ASSERT(m_pMemStream == m_streamVertsRefine.GetStream());
  m_pMemStream->Flush();

  int nrBytes = m_pMemStream->GetBitCount() / 8;
  MStreamLevel &meshLevel = m_streamDescr.GetMeshLevel(m_level);
  meshLevel.SetNrBytesVertsRefineCoding(m_roiIdx, nrBytes);

  m_pMemStream = NULL;
}

void CMGStreamParser::OnEndMeshCodingROI(bool bEmpty)
{
  MStreamLevel &meshLevel = m_streamDescr.GetMeshLevel(m_level);
  meshLevel.SetCodingLookup(m_roiIdx, !bEmpty);
}

void CMGStreamParser::StoreMeshHeaderPart()
{
  // store the mesh header for the level & byte allign the data
  int partSize = 0;

  // compute the size of the part, store tag, size, level and data
  StoreLevel(m_level, &partSize);
  StoreMeshHeaderLevelDetails(m_level, &partSize);

  StoreTagAndSize(MGMeshInfoTag, partSize);

  StoreLevel(m_level, NULL);
  StoreMeshHeaderLevelDetails(m_level, NULL);
}

void CMGStreamParser::StoreMeshConnectivityPart()
{
  int partSize = 0;

  // compute the size of the part, store tag, size, level and data
  StoreLevel(m_level, &partSize);
  StoreROICodingLookup(m_level, &partSize);
  StoreConnectivityCodingHeader(m_level, &partSize);
  StoreCodedConnectivity(m_level, &partSize);

  StoreTagAndSize(MGMeshConnectivityInfoTag, partSize);

  StoreLevel(m_level, NULL);                          
  StoreROICodingLookup(m_level, NULL);                
  StoreConnectivityCodingHeader(m_level, NULL);       
  StoreCodedConnectivity(m_level, NULL);              

  m_streamConnectivity.DeleteStream();
}

void CMGStreamParser::StoreMeshRefinementPart()
{
  // store the coded vertices' refinement bits for the level & byte allign the data
  int partSize = 0;

  // compute the size of the part, store tag, size, level and data
  if (m_streamDescr.IsRefinementInfoAvailable(m_level))
  {
    // store the part identifier and level
    StoreLevel(m_level, &partSize);
    StoreRefinementCodingHeader(m_level, &partSize);
    StoreCodedVertsRefineBits(m_level, &partSize);  

    StoreTagAndSize(MGVerticesRefinementInfoTag, partSize);

    StoreLevel(m_level, NULL);                            
    StoreRefinementCodingHeader(m_level, NULL);           
    StoreCodedVertsRefineBits(m_level, NULL);             
  }
  m_streamVertsRefine.DeleteStream();
}

void CMGStreamParser::StoreMeshRepositionPart()
{
  // store the coded vertices' reposition bits for the level & byte allign the data
  int partSize = 0;

  // compute the size of the part, store tag, size, level and data
  if (m_streamDescr.IsRepositionInfoAvailable(m_level))
  {
    // store the part identifier and level
    StoreLevel(m_level, &partSize);
    StoreRepositionCodingHeader(m_level, &partSize);
    StoreCodedVertsReposBits(m_level, &partSize);   

    StoreTagAndSize(MGVerticesRepositionInfoTag, partSize);

    StoreLevel(m_level, NULL);                       
    StoreRepositionCodingHeader(m_level, NULL);                 
    StoreCodedVertsReposBits(m_level, NULL);                    
  }
  m_streamVertsRepos.DeleteStream();
}

void CMGStreamParser::OnEndMeshCodingLevel()
{
  // do we need to encode something?
  if (!m_streamDescr.IsConnectivityInfoAvailable(m_level) &&
      !m_streamDescr.IsRefinementInfoAvailable(m_level) &&
      !m_streamDescr.IsRepositionInfoAvailable(m_level))
    return;

  // store the commun header
  StoreMeshHeaderPart();

  if (m_streamDescr.IsConnectivityInfoAvailable(m_level))
    StoreMeshConnectivityPart();

  if (m_streamDescr.IsRefinementInfoAvailable(m_level))
    StoreMeshRefinementPart();

  if (m_streamDescr.IsRepositionInfoAvailable(m_level))
    StoreMeshRepositionPart();

  // remove the reference to the stream
  m_pMemStream = NULL;  
}

bool CMGStreamParser::OnMeshDecodingROI(int roiIdx, int level)
{
  // return false if there is no mesh inside the ROI
  if (!SetStreamPosition(Q_VERTEX, level, roiIdx))
    return false;

  // clear the stream buffer
  FlushBitStream();

  // reset the counters that keep track of the amount of vertices
  // that have been recovered
  m_nrVertRoot = 0;

  return true;
}

bool CMGStreamParser::OnMeshDecodingROIRefine(int roiIdx, int level)
{
  // return false if there is no mesh inside the ROI
  if (!SetStreamPosition(Q_REFINE, level, roiIdx))
    return false;

  // clear the stream buffer
  FlushBitStream();

  return true;
}

bool CMGStreamParser::OnMeshDecodingROIReposition(int roiIdx, int level)
{
  // return false if there is no mesh inside the ROI
  if (!SetStreamPosition(Q_REPOS, level, roiIdx))
    return false;

  // clear the stream buffer
  FlushBitStream();

  return true;
}

/******************************************************************************/
// Grid Coding

void CMGStreamParser::OnBeginGridCodingDir(int dir, int level)
{
  ASSERT(dir >= X_GRID && dir <= Z_GRID);

  // initialise the stream to encode the grid
  GStreamLevel &gridLevel = m_streamDescr.GetGridLevel(level);
  m_pMemStream = m_streamGrid[dir].InitStream(gridLevel.GetNrBytesTileCoding(dir, 0) * 8);
}

void CMGStreamParser::OnEndGridCodingDir(int dir)
{
  // delete the stream
  m_streamGrid[dir].DeleteStream();
  m_pMemStream = NULL;
}

void CMGStreamParser::OnBeginGridCodingLevel(int levelG)
{
  ASSERT(levelG >= 0 && levelG < m_streamDescr.GetMaxNrLevelsGrid());
}

void CMGStreamParser::OnEndGridCodingLevel(int levelG)
{
  ASSERT(levelG >= 0 && levelG < m_streamDescr.GetMaxNrLevelsGrid());

  // store the part identifier and level, store the header & byte allign the data
  int partSize = 0;
  StoreLevel(levelG, &partSize);
  StoreGridHeaderLevelDetails(levelG, &partSize);

  StoreTagAndSize(MGGridInfoTag, partSize);      

  StoreLevel(levelG, NULL);
  StoreGridHeaderLevelDetails(levelG, NULL);     

  // store the part identifier and level, store each of the X, Y, Z coded buffers & byte allign the data
  partSize = 0;
  StoreLevel(levelG, &partSize);
  StoreTilesCodingHeader(levelG, &partSize);
  StoreCodedTiles(levelG, &partSize);       

  StoreTagAndSize(MGGridCoefficientsInfoTag, partSize);      

  StoreLevel(levelG, NULL);
  StoreTilesCodingHeader(levelG, NULL);
  StoreCodedTiles(levelG, NULL);       
}

void CMGStreamParser::OnBeginGridCodingTile(int dir)
{
  ASSERT(dir >= X_GRID && dir <= Z_GRID);

  m_streamGrid[dir].ResetStream();
}

void CMGStreamParser::OnEndGridCodingTile(int dir, int levelG, int tileIdx)
{
  ASSERT(dir >= X_GRID && dir <= Z_GRID);
  ASSERT(levelG >= 0 && levelG < m_streamDescr.GetMaxNrLevelsGrid());

  GStreamLevel &gridLevel = m_streamDescr.GetGridLevel(levelG);

  m_pMemStream->Flush();
  int nrBytes = m_pMemStream->GetBitCount() / 8;

  gridLevel.SetNrBytesTileCoding(dir, tileIdx, nrBytes);
}

bool CMGStreamParser::OnBeginGridDecodingTile(QueryType qType, int level, int tileIdx)
{
  return SetStreamPosition(qType, level, tileIdx);
}

void CMGStreamParser::OnEndGridDecodingTile(int tileIdx)
{
  m_pMemStream = NULL;
}

/******************************************************************************/
  
bool CMGStreamParser::SetStreamPosition(QueryType qType, int level, int roiIdx)
{
  // ToDo check if the information exists: the ROI contains information or the tile code != 0
  switch (qType)
  {
  case Q_VERTEX:  return SetStreamPositionConnectivity(level, roiIdx);
  case Q_REFINE:  return SetStreamPositionRefine(level, roiIdx);
  case Q_REPOS:   return SetStreamPositionReposition(level, roiIdx);
  case Q_GRIDX:   return SetStreamPositionGrid(X_GRID, level, roiIdx);
  case Q_GRIDY:   return SetStreamPositionGrid(Y_GRID, level, roiIdx);
  case Q_GRIDZ:   return SetStreamPositionGrid(Z_GRID, level, roiIdx);
  default: ASSERT(0); return false;
  }
}

bool CMGStreamParser::SetStreamPositionGrid(int dir, int level, int tileidx)
{
  GStreamLevel &gridLevel = m_streamDescr.GetGridLevel(level);

  int nrBits = gridLevel.GetNrBytesTileCoding(dir, tileidx) * 8;
  if (nrBits == 0) return false;

  m_streamMG.SetPositionAndLimit(gridLevel.GetPosStreamTile(dir, tileidx),
                                 nrBits, CBitStream::OP_READ);
  m_pMemStream = m_streamMG.GetStream();

  return true;
}

bool CMGStreamParser::SetStreamPositionConnectivity(int level, int roiIdx)
{
  MStreamLevel &meshLevel = m_streamDescr.GetMeshLevel(level);

  if (meshLevel.GetNrBytesConnectivityCoding(roiIdx) == 0)
    return false;

  m_streamMG.SetPositionAndLimit(meshLevel.GetPosStreamConnectivity(roiIdx),
                                 -1, CBitStream::OP_READ);
  m_pMemStream = m_streamMG.GetStream();

  return true;
}

bool CMGStreamParser::SetStreamPositionReposition(int level, int roiIdx)
{
  MStreamLevel &meshLevel = m_streamDescr.GetMeshLevel(level);

  if (meshLevel.GetNrBytesVertsReposCoding(roiIdx) == 0)
    return false;

  m_streamMG.SetPositionAndLimit(meshLevel.GetPosStreamReposition(roiIdx),
                                  -1, CBitStream::OP_READ);
  m_pMemStream = m_streamMG.GetStream();

  return true;
}

bool CMGStreamParser::SetStreamPositionRefine(int level, int roiIdx)
{
  MStreamLevel &meshLevel = m_streamDescr.GetMeshLevel(level);

  if (meshLevel.GetNrBytesVertsRefineCoding(roiIdx) == 0)
    return false;

  m_streamMG.SetPositionAndLimit(meshLevel.GetPosStreamRefine(roiIdx),
                                  -1, CBitStream::OP_READ);
  m_pMemStream = m_streamMG.GetStream();

  return true;
}

/******************************************************************************/

void CMGStreamParser::CloseStream()
{
  m_streamMG.DeleteStream();
}

void CMGStreamParser::CreateStream(CFile *file, UINT op)
{
	// initialize bitStream
  if (file == NULL)
  {
    ASSERT(op == CBitStream::OP_WRITE);

    // write to temporary mem file and only
    // at the end copy the size and content to the file
    m_streamMG.Initialise(&m_memFile, op);
  }
  else
  {
    m_streamMG.Initialise(file, op);
  }
}

/******************************************************************************/

int CMGStreamParser::GetPower2(int value)
{
  int count = 0;
  while(value >= 1) { ++count;    value >>= 1; }
  return count;
}

int CMGStreamParser::GetNrBits(int value)
{
  int count = 1;
  while(value > 1) { ++count;    value >>= 1; }
  return count;
}

int CMGStreamParser::GetNrBits(const CPointUVW &p)
{
  return GetNrBits(max(p.u, max(p.v, p.w)));
}

/******************************************************************************/

void CMGStreamParser::FlushBitStream()
{
	m_streamMG.Flush();
}

/******************************************************************************/

void CMGStreamParser::DumpStream(CStringArray &msgs)
{
  m_streamDescr.DumpStream(msgs);
}

