/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

// MGStreamParse.h: interface for the CMGStreamParser class.
//
//////////////////////////////////////////////////////////////////////

class MeshGridDecoderSpecificInfo;

#if !defined(AFX_MGSTREAMPARSE_H__8973651B_CBD9_4473_BC92_F390AD903F76__INCLUDED_)
#define AFX_MGSTREAMPARSE_H__8973651B_CBD9_4473_BC92_F390AD903F76__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "BasicTypes.h"
#include "MGErrMsg.h"
#include "MGStreamDescription.h"

class CMGMesh;
class CMGGrid;
class CBitStream;

enum RepositionMode {
  REPOSITION_NO = 0,                                        // the reposition bits are all 0
  REPOSITION_YES = 1,                                       // the reposition bits are all 1
  REPOSITION_ANY = 2                                        // no default reposition bits value
};

enum {                                                      // connectivity links values 
  SAME_DIR  = 0,		                                        // 00
  CCW90_DIR	= 1,		                                        // 01
  CW90_DIR  = 2,		                                        // 10
  NO_LINK	  =	3	                                            // 11
};

enum {
  GENERIC_MESH = 0,                                         // The encoded mesh is generic, meaning that the triangulation
                                                            // may contain triangles, quadrilaterals, pentagons, hexagons and heptagons
  TRI_MESH = 1,                                             // The encoded mesh contains only triangles
  QUADRI_MESH = 2,                                          // The encoded mesh contains only quadrilaterals
  HEXA_MESH = 3,                                            // The encoded mesh contains only hexagons
};

enum {
  CYCLIC_NONE = 0,                                          // The encoded mesh is not cyclic
  CYCLIC_U = 1,                                             // The encoded mesh is cyclic in direction "U"
  CYCLIC_V = 2,                                             // The encoded mesh is cyclic in direction "V"
  CYCLIC_UV = 3,                                            // The encoded mesh is cyclic in direction "UV"
  CYCLIC_W = 4,                                             // The encoded mesh is cyclic in direction "W"
  CYCLIC_UW = 5,                                            // The encoded mesh is cyclic in direction "UW"
  CYCLIC_VW = 6,                                            // The encoded mesh is cyclic in direction "VW"
  CYCLIC_UVW = 7                                            // The encoded mesh is cyclic in direction "UVW"
};

enum {
  TAG_BITS = 8,                                             // number of bits to encode the tag of encoded part
  ForbiddenTag = 0,                                         // not allowed
  MGInfoTag = 1,                                            // mesh grid header
  MGMeshInfoTag = 2,                                        // mesh coding header
  MGGridInfoTag = 3,                                        // grid coding header
  MGMeshConnectivityROIInfoTag = 4,                         // mesh connectivity ROI part 
  MGMeshConnectivityInfoTag = 5,                            // mesh connectivity part
  MGVerticesRepositionROIInfoTag = 6,                       // vertices reposition bits ROI part
  MGVerticesRepositionInfoTag = 7,                          // vertices reposition bits part
  MGVerticesRefinementROIInfoTag = 8,                       // vertices refinement bits ROI part
  MGVerticesRefinementInfoTag = 9,                          // vertices refinement bits part
  MGGridCoefficientsROIInfoTag = 10,                        // grid coefficients ROI bitplanes
  MGGridCoefficientsInfoTag = 11                            // grid coefficients bitplanes
};

enum GridCoord {
  X_GRID = 0,                                               // the index of the buffer containing the X coordinates
  Y_GRID = 1,                                               // the index of the buffer containing the Y coordinates
  Z_GRID = 2,                                               // the index of the buffer containing the Z coordinates
};

class CMGStreamParser  
{
public:
  CMGStreamParser(CFile *file, UINT op);
  CMGStreamParser();
  ~CMGStreamParser();
  
  enum {  MG_SIG = 0x10 };
  enum {
          LEVEL_BITS = 5,
          MIN_LEVEL = 1,                  
          MAX_LEVEL_BITS = (1 << LEVEL_BITS) - 1,                            
          SLICES_BITS = 5,                                          
          MAX_SLICES_BITS = (1 << SLICES_BITS) - 1,                 // max value 31
          MIN_SLICES = 1,                                           // min 1 slice
          MAX_SLICES = UINT_MAX,
          REFINE_BITS = 5,
          MAX_REFINE_BITS = (1 << REFINE_BITS) - 1,
          LOD_BITS = 2,                                             // 3 possible values
          FILTER_BITS = 4,                                          // different choices of the filter
                                                                    // 0 - no LOD, 1 - constant density, 2 - constant size
          ROI_BITS = 5,                                             // max value 31
          MAX_ROI_BITS = (1 << ROI_BITS) - 1,
          MAX_ROI_SIZE = UINT_MAX,
          MIN_ROI_SIZE = 2,                                         // min value of the ROI size
                                                                    // and the position of the start Vertex
          TILE_BITS = 5,
          MAX_TILE_BITS = (1 << TILE_BITS) - 1,
          MAX_TILE_SIZE = UINT_MAX,
          FIELD_BITS = 5,
          MAX_FIELD_BITS = (1 << FIELD_BITS) - 1,
          MIN_SCALE = 1,
          MAX_SCALE = UINT_MAX,
          QUANT_BITS = 5,
          MAX_QUANT_BITS = (1 << QUANT_BITS) - 1,                   // max value 31
          MIN_DC_FIELD = INT_MIN,
          MAX_DC_FIELD = INT_MAX,
          GRID_BITS =  5,
          MAX_GRID_BITS = (1 << GRID_BITS) - 1,
          VERTS_BITS = 5,
          MAX_VERTS_BITS = (1 << VERTS_BITS) - 1,
          DIR_BITS = 3,                                             // 6 choices for the absolute direction
          BITS_PER_CONNECT = 2,                                     // nr of bits per vertex for a connectivity
          BITS_PER_REPOSITION  = 1,                                 // nr of bits per vertex inter-level refine (reposition)
          RESERVED_BITS = 9,                                        // reserved bits
          CYCLIC_MODE_BITS = 3,                                     // to store the the cyclic mode
          MESH_TYPE_BITS = 2,                                       // to store the type of mesh
          REPOS_MODE_BITS = 2,                                      // to store the default reposition mode
  };
  enum QueryType {                                                  
          Q_VERTEX = 1,                                             // query vertex connectivity bits
          Q_REFINE = 2,                                             // query vertex refine bits
          Q_REPOS  = 3,                                             // query vertex reposition bits
          Q_GRIDX  = 4,                                             // query grid stream
          Q_GRIDY  = 5,                                             // query grid stream
          Q_GRIDZ  = 6,                                             // query grid stream
  };

public:
  static CMGStreamParser *InitialiseStreamParser(MeshGridDecoderSpecificInfo *pinfo,
                                                 CString &error);            // create and initialise the pasers object
  void AttachToStream(CFile *file, UINT op);                                 // attach the paser to the stream
  void Initialise(MeshGridDecoderSpecificInfo *pinfo);                       // initialise the parser
  static CMGStreamParser *ParseMeshGridHeader(CFile *file, CString &error,
                                              int &headerSize);              // do the parsing of the stream headers

  void  StoreMeshGridHeader(bool keepHeader);
  void  WriteOutput(CFile *file);
  void  StoreValue(MGStream &stream, int number);
  void  StoreTagAndSize(BYTE tag, int size);
  void  StoreLevel(int level, int *partSize);
  int   ParseMeshGridHeader();                                                // parse stream header description
  void  ParseMeshGridStream(int streamSize);                                  // parse the description for each level
  int   ParseStreamSize(int &counterLength);
  int   ParseValue(int &number);
  int   ParseTagAndSize(BYTE &tag, int &size);
  void  ParseLevel(int &level);
  void  MeshDescriptor(int level);                                            // retrieve mesh header level coding details
  void  DefaultMeshDescriptor(int level);                                     // initialise mesh header level for auto resolution
  int   GridDescriptor(int levelG);                                           // retrieve the grid coding level header
  int   MeshConnectivityROIHeader(int level) { return 0; };                                    // retrieve the connectivity coding header
  int   MeshConnectivityHeader(int level);                                    // retrieve the connectivity coding header
  int   VertexRepositionROIHeader(int level) { return 0; };                                    // retrieve the reposition coding header
  int   VertexRepositionHeader(int level);                                    // retrieve the reposition coding header
  int   VertexRefineROIHeader(int level) { return 0; };                                        // retrieve the refine coding header
  int   VertexRefineHeader(int level);                                        // retrieve the refine coding header
  int   GridCoefficientsROIHeader(int level) { return 0; };                                    // retrieve the tiles coding header
  int   GridCoefficientsHeader(int level);                                    // retrieve the tiles coding header
  void  ParseROICodingLookup(int level);                                      // retrieve the ROI coding lookup
  void  ParseConnectivityCodingHeader(int level);                             // retrieve the nr of bytes/ROI for the connectivity coding
  void  ParseRepositionCodingHeader(int level);                               // retrieve the nr of bytes/ROI for the reposition bits coding
  void  ParseRefinementCodingHeader(int level);                               // retrieve the nr of bytes/ROI for the refinment bits coding
  void  ParseTilesCodingHeader(int level);                                    // retrieve the counters for each Tile per level

  void  Attach(BYTE *data, int nrBytes);

  void  ParseMeshGridHeader(BYTE *data, int nrBytes);                         // parse the stream header description from the buffer
  void  ParseMeshLevelHeader(BYTE *data, int nrBytes, int level);             // parse the mesh level header description from the buffer
  void  ParseGridLevelHeader(BYTE *data, int nrBytes, int level);             // parse the grid level header description from the buffer

  bool  ReadMeshGridHeader(BYTE *&buf);                                       // read the stream header into the buffer
  bool  ReadMeshLevelHeader(BYTE *&buf, int level);                           // read the mesh level header into the buffer
  bool  ReadGridLevelHeader(BYTE *&buf, int level);                           // read the grid level header into the buffer

  // retrive the stream header information
  const CPointUVW &GetMaxNrSlices() const;                                    // get the max number of slices
  int   GetMaxNrLevelsMesh() const;                                           // get the max number of resolution levels of the mesh
  const CPointUVW &GetNrLevelsMesh() const;                                   // get the number of resolution levels of the mesh
  int   GetMaxNrLevelsGrid() const;                                           // get the number of resolution levels of the grid

  CBitStream *GetStream() { return m_pMemStream; };
  void  EndOfPatch(int nrVertices, bool &bFinish, int roiIdx, int level);     // mark for the end of a connectivity patch in ROI
  void  EndOfPart() { FlushBitStream(); };                                    // mark for the end of a stream part

  void  SetROISize(const CPointUVW &szROI, const CPointUVW &maxSzROI, int level); // set the size of the ROI per level
                                                                              
  // streming mesh                                                            
  void  OnBeginMeshCodingLevel(int level);                                    // the coding of a new mesh level starts
  void  OnBeginMeshCodingROI(int roiIdx);                                     // the coding of a new mesh ROI starts
  void  OnBeginMeshCodingConnectivityROI();                                   // the coding of the connectivity description starts
  void  OnBeginMeshCodingVertsReposROI();                                     // the coding of the vertices' reposition bits starts
  void  OnBeginMeshCodingVertsRefineROI();                                    // the coding of the vertices' refinement bits starts
  bool  OnMeshDecodingROI(int roiIdx, int level);                             // position the stream for ROI decoding
  bool  OnMeshDecodingROIRefine(int roiIdx, int level);                       // position the stream for ROI Refine decoding
  bool  OnMeshDecodingROIReposition(int roiIdx, int level);                   // position the stream for ROI Reposition decoding
                                                                              
  void  OnEndMeshCodingLevel();                                               // the coding of a new mesh level ends
  void  OnEndMeshCodingROI(bool bEmpty);                                      // the coding of a new mesh ROI ends
	void  OnEndMeshCodingConnectivityROI();                                     // the coding of the connectivity description ends
  void  OnEndMeshCodingVertsReposROI();                                       // the coding of the vertices' reposition bits ends
  void  OnEndMeshCodingVertsRefineROI();                                      // the coding of the vertices' refinement bits ends
                                                                              
  // streaming grid                                                           
  void  OnBeginGridCodingDir(int dir, int leve);                              // the coding of a new grid direction starts
  void  OnBeginGridCodingLevel(int level);                                    // the coding of a new grid resolution level starts
  void  OnBeginGridCodingTile(int dir);                                       // the coding of a new grid tile start
  bool  OnBeginGridDecodingTile(QueryType qType, int level, int tileIdx);     // the decoding of a new grid tile start
                                                                              
  void  OnEndGridCodingDir(int dir);                                          // the coding of a new grid direction ends
  void  OnEndGridCodingLevel(int level);                                      // the coding of a new grid resolution level ends
  void  OnEndGridCodingTile(int dir, int level, int tileIdx);                 // the coding of a new grid tile ends
  void  OnEndGridDecodingTile(int tileIdx);                                   // the decoding of a new grid tile ends

  // position stream Mesh and Grid
  bool  SetStreamPosition(QueryType qType, int level, int roi);               // set the stream position
  bool  SetStreamPositionConnectivity(int level, int roi);                    // set the stream position for mesh connectivity in ROI
  bool  SetStreamPositionRefine(int level, int roi);                          // set the stream position for refining the vertices in ROI
  bool  SetStreamPositionReposition(int level, int roi);                      // set the stream position for repositioning the vertices in ROI
  bool  SetStreamPositionGrid(int dir, int level, int tileidx);               // set the stream position for a grid tile
  void  SetStreamPositon(int pos) { m_streamMG.SetPosition(pos); };           // set the position in the stream
  void  MoveStreamPosition(int size) { m_streamMG.MovePosition(size); };      // move the stream position
  int   GetStreamPositon() const { return m_streamMG.GetPosition(); };        // retrieve the position in the stream
                                                                              
  // read / write mesh coding symbols                                         
  void  GetAbsoluteVertex(CPointUVW &pos3D, BYTE &dir, int level);            // get the absolute position/direction of the starting vertex
  void  GetVertexConnectivity(BYTE &symbol);                                  // get a connectivity symbol for a vertex
  void  GetVertexRefineBit(BYTE &symbol);                                     // get a refine bit for a vertex
  void  GetVertexReposition(BYTE &symbol, bool bSpecialCase);                 // get a refine bit for a vertex
                                                                              
  void  SetAbsoluteVertex(const CPointUVW &pos3D, BYTE dir, int level);
  void  SetVertexConnectivity(BYTE symbol);
  void  SetVertexRefineBit(BYTE symbol);                                      
  void  SetVertexReposition(BYTE symbol, bool bSpecialCase);                                      
                                                                                                                                                            
  void  DumpStream(CStringArray &msgs);                                       // display the fields from the stream

protected:
  void  CreateStream(CFile *file, UINT op);
  void  CloseStream();
  void  InitGridCodingHeader();

  int   AlignInBits(int value) const { return AlignInBytes(value) * 8; };
  int   AlignInBytes(int value) const { return (value + 7) / 8; };
  
  // store / retrieve main MeshGrid information in / from the stream
  void  Store3Fields(const CPointUVW &var, int lenBits, int maxLen, int *partSize); // store a 3-field variable
  void  StoreNrLevels(bool inBinary, bool inASCII, int *partSize);            // store the number of levels mesh
  void  StoreNrSlices(bool inBinary, bool inASCII, int *partSize);            // store the number of slices
  void  StoreMeshType(bool inBinary, bool inASCII, int *partSize);            // store the type of mesh, homogeneous or not
  void  StoreCyclicMode(bool inBinary, bool inASCII, int *partSize);          // store the cyclic mode flag
  void  StoreNrRefineBitsVertex(bool inBinary, bool inASCII, int *partSize);  // store the number of refine bits for the vertices' positions
  void  StoreReservedBits(bool inBinary, bool inASCII, int *partSize);        // store reserved bits
  void  StoreFilterType(bool inBinary, bool inASCII, int *partSize);          // store the filter type
  void  StoreFlags(bool inBinary, bool inASCII, int *partSize);               // store connectivity, refine and grid availability flags
  void  StoreGridCodingParam(bool inBinary, bool inASCII, int *partSize);     // store the grid coding parameters
  void  StoreGridCorners(bool inBinary, bool inASCII, int *partSize);         // store the grid corners bits
                                                                              
  // parse the mesh and grid description from the stream
  void  Parse3Fields(CPointUVW &var, int lenBits, int errNr);                // retrieve a 3-field variable
  void  ParseNrLevels();                                                      // retrieve the number of levels mesh
  void  ParseNrSlices();                                                      // retrieve the number of slices
  void  ParseMeshType();                                                      // retrieve the type of mesh, homogeneous or not
  void  ParseCyclicMode();                                                    // retrieve cyclic mode of the mesh
  void  ParseNrRefineBitsVertex();                                            // retrieve the number of refine bits for the vertices' positions
  void  ParseReservedBits();                                                  // retrieve reserved bits
  void  ParseFilterType();                                                    // retrieve the filter type
  void  ParseFlags();                                                         // retrieve connectivity, refine and grid availability flags
  void  ParseGridCodingParam();                                               // retrieve the grid coding parameters
  void  ParseGridCorners();                                                   // retrieve the grid corners bits

  // init the mesh and grid description from the decoder specific info
  void  InitNrLevels(MeshGridDecoderSpecificInfo *pinfo);                     // retrieve the number of levels mesh
  void  InitNrSlices(MeshGridDecoderSpecificInfo *pinfo);                     // retrieve the number of slices
  void  InitFlags(MeshGridDecoderSpecificInfo *pinfo);                        // retrieve connectivity, refine and grid availability flags
  void  InitMeshType(MeshGridDecoderSpecificInfo *pinfo);                     // retrieve the type of mesh, homogeneous or not
  void  InitCyclicMode(MeshGridDecoderSpecificInfo *pinfo);                   // retrieve cyclic mode of the mesh
  void  InitNrRefineBitsVertex(MeshGridDecoderSpecificInfo *pinfo);           // retrieve the number of refine bits for the vertices' positions
  void  InitReservedBits(MeshGridDecoderSpecificInfo *pinfo);                 // retrieve reserved bits
  void  InitFilterType(MeshGridDecoderSpecificInfo *pinfo);                   // retrieve the filter type
  void  InitGridCodingParam(MeshGridDecoderSpecificInfo *pinfo);              // retrieve the grid coding parameters
  void  InitGridCorners(MeshGridDecoderSpecificInfo *pinfo);                  // retrieve the grid corners bits

  // store / retrieve resolution level mesh information in / from the stream
  void  StoreMeshHeaderPart();                                                // compute size & store mesh header level coding details
  void  StoreMeshConnectivityPart();                                          // compute size & store connectivity coding details
  void  StoreMeshRefinementPart();                                            // compute size & store refinement coding details
  void  StoreMeshRepositionPart();                                            // compute size & store reposition coding details

  void  StoreMeshHeaderLevelDetails(int level, int *partSize);                // store mesh header level coding details
  bool  StoreNrROIs(int level, int *partSize);                                // store the number of ROIs
  void  StoreMeshCodingHeader(int level, int *partSize);                      // store a lookup of coded ROIs
  void  StoreROICodingLookup(int level, int *partSize);                       // store the ROI coding lookup
  void  StoreConnectivityCodingHeader(int level, int *partSize);              // store the nr of bytes/ROI for the connectivity coding
  void  StoreRepositionCodingHeader(int level, int *partSize);                // store the nr of bytes/ROI for the reposition bits coding
  void  StoreRefinementCodingHeader(int level, int *partSize);                // store the nr of bytes/ROI for the refinrment bits coding
  void  StoreCodedConnectivity(int level, int *partSize);                     // store the coded connectivity bits
  void  StoreCodedVertsReposBits(int level, int *partSize);                   // store the coded vertices' reposition bits
  void  StoreCodedVertsRefineBits(int level, int *partSize);                  // store the coded vertices' refinement bits
                                                                              
  int   ParseNrROIs(int level);                                               // retrieve the number of the ROIs
  void  ParseMeshCodingHeader(int level);                                     // retrieve the coded mesh bits

  // store / retrieve resolution level grid information in / from the stream
  void  StoreGridHeaderLevelDetails(int level, int *partSize);                // store grid header level coding details
  bool  StoreNrTiles(int level, int *partSize);                               // store the number of tiles
  void  StoreThreshold(int level, int *partSize);                             // store the threshold for the coded grid
  void  StoreGridCodingHeader(int level, int *partSize);                      // store the nr of bits for each Tile per level
  void  StoreTilesCodingHeader(int level, int *partSize);                     // store the lookup and counters for each Tile per level
  void  StoreCodedTiles(int level, int *partSize);                            // store the coded grid bits
  
  int   ParseNrTiles(int levelG, int levelM);                                 // retrieve the number of tiles
  void  ParseThreshold(int level);                                            // retrieve the threshold for the coded grid
  void  ParseGridCodingHeader(int level);                                     // retrieve the nr of bits for each Tile per level

  // updating references to the stream
  int   UpdateConnectivityFilePointers(int level);
  int   UpdateRepositionFilePointers(int level);
  int   UpdateRefineFilePointers(int level);
  int   UpdateTilesFilePointers(int level);

  // process the ROIs
	void  FlushBitStream();
  int   GetNrBits(int value);
  int   GetPower2(int value);
  int   GetNrBits(const CPointUVW &value);
  void  CheckNumberSlices(const CPointUVW  &nrSlices);
  bool  IsRemainingEmpty();

public:
  MGStreamDescr   m_streamDescr;        // description of the MeshGrid stream

protected:
  int             m_level;              // the resolution level being coded
  int             m_roiIdx;             // index of the ROI being coded
  CMemFile        m_memFile;            // memory file for view-dependent decoding
  CBitStream    * m_pMemStream;         // temporary stream pointing to the currently selected stream

  MGStream        m_streamMG;           // the stream containing the coded MeshGrid
  MGStream        m_streamConnectivity; // temporary stream for storing the coded mesh
  MGStream        m_streamVertsRepos;   // temporary stream for storing the coded mesh
  MGStream        m_streamVertsRefine;  // temporary stream for storing the coded mesh
  MGStream        m_streamGrid[3];      // temporary stream for storing the coded grid

  int             m_nrVertRoot;         // keep the number of root vertices in a ROI

  FILE          * m_decSpecInfoFile;    // dump the decoder specific info
};

inline CMGStreamParser *CMGStreamParser::InitialiseStreamParser(MeshGridDecoderSpecificInfo *pinfo, CString &error)
{
  CMGStreamParser *pParser = NULL;

  // create the parser object
	try
	{
    pParser = new CMGStreamParser();
	}
	catch (CMemoryException *e)
	{
		GetErrorMessage(e, error);  return NULL;
	}

  try
  {
    // do the parsing of the stream headers
    pParser->Initialise(pinfo);
  }
  catch (CMGDataException &e)
  {
    error = e.GetErrorMessage();  return NULL;
  }

  return pParser;
}

inline CMGStreamParser *CMGStreamParser::ParseMeshGridHeader(CFile *file, CString &error, int &headerSize)
{
  CMGStreamParser *pParser = NULL;

  // create the parser object
	TRY
	{
    pParser = new CMGStreamParser(file, CBitStream::OP_READ);
	}
	CATCH_ALL(e)
	{
		GetErrorMessage(e, error);  return NULL;
	}
	END_CATCH_ALL

  try
  {
    // do the parsing of the stream headers
    headerSize = pParser->ParseMeshGridHeader();
  }
  catch (CMGDataException &e)
  {
    error = e.GetErrorMessage();
    delete pParser;   return NULL;
  }

  return pParser;
}

#endif // !defined(AFX_MGSTREAMPARSE_H__8973651B_CBD9_4473_BC92_F390AD903F76__INCLUDED_)
