/*
	This software module was originally developed by SGDL Systems Inc. in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1)
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license
to this software module or modifications thereof for use in hardware or software
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1).
Those intending to use this software module in hardware or software products are
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability for use of
this software module or modifications thereof in an implementation. Copyright is
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.
	SGDL Systems Inc. retains full right to use the code for his/her own purpose, assign or
donate the code to a third party and to inhibit third parties from using the code
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
						Copyright  2001
*/

//*******************************************************************
// SGDLNode
//*******************************************************************

#include "Opengl/AFX/Solid/VolumeTreeNode/SGDLTools.h"
#include "Opengl/AFX/Solid/SGDLPolygonise.h"

//*******************************************************************
//  SGDLNode constructor
//
//  By Eric Paquet
//  SGDL Systems inc 
//  Date : December 10th, 2001.
//*******************************************************************
SGDLNode::SGDLNode(const double coeff_ptr[],
			       int          coeff_number,
			       bool         is_solid,
				   const float  bbox_ptr[],
				   int          Node_Density[2]
                   ):
Volume_Tree_Node(SGDL_NODE, bbox_ptr)
{
	assert(coeff_ptr	!= NULL &&
	       bbox_ptr	    != NULL && 
		   (coeff_number == K_QuadricNumberOfCoeff ||
		    coeff_number == K_QuarticNumberOfCoeff));

	init();

    coeff = new double[coeff_number];

	//Get a copy of the coefficients
	for ( int i = 0; i < coeff_number; i++)
        coeff[i] = coeff_ptr[i];

	nCoeff		  = coeff_number;
	solid		  = is_solid;
    density[0]	  = Node_Density[0];	
    density[1]	  = Node_Density[1];
    Manage_At_The_Top();
}

SGDLNode::SGDLNode(const float coeff_ptr[],
			       int         coeff_number,
			       bool        is_solid,
		           const float bbox_ptr[],
		           int         Node_Density[2]):
Volume_Tree_Node(SGDL_NODE, bbox_ptr)
{
	assert(coeff_ptr     != NULL &&
	       bbox_ptr      != NULL && 
		   (coeff_number == K_QuadricNumberOfCoeff ||
		    coeff_number == K_QuarticNumberOfCoeff));

    init();
	coeff = new double[coeff_number];

	//Get a copy of the coefficients
	for ( int i = 0; i < coeff_number; i++)
		coeff[i] = static_cast<double>(coeff_ptr[i]);

	nCoeff		= coeff_number;
	solid		= is_solid;  
    density[0]	= Node_Density[0];
    density[1]	= Node_Density[1];
	Manage_At_The_Top();
}

//******************************************************************
//	This function takes a point x, y, z and return true if the point
//	is inside the solid and false if the volume is outside.
//
//	Author : Eric Paquet
//	Date : October 9th, 2001
//	Modif : chloe ... for hybrid uses SGDL_GetDensity
//******************************************************************
int SGDLNode::EvaluateDensity(const double& x, const double& y, const double& z)const
{    
	assert(coeff != NULL &&
	      (nCoeff == K_QuadricNumberOfCoeff ||
	       nCoeff == K_QuarticNumberOfCoeff));

    int value = 2;

    if (solid)
    {
		//This call may modify is_inside
		Bounding_Box_Correction(bbox, &value, x, y, z);
    }

    if(value != 0)
    {
        if ( nCoeff == K_QuadricNumberOfCoeff )
		    value = DLz2_coe(coeff, x, y, z);
	    else if ( nCoeff == K_QuarticNumberOfCoeff )
		    value = DLz4_coe(coeff, x, y, z);
        if (density[K_INT_INTERIOR_INDEX] != -1 && value > 0)
		    value = density[K_INT_INTERIOR_INDEX];
    }
    
    return value;
}

//*******************************************************************
//  SGDLNode destructor 
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
SGDLNode::~SGDLNode(void)
{
    if ( ref == 0 && coeff != NULL)
        delete coeff;
}

//***************************************************
//	
//  SGDLNode::EvaluateGradient()
//
//	Author : Simon Lavallee, SGDL Systems
//	Date   : March 27th, 2002
//
//  Input  : double x           -> X coordinate
//           double y           -> Y coordinate
//           double z           -> Z coordinate
//
//  Output : double MinValue    -> FieldValue of the implicit function
//           float  Color       -> Color of the point
//           double *Gradient   -> Gradient of the function
//
//  Desc   : This function evaluate the gradient
//           of the imlicit function at point x, y, z.
//           The gradient of the function return a vector
//           that is normal to the surface of the implicit
//           function.
//***************************************************
bool SGDLNode::EvaluateGradient(const double& x,  const double& y, const double& z, 
                                double& Distance, double *Normal,   float *Color) const
{

    assert(coeff  != NULL &&
          (nCoeff == K_QuadricNumberOfCoeff || 
           nCoeff == K_QuarticNumberOfCoeff));
    
    double NormalLength = 0.0;
    double ToHighToBeTheGoodOne = 0.0;

    if (nCoeff == K_QuadricNumberOfCoeff)
    {
        //Get field value of the quadric (very approximative distance)
        Distance = DLz2_coeR(coeff, x, y, z);
        //Get the normal, using the gradient of the quadric
        Dlz2_Gradient(coeff, x, y, z, Normal);

        ToHighToBeTheGoodOne = 1.0;
    }
    
	else if (nCoeff == K_QuarticNumberOfCoeff)
    {
        //Get field value of the quartic (very approximative distance)
	    Distance = DLz4_coeR(coeff, x, y, z);
        //Get the normal, using the gradient of the quartic
        Dlz4_Gradient(coeff, x, y, z, Normal);

        ToHighToBeTheGoodOne = 10.0;
    }
    
    //For now, the distance we have is the field value of the 
    //implicit function that is actually not a very good approximation
    //of the distance between the point and the surface. The next step 
    //will project the point to the surface along the normal and return 
    //the distance between the initial and the final point.    
    
    if ((fabs(Distance) < ToHighToBeTheGoodOne) && !EpsilonEqual(Distance, 0.0))
    {
        //We compute the distance between the point and the surface only
        //if the field value isn't greater than one (that is certainly too 
        //far to be the good one) or isn't already on the surface (== 0.0).
        
        NormalLength = sqrt(Normal[0]*Normal[0] + 
                            Normal[1]*Normal[1] + 
                            Normal[2]*Normal[2]);

        //Normalize the vector
        Normal[0] /= NormalLength;
        Normal[1] /= NormalLength;
        Normal[2] /= NormalLength;

        //Get the distance between the point and the surface
        Distance = GetDistanceToSurface(coeff, nCoeff, Normal, 
                                        Distance, x, y, z);
    }

    if (solid && EpsilonGreaterOrEqual(Distance, 0.0))
    {
        //The gradient and the distance can be modified by the bbox.
        Bounding_Box_Gradient_Correction(bbox, Normal, &Distance, x, y, z);        
    }

    if (Color && !Undefined_Color(color))
    {
        //Get the color
        Color[0] = color[0];
        Color[1] = color[1];
        Color[2] = color[2];
    }

    return true;
}

//*******************************************************************
//  This function manages the variable At_The_Top
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
void SGDLNode::Manage_At_The_Top()
{
	At_The_Top = true;
}

//****************************************************
//	
//  Get the color of the point, based on the fact
//  that the point is in or outside the shape.
//
//	Author : Eric Baillargeon, SGDL Systems
//	Date : October 15th, 2001
//  Modified: Simon Lavallee, SGDL Systems
//            April 15th, 2002
//
//  Input  : x,y,z   
//
//  Output : float set_color[3] -> color to set
//           bool  is_inside    -> true if the point is inside
//                                 the shape.
//
//***************************************************
bool SGDLNode::GetColor(const double &x, const double &y, const double &z, float set_color[])const
{
    bool is_inside = isInsideWithoutDensityCheck(x,y,z);
    
	//Set the color
	if ( is_inside && !Undefined_Color(color) )
	{
		set_color[0] = color[0];
		set_color[1] = color[1];
		set_color[2] = color[2];
	}

    return is_inside;
}

//*******************************************************************
//  This function sets the solidity of the SGDLNode
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
void SGDLNode::Set_Solidity(bool to_set)
{
	solid = to_set;
}

//*******************************************************************
//  This function the SGDLNode coeff(use for the gradient)
//
//  By Eric Baillargeon
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
double * SGDLNode::GetCoeff(double x, double y , double z)
{
    return coeff;
}

//*******************************************************************
//  This function the SGDLNode number of coeff (use for the gradient)
//
//  By Eric Baillargeon
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
int  SGDLNode::GetNumOfCoeff(double x, double y , double z)
{
    return nCoeff;
}

//*******************************************************************
//  This function removes a reference from the SGDLNode 
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
void SGDLNode::UnRef()
{
    ref--;

    if ( ref == 0 )
        delete this;
}

//EOF***************************************************************