/*
	This software module was originally developed by SGDL Systems Inc. in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1)
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license
to this software module or modifications thereof for use in hardware or software
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1).
Those intending to use this software module in hardware or software products are
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability for use of
this software module or modifications thereof in an implementation. Copyright is
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.
	SGDL Systems Inc. retains full right to use the code for his/her own purpose, assign or
donate the code to a third party and to inhibit third parties from using the code
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
						Copyright  2001
*/

//*******************************************************************
//Volume_Tree_Matrix
//*******************************************************************

#include "Opengl/AFX/Solid/VolumeTreeNode/SGDLTools.h"
#include "../z2_coo.h"

//*******************************************************************
//  Volume_Tree_Matrix constructor 
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
Volume_Tree_Matrix::Volume_Tree_Matrix(Volume_Tree_Node *The_Node,
						               const Matrix &the_transfo)
									   : Volume_Tree_Node(SGDL_VT_MATRIX)
{
	assert ( The_Node != NULL );
    v = The_Node;
    The_Matrix = the_transfo;

    this->Minimize(The_Node);
	SGDL_matInv(&The_Matrix_Inverted[0], &The_Matrix[0]);
	Manage_At_The_Top();    
}

//*******************************************************************
//  Volume_Tree_Matrix constructor  
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
Volume_Tree_Matrix::Volume_Tree_Matrix(Volume_Tree_Node *The_Node,
				   const float bbox_ptr[],
				   const Matrix &the_transfo):
Volume_Tree_Node(SGDL_VT_MATRIX, bbox_ptr)
{
	assert ( The_Node != NULL );
    v = The_Node;
    The_Matrix = the_transfo;

    this->Minimize(The_Node);

	SGDL_matInv(&The_Matrix_Inverted[0], &The_Matrix[0]);

	Manage_At_The_Top();
}

//*******************************************************************
//  This function compress two or more VolumeTreeMatrix in a row into
//  only one node with the product of all the matrix.
//
//  By Simon Lavallee
//  SGDL Systems inc
//  Date : October 7th, 2002.
//*******************************************************************
void Volume_Tree_Matrix::Minimize(Volume_Tree_Node *The_Node)
{
    Volume_Tree_Matrix *TriedVTMatrix =
		The_Node->getId()==SGDL_VT_MATRIX ? (Volume_Tree_Matrix*)The_Node : 0;

    while((TriedVTMatrix != NULL) && (TriedVTMatrix != this))
    {
        The_Matrix *= TriedVTMatrix->Get_Matrix();
        v = TriedVTMatrix->v;
        delete TriedVTMatrix;
        TriedVTMatrix = v->getId()==SGDL_VT_MATRIX ? (Volume_Tree_Matrix*)v : 0;
    }    
}

//*******************************************************************
//  This function returns the density to the point x, y, z
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
int Volume_Tree_Matrix::EvaluateDensity(const double& x, const double& y, const double& z)const
{
	double new_Pt_x = x;
	double new_Pt_y = y;
	double new_Pt_z = z;

	int to_return = 0;
	if ( !SGDL_isIdentity(&The_Matrix[0]) )	{
	    Multiply_Mat_Double(The_Matrix_Inverted, x, y, z, 
                            &new_Pt_x, &new_Pt_y, &new_Pt_z);
	}

	to_return = v->EvaluateDensity(new_Pt_x, new_Pt_y, new_Pt_z);

	if ( At_The_Top )
		Bounding_Box_Correction(bbox, &to_return, new_Pt_x, new_Pt_y, new_Pt_z);

	return to_return;
}

//***************************************************
//	
//  Volume_Tree_Matrix::EvaluateGradient()
//
//	Author : Simon Lavallee, SGDL Systems
//	Date   : March 27th, 2002
//
//  Input  : const double &x   -> X coordinate
//           const double &y   -> Y coordinate
//           const double &z   -> Z coordinate
//
//  Output : double &Distance  -> Distance between the point 
//                                and the surface
//           float  *Color     -> Color of the point
//           double *Normal    -> Normal of the point
//           bool              -> If we've been able to get the normal
//
//  Desc   : This function evaluate the gradient of its volume and 
//           modify it, depending if the matrix has rotation.
//***************************************************
bool Volume_Tree_Matrix::EvaluateGradient(const double& x,  const double& y, const double& z, 
                                          double& Distance, double *Normal, float *Color) const
{
    //Get a copy of the points
    double new_Pt_x = x;
	double new_Pt_y = y;
	double new_Pt_z = z;

    //Flag to know if the normal calculation went well
    bool   NormalIsValid = false;
	if ( !SGDL_isIdentity(&The_Matrix[0]) ) {
        //If the matrix isn't identity, multiply the point by
        //the inverted matrix since the function will be evaluated
        //as if it has no transformation matrix.
		Multiply_Mat_Double(The_Matrix_Inverted, x,	y, z,
							&new_Pt_x, &new_Pt_y, &new_Pt_z);
	}
    //Get the normal at this transformed point
    NormalIsValid = v->EvaluateGradient(new_Pt_x, new_Pt_y, new_Pt_z, 
                                        Distance, Normal, Color);

	// apply matrix on normal
	float norm[4];
	norm[0] = Normal[0];
	norm[1] = Normal[1];
	norm[2] = Normal[2];
	norm[3] = 0;
	SGDL_z__pro(norm, norm, &The_Matrix[0]);
	float length = sqrt(norm[0]*norm[0] + norm[1]*norm[1] + norm[2]*norm[2]);
	Normal[0] = norm[0]/length;
	Normal[1] = norm[1]/length;
	Normal[2] = norm[2]/length;

    if (At_The_Top && NormalIsValid)
    {
        //The bounding box can modify the field value and the gradient
        Bounding_Box_Gradient_Correction(bbox, Normal, &Distance, x, y, z);        
    }

    if (Color && !Undefined_Color(color) && NormalIsValid)
    {
        Color[0] = color[0];
        Color[1] = color[1];
        Color[2] = color[2];
    }

	return NormalIsValid;
}

//*******************************************************************
//  This function returns the coeffs of the Volume_Tree_Matrix's VTN
//
//  By Eric Baillargeon
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
double * Volume_Tree_Matrix::GetCoeff(double x, double y , double z)
{
    return v->GetCoeff(x,y,z);
}

//*******************************************************************
//  This function returns the vumber of coeffs of the Volume_Tree_Matrix's VTN 
//
//  By Eric Baillargeon
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
int Volume_Tree_Matrix::GetNumOfCoeff(double x, double y , double z)
{
    return v->GetNumOfCoeff(x,y,z);
}

//*******************************************************************
//  This function set the color of a given point, based on the fact 
//  that is in or outside the volume.
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//  Modified : Simon Lavallee (April 15th 2002)
//*******************************************************************
bool Volume_Tree_Matrix::GetColor(const double &x, const double &y, const double &z, float set_color[])const
{  
    double new_Pt_x = x;
    double new_Pt_y = y;
    double new_Pt_z = z;

	if ( !SGDL_isIdentity(&The_Matrix[0]) ) {		
		//Transform the point if the matrix isn't identity
        Multiply_Mat_Double(The_Matrix_Inverted, x, y, z,
                            &new_Pt_x, &new_Pt_y, &new_Pt_z);
    }    

	bool inside_volume = v->GetColor(new_Pt_x, new_Pt_y, new_Pt_z, set_color);

    if (!Undefined_Color(color) && (inside_volume || At_The_Top))
    {
        //If this node has his own color, keep it.
		set_color[0] = color[0];
		set_color[1] = color[1];
		set_color[2] = color[2];
    }

    return inside_volume;
}

//*******************************************************************
//  This function manages At_The_Top
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
void Volume_Tree_Matrix::Manage_At_The_Top()
{
	v->At_The_Top = false;
	At_The_Top = true;
}

//*******************************************************************
//  This function adds a reference to a Volume_Tree_Matrix
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
void Volume_Tree_Matrix::Ref()
{
    ref++;

    if (v != 0)   //Ref only if v exist (SL SGDL-sys 30-01-02) TBM
       v->Ref();
}

//*******************************************************************
//  This function removes a reference to a Volume_Tree_Matrix
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
void Volume_Tree_Matrix::UnRef()
{
    ref--;

    if (v != 0)   //UnRef only if v exist (SL SGDL-sys 30-01-02) TBM
       v->UnRef();

    if ( ref == 0 )
       delete this;
}

void Volume_Tree_Matrix::Set_Matrix(const Matrix& the_other)
{
   The_Matrix = the_other;
   SGDL_matInv(&The_Matrix_Inverted[0], &The_Matrix[0]);
}
