/*
	This software module was originally developed by SGDL Systems Inc. in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1)
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license
to this software module or modifications thereof for use in hardware or software
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1).
Those intending to use this software module in hardware or software products are
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability for use of
this software module or modifications thereof in an implementation. Copyright is
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.
	SGDL Systems Inc. retains full right to use the code for his/her own purpose, assign or
donate the code to a third party and to inhibit third parties from using the code
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
						Copyright  2001
*/

//*******************************************************************
//  Volume_Tree_Unary_Op
//*******************************************************************

#include "Opengl/AFX/Solid/VolumeTreeNode/SGDLTools.h"

//*******************************************************************
//  Volume_Tree_Unary_Op constructor 
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
Volume_Tree_Unary_Op::Volume_Tree_Unary_Op(SGDLUnaryOperator the_Op,
										   const Volume_Tree_Node *To_Be_V,
                                           const float bbox_ptr[])
	: Volume_Tree_Node(SGDL_VT_UNOP, bbox_ptr)
{
	assert(To_Be_V != NULL);

	the_Operation = the_Op;
	v = const_cast<Volume_Tree_Node *>(To_Be_V);
	Manage_At_The_Top();
}

//*******************************************************************
//  This function return true(1) if the point x, y, z is inside and
//  false(0) if the point is outside
//
//  Author : Eric Paquet
//  Date : October 4th,2001
//
//  Modifs :
//*******************************************************************
int Volume_Tree_Unary_Op::EvaluateDensity(const double& x, const double& y, const double& z)const
{
    ASSERT(v != NULL);

    int to_return = 0;

    switch(the_Operation)
    {
        case SGDLNot :
	    {
            int dens = v->EvaluateDensity(x, y, z) % 3;
			to_return = 2 - dens;
            break;
	    }
        case SGDLSquareRoot :
        {
            to_return = static_cast<int>(sqrt(v->EvaluateDensity(x, y, z)));
            break;
        }
        case SGDLCubeRoot :
        {
            to_return = static_cast<int>(pow(v->EvaluateDensity(x, y, z), 1 / 3.0));
            break;
        }
        case SGDLEven :
        {
            int v_density = v->EvaluateDensity(x, y, z);
            if ( v_density % 2 == 0 )
                to_return = v_density;
            break;
        }
        case SGDLOdd :
        {
            int v_density = v->EvaluateDensity(x, y, z);
            if ( v_density % 2 != 0 )
                to_return = v_density;
            break;
        }

        default:
	    {
	    assert(false); //Do not know this operation
        }
    }
    
    if ( At_The_Top )
	    //This call may modify to_return
	    Bounding_Box_Correction(bbox, &to_return, x, y, z);	
    
    return to_return;
}

//***************************************************
//	
//  Volume_Tree_Unary_Op::EvaluateGradient()
//
//	Author : Simon Lavallee, SGDL Systems
//	Date   : March 27th, 2002
//
//  Input  : double x        -> X coordinate
//           double y        -> Y coordinate
//           double z        -> Z coordinate
//
//  Output : double Distance -> Distance between the point and the surface
//           double *Normal  -> Normal of the point
//           float  Color    -> Vertex color
//           bool            -> true if we've been able to compute the normal    
//
//  Desc   : This function evaluate the normal of its 
//           volume and modify it, depending on the operator.
//***************************************************
bool Volume_Tree_Unary_Op::EvaluateGradient(const double& x, const double& y, const double& z, 
                                            double& Distance, double *Normal, float *Color) const
{   
    bool NormalIsValid = v->EvaluateGradient(x, y, z, Distance, Normal, Color);

    if (NormalIsValid)
    {
        switch(the_Operation)
        {
            case SGDLNot :
            {
                //Reverse the gradient and the distance
                Normal[0] *= -1;
                Normal[1] *= -1;
                Normal[2] *= -1;

                Distance  *= -1;
                break;
            }

            case SGDLSquareRoot :
            case SGDLCubeRoot :            
            case SGDLEven :
            case SGDLOdd :
            {
                //These operators aren't supposed to modify the gradient
                break;
            }

            default:
            {
                assert(false); //Do not know this operation
            }
        }
     }
    
    if (At_The_Top && NormalIsValid)
    {
        //The gradient can be modified by the bounding box
        Bounding_Box_Gradient_Correction(bbox, Normal, &Distance, x, y, z);
    }

    if (Color && !Undefined_Color(color) && NormalIsValid)
    {
        //If this node has its own color, keep it.
        Color[0] = color[0];
        Color[1] = color[1];
        Color[2] = color[2];
    }

    return NormalIsValid;
}

/********************************************************************\
Function:  

  Input:     
  Output:     
  
    Author:	Eric Baillargeon SGDL Systme Inc.
    Created:	2001/10/26	
    
      
*********************************************************************/
double * Volume_Tree_Unary_Op::GetCoeff(double x, double y , double z)
{
    return (v->GetCoeff(x,y,z));    
}

/********************************************************************\
Function:  

  Input:     
  Output:     
  
    Author:	Eric Baillargeon SGDL Systme Inc.
    Created:	2001/10/26	
    
      
*********************************************************************/
int Volume_Tree_Unary_Op::GetNumOfCoeff(double x, double y , double z)
{
    return (v->GetNumOfCoeff(x,y,z));
}

//*******************************************************************
//  This function manages the operator = between two Volume_Tree_Unary_Op
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
const Volume_Tree_Unary_Op& Volume_Tree_Unary_Op::operator=(const Volume_Tree_Unary_Op& The_Other)
{
	the_Operation = The_Other.the_Operation;
	v = The_Other.v;
	return *this;
}

//*******************************************************************
//  This function return the color of a point based on the fact that is
//  in or ouside the volume.
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//  Modif: Simon Lavallee (04-16-2002) (Now just get the color)
//*******************************************************************
bool Volume_Tree_Unary_Op::GetColor(const double &x, const double &y, const double &z, float set_color[])const
{
	bool inside_volume = v->GetColor(x, y, z, set_color);
	
	if (!Undefined_Color(color) && (inside_volume || At_The_Top))
	{
		set_color[0] = color[0];
		set_color[1] = color[1];
		set_color[2] = color[2];
	}

    return inside_volume;
}

//*******************************************************************
//  This function manages At_The_Top
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
void Volume_Tree_Unary_Op::Manage_At_The_Top()
{
	v->At_The_Top = false;
	At_The_Top = true;
}

//*******************************************************************
//  This adds a reference to a Volume_Tree_Unary_Op
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
void Volume_Tree_Unary_Op::Ref()
{
    ref++;
    v->Ref();
}

//*******************************************************************
//  This removes a reference to a Volume_Tree_Unary_Op
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
void Volume_Tree_Unary_Op::UnRef()
{
    ref--;
    v->UnRef();

    if ( ref == 0 )
        delete this;
}

//EOF***************************************************************************