
//	This software module was originally developed by Massimo Balestri
//	(Telecom Italia Lab S.p.A.), and edited by Claudio Alberti (EPFL) 
//  in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1) 
//  standard. This software module is an implementation of a part of one 
//  or more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the 
//  MPEG-4 Systems (ISO/IEC 14496-1) standard.
//
//	ISO/IEC gives users of the  the MPEG-4 Systems (ISO/IEC 14496-1) free
//	license to this software module or modifications thereof for use in hardware
//	or software products claiming conformance to the MPEG-4 Systems (ISO/IEC
//	14496-1).
//
//	Those intending to use this software module in hardware or software
//	products are advised that its use may infringe existing patents.
//
//	The original developer of this software module and his/her company, the
//	subsequent editors and their companies, and ISO/IEC have no liability for
//	use of this software module or modifications thereof.
//
//	Copyright is not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming
//	products.
//
//	Telecom Italia Lab S.p.A retains full right to use the code for its own purpose,
//	assign or donate the code to a third party and to inhibit third parties from
//	using the code for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.
//
//	This copyright notice must be included in all copies or derivative works.
//
//	Copyright (c) 2000.


// IPMP_Message.cpp

#include "IPMP_Message.h"
#include "IPMP_Data_BaseClass.h"
#include <memory.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

USING_MOSES_NAMESPACE

IPMP_Message::IPMP_Message()			// constructor
{
	init();
}


IPMP_Message::~IPMP_Message()			// destructor
{
	cleanUp();
}

IPMP_Message::IPMP_Message(const IPMP_Message & message)	// copy constructor
	:ExpandableBaseClass(message)
{
	init();
	switch (message.getType())
	{
		case HOOKS_TYPE:
		{
			bool b = setHooksType(
				message.getIPMPS_Type(),
				message.getIPMP_data(),
				message.getSizeOfData());
			assert(b);
			break;
		}
		case URL_TYPE:
		{
			bool b = setUrlType( message.getURLString());
			assert(b);
			break;		
		}
		case IPMPX_TYPE:
		{
			bool b = setIPMPXType(message.getIPMP_DescriptorID());
			assert(b);
			
			for(int x=0;x < message.getSizeOfData(); x++ )
			{
				bool b = addIPMPXMessage( *message.getIPMPXMessage(x) );
				assert(b);
			}

			break;
		}
	}	// end switch
}

void IPMP_Message::init()
{
	messageType = IPMPX_TYPE;					// default is IPMPX, not hooks
	IPMP_DescriptorID = 0 ;
	IPMPS_Type = 0;	
	IPMPX_data = new IPMP_Data_BaseClass * [MAX_IPMPX_DATA_MESSAGES] ;
	
	for (int i=0; i<MAX_IPMPX_DATA_MESSAGES; i++)
		IPMPX_data[i] = NULL;

	URLString = NULL ;
	IPMP_data = NULL ;
	sizeOfData = 0 ;
}

void IPMP_Message::cleanUp() 
{
	if (URLString != NULL) 
		delete[] URLString ;

	if (IPMP_data != NULL)
		delete[] IPMP_data ;
	
	if (IPMPX_data != NULL)
	{
		for (unsigned int i=0;i<sizeOfData;i++)
			delete IPMPX_data[i];

		delete[] IPMPX_data ;
	}
}


IPMP_Message::MessageTypeEnum IPMP_Message::getType() const
{
	return messageType;
}


bool IPMP_Message::setUrlType(const char *URL)
{
	if (URL == NULL)				// check if the URL exists
		return false;

	size_t size = strlen(URL) + 1;
	if (size > 2048)				// check if the URL size is reasonable
		return false;				// return without changing anything

	messageType = URL_TYPE;
	IPMPS_Type = 0x0000;

	if( URLString )
		delete[] URLString ;

	URLString = new char[ size ] ;
	strncpy( URLString, URL, size ) ;
	return true;
}

bool IPMP_Message::setIPMPXType(unsigned short descrId)
{
	if ((descrId == 0) || (descrId == 0xFFFF))
		return false;								// forbidden

	messageType = IPMPX_TYPE;
	IPMPS_Type = 0x0001;
	IPMP_DescriptorID = descrId;
	return true;
}

bool IPMP_Message::setHooksType(unsigned short mytype, byte * ipmpdata, int ipmpdatasize)
{
	// check param values
	if (mytype < 0x2000)
		return false;				// cannot be zero (=URL) or reserved values (1..0x2000)

	messageType = HOOKS_TYPE;
	IPMPS_Type = mytype;

	if( IPMP_data )
		delete[] IPMP_data ;
	IPMP_data = new byte[ ipmpdatasize ] ;
	memcpy( IPMP_data, ipmpdata, ipmpdatasize) ;	
	sizeOfData = ipmpdatasize;
	return true;
}

// copy the message into the local storage
bool IPMP_Message::addIPMPXMessage(IPMP_Data_BaseClass &msg )
{
	if (messageType != IPMPX_TYPE)
		return false;

	if( sizeOfData < MAX_IPMPX_DATA_MESSAGES )
	{
		IPMPX_data[ sizeOfData++ ] = IPMP_Data_BaseClass::createMessage(msg);		// copy the message
		return true ;
	}
	else
		return false ;
}

byte * IPMP_Message::getIPMP_data() const
{
	return IPMP_data;
}

char * IPMP_Message::getURLString( ) const
{
	return URLString;
}

int IPMP_Message::getSizeOfData() const
{
	return sizeOfData ;
}

unsigned short IPMP_Message::getIPMP_DescriptorID() const
{
	return (unsigned short) IPMP_DescriptorID ;	
}

unsigned short IPMP_Message::getIPMPS_Type( ) const
{
	return (unsigned short) IPMPS_Type ;	
}


IPMP_Data_BaseClass* IPMP_Message::getIPMPXMessage( unsigned int nIndex ) const
{
	if (messageType != IPMPX_TYPE)
		return NULL;

	if( nIndex < sizeOfData )
	{
		return IPMPX_data[ nIndex ] ;
	}
	else
		return NULL;
}


int IPMP_Message::fromBuffer(const byte * buf)
{
	ExpandableBaseClass::fromBuffer(buf);		// read the base descriptor and set "buf" as read buffer for this class.
	cleanUp() ;	
	init() ;

	IPMPS_Type = get(16 ) ;								// 16 bits

	if (IPMPS_Type == 0)
	{
		messageType = URL_TYPE;
		sizeOfData = getSizeOfInstance() - 2;
		URLString = new char[sizeOfData];
		getBuffer(URLString, sizeOfData * 8);
	}
	else if (IPMPS_Type == 1)
	{
		messageType = IPMPX_TYPE;
		IPMP_DescriptorID = get(16);					// 16 bits
		sizeOfData=0;
		int buffersize = getSizeOfInstance() - 2;
		while( (getReadCounter() / 8) < buffersize )
		{
			unsigned int nbit;
			IPMPX_data[sizeOfData++] = IPMP_Data_BaseClass::createMessage(getReadPtr(), nbit);
			skipReading( nbit );
		}
	}
	else
	{
		messageType = HOOKS_TYPE;
		sizeOfData = getSizeOfInstance() - 2;
		IPMP_data = new byte[sizeOfData];
		getBuffer(IPMP_data,  sizeOfData * 8);
	}
	return getReadCounter();				// return the number of read bits	
}

int IPMP_Message::toBuffer(byte * buf)
{
	ExpandableBaseClass::toBuffer(buf);		// read the base descriptor and set "buf" as read buffer for this class.

	put( IPMPS_Type, 16 ) ;						// 16 bits

	if (IPMPS_Type == 0)
	{
		putBuffer(URLString, 8 * (strlen(URLString) + 1));
	} 
	else if (IPMPS_Type == 0x0001)
	{
		put( IPMP_DescriptorID, 16 ) ;				// 16 bits
		
		for( unsigned int x=0;x<sizeOfData;x++ )
		{
			skipWriting( IPMPX_data[x]->toBuffer( getWritePtr() )) ;
		}
	} 
	else
	{
		putBuffer(IPMP_data, 8 * sizeOfData);
	}
	return getWriteCounter();				// return the number of read bits
}

unsigned int IPMP_Message::sizeOfInstance() const
{
	unsigned int size = 2 ;				// IPMPS_Type

	if (IPMPS_Type == 0)
	{
		size += (strlen(URLString) + 1);
	}
	else if (IPMPS_Type == 1)
	{
		size += 2 ;				// IPMP_DescriptorID
		
		for( unsigned int x=0;x<sizeOfData;x++ )
		{
			size += IPMPX_data[x]->getSize();
		}
	}
	else
	{
		size += sizeOfData ;
	}

	return size;
}


































//======================= OLD METHODS ===========================
#ifdef __REMOVED__
void IPMP_Message::setURLString(byte* str)
{
	if (URLString != NULL)
		delete [] URLString;
	URLString = new byte[sizeOfMsg];
	memcpy(URLString, str, sizeOfMsg);
}


void IPMP_Message::getURLString(byte* str)
{
	memcpy(str, URLString, sizeOfMsg);
}


void IPMP_Message::setIPMP_data(byte* data)
{
	if (IPMP_data != NULL)
		delete [] IPMP_data;
	IPMP_data = new byte[sizeOfMsg];
	memcpy (IPMP_data, data, sizeOfMsg);
}


bool IPMP_Message::getIPMP_data(byte* data)
{
	if (IPMP_data == NULL)
		return false;
	memcpy (data, IPMP_data, sizeOfMsg);
	return true;
}


void IPMP_Message::setIPMP_ExtendedData(int i, byte tag)
{
	if(IPMP_ExtendedData[i].message)
		delete IPMP_ExtendedData[i].message;
	IPMP_ExtendedData[i].message = IPMP_Data_BaseClass::createMessage(tag);
}


IPMP_Data_BaseClass *IPMP_Message::getIPMP_ExtendedData(int i)
{
	return IPMP_ExtendedData[i].message;
}

void IPMP_Message::setExtendedDataSize(unsigned int n)
{
	extendedDataSize = n;
}

unsigned int IPMP_Message::calculateNumOfMessages(const byte *buf, unsigned int size, unsigned long bitindex)
{	
	unsigned int numOfMessages = 0;
	IPMP_Data_BaseClassList tempMessageList[MAX_TRANSPORTED_MESSAGES];
	IPMP_Data_BaseClass *tempMessage = NULL;

	while( (bitindex / 8) < size )
		{
			tempMessageList[numOfMessages].message = IPMP_Data_BaseClass::createMessage(buf + bitindex/8);
			bitindex += (tempMessageList[numOfMessages++].message->getSize() * 8);
		}

	return numOfMessages;
}


#endif // __REMOVED__