
//	This software module was originally developed by Massimo Balestri (TILAB)
//	in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1) standard. 
//	This software module is an implementation of a part of one or more MPEG-4 
//	Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 
//	14496-1) standard. 
//
//	ISO/IEC gives users of the  the MPEG-4 Systems (ISO/IEC 14496-1) free 
//	license to this software module or modifications thereof for use in hardware 
//	or software products claiming conformance to the MPEG-4 Systems (ISO/IEC 
//	14496-1). 
//
//	Those intending to use this software module in hardware or software 
//	products are advised that its use may infringe existing patents. 
//
//	The original developer of this software module and his/her company, the 
//	subsequent editors and their companies, and ISO/IEC have no liability for 
//	use of this software module or modifications thereof. 
//
//	Copyright is not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming 
//	products. 
//
//	Telecom Italia Lab S.p.A. retains full right to use the code for its own purpose, 
//	assign or donate the code to a third party and to inhibit third parties from 
//	using the code for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.
// 
//	This copyright notice must be included in all copies or derivative works. 
//
//	Copyright (c) 2002.

// IPMP_Tool.h

#ifndef	IPMP_TOOL_H
#define	IPMP_TOOL_H
#include "namespace.h"
#include "BaseDescriptor.h"
#include "IPMP_ParametricDescription.h"
#include "IPMP_ByteArray.h"

BEGIN_MOSES_NAMESPACE

/** 
 *  @class IPMP_Tool
 *  This class identifies one IPMP Tool that is required by the Terminal to consume the content.
 *  The IPMP Toolcan be specified either as a unique implementation, as one of a list of alternatives, or through
 *  a parametric description.
 *  
 *  @author Massimo Balestri
 *  @date 2002
 */
class IPMP_Tool: public BaseDescriptor 
{
private:
	byte *IPMP_ToolID;					// the 128 bit id of the Tool	
	unsigned int isAltGroup ;
	unsigned int isParametric ;

	unsigned int numAlternates ;
	byte ** specificToolID;
	IPMP_ParametricDescription toolParamDesc;
	Container<IPMP_ByteArray>  toolURL ;

	void init() ;						// local initialization 

public:
	/**
	 * builds a copy of the same object
	 * @param tool the object to clone
	 */
	IPMP_Tool(const IPMP_Tool &tool);

	/**
	 *  set the 182 bit ID of the IPMP Tool described by this descriptor.
	 *  @param toolId a 128 bit identifier of the IPMP Tool.
	 */
	void setIPMP_ToolID( const byte * toolId);

	/** 
	 *  get the 182 bit ID of the IPMP Tool described by this descriptor.
	 *  @return a pointer to a 128 bit identifier of the IPMP Tool.
	 */
	byte * getIPMP_ToolID() const;

	/**
	 * set the Tool ID of one alternate tool (a list of alternate tools can be 
	 * provided in this message).
	 * @param toolId the 128 bit id of this alternate tool
	 * @param index the index of this tool in the list of alternate tools
	 */
	void setSpecificToolID(const byte *toolId, unsigned int index);


	/**
	 * get the Tool ID of one alternate tool (a list of alternate tools can be 
	 * provided in this message).
	 * @param index the index of this tool in the list of alternate tools
	 * @return a poiter to the 128 bit tool ID
	 */
	byte * getSpecificToolID(unsigned int index) const;
	
	/**
	 * set or reset the flag indicating if alternative tools exist for this tool.
	 * @param alt the value of the flag
	 */
	void setIsAltGroup(bool alt);

	/**
	 * get the flag indicating if alternative tools exist for this tool.
	 * @return the value of the flag. TRUE indicates that alternative tools exist.
	 */
	bool getIsAltGroup() const;

	/**
	 * set or reset the flag indicating if the tool is a parametric tool.
	 * @param parametric the value of the flag
	 */
	void setIsParametric(bool parametric);

	/**
	 * get the flag indicating if the tool is a parametric tool.
	 * @return the value of the flag. TRUE indicates that the tool is parametric.
	 */
	bool getIsParametric() const;

	/**
	 * set the number of alternate tools.
	 * @param num the number of alternate tools.
	 */
	void setNumAlternates(unsigned int num);

	/**
	 * get the number of alternate tools.
	 * @return the number of alternate tools. 
	 */
	unsigned int getNumAlternates() const;

	/**
	 * get the number of URLs associated to this tool.
	 * @return the number of URLs 
	 */
	unsigned int getNumURLs() const;

	/**
	 * set the parametric description associated to this tool.
	 * @param descr the parametric description
	 */
	void setToolParamDesc(const IPMP_ParametricDescription & descr );

	/**
	 * get the parametric description associated to this tool.
	 * @return the parametric description
	 */
	const IPMP_ParametricDescription & getToolParamDesc() const;

	/**
	 * get the size in bytes of the parametric description associated to this tool.
	 * @return the length of the parametric description
	 */
	unsigned int getToolParamDescSize() const;

	/**
	 * Add one URL to the list of URLs.
	 * This is one informative URL from which the tool specified in the 
	 * message may be obtained.
	 * @param url the URL
	 * @param size the URL length in bytes
	 */
	void addToolURL(byte *url, int size);

	/**
	 * get one of the informative URLs from which the tool specified in the 
	 * message may be obtained.
	 * @param index the position of this URL in the list of URLs
	 * @return a pointer to the URL
	 */
	byte * getToolURL(unsigned int index) const;
	
	/**
	 * get the size of one of the informative URLs from which the tool specified in the 
	 * message may be obtained.
	 * @param index the position of this URL in the list of URLs
	 * @return the URL length in bytes
	 */
	unsigned int getToolURLsize(unsigned int index) const;
	
	/**
	 * constructor.
	 */
	IPMP_Tool();	

	/**
	 * destructor.
	 */
	virtual ~IPMP_Tool();		

	/**
	 * method for writing the message using the data of the class.
	 * @param buf storage location for the byte sequence. 
	 * @return number of bit written.
	 */
	virtual int toBuffer(byte * buf);	
	
	/**
	 * method for reading the contents of the message.
	 * @param buf message storage location for the byte sequence. 
	 * @return number of bit read.
	 */
	virtual int fromBuffer(const byte * buf);

	/**
	 * get the total size of the class,when converted into a bitstream. 
	 * This size does NOT include the object_id and the length of the base class. 
	 * @return the size in bytes. 
	 */
	virtual unsigned int sizeOfInstance() const;
};


END_MOSES_NAMESPACE

#endif
