
//	This software module was originally developed by ADETTI in the course of
//	development of the MPEG-4 Systems (ISO/IEC 14496-1) standard. 
//	This software module is an implementation of a part of one or more MPEG-4 
//	Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 
//	14496-1) standard. 
//
//	ISO/IEC gives users of the  the MPEG-4 Systems (ISO/IEC 14496-1) free 
//	license to this software module or modifications thereof for use in hardware 
//	or software products claiming conformance to the MPEG-4 Systems (ISO/IEC 
//	14496-1). 
//
//	Those intending to use this software module in hardware or software 
//	products are advised that its use may infringe existing patents. 
//
//	The original developer of this software module and his/her company, the 
//	subsequent editors and their companies, and ISO/IEC have no liability for 
//	use of this software module or modifications thereof. 
//
//	Copyright is not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming 
//	products. 
//
//	ADETTI retains full right to use the code for its own purpose, 
//	assign or donate the code to a third party and to inhibit third parties from 
//	using the code for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.
// 
//	This copyright notice must be included in all copies or derivative works. 
//
//	Copyright (c) 2002.


// IPMP_UserQueryResponse.h

#ifndef	IPMP_USERQUERYRESPONSE_H
#define	IPMP_USERQUERYRESPONSE_H

#include "IPMP_Data_BaseClass.h"
#include "Container.h"
#include "namespace.h"

BEGIN_MOSES_NAMESPACE

class RTArray;

/** @class IPMP_UserQueryResponse
 *  Carries an user's answer to an IPMP_UserQuery.
 *
 *  @author ADETTI
 *  @date 2002
 */
class IPMP_UserQueryResponse: public IPMP_Data_BaseClass 
{
private:
	bool inclReplyText;
	bool inclOptionSelect;
	word32 languageCode;
	word32 numOfOptions;
	bool * pOptionResults;	// if inclOptionSelect

	virtual void clear();	// clear all internal data

public:
	Container<RTArray> replyText;		///< list of reply text

	/**
	 * Constructor.
	 */
	IPMP_UserQueryResponse();
	
	/**
	 * Destructor.
	 */
	virtual ~IPMP_UserQueryResponse();
	
	virtual int toBuffer(byte * buf);		
	virtual int fromBuffer(const byte * buf);
	virtual unsigned int sizeOfInstance() const;

	/**
	 * Set the language code for this IPMP_UserQueryResponse instance.
	 * Language code is a 3 character value according to ISO 639-2:1998.
	 * @param languageCode 3 character language code
	 */
	void setLanguageCode(word32 languageCode);
	/**
	 * Clear the language code for this IPMP_UserQueryResponse instance.
	 */
	void clearLanguageCode();
	/**
	 * Get the language code for this IPMP_UserQueryResponse instance.
	 * Language code is a 3 character value according to ISO 639-2:1998.
	 * @return 3 character language code
	 */
	word32 getLanguageCode() const;

	/**
	 * Check if reply text information is available.
	 * @return <i>true</i> if available
	 */
	bool hasReplyText() const;

	/**
	 * Set the array of option results.
	 * @param pOptionResults array with option results
	 * @param numOfOptions length of the pOptionResult array
	 */
	void setOptionResults(bool * pOptionResults, word32 numOfOptions);
	
	/**
	 * Clear the array of option results.
	 */
	void clearOptionResults();
	
	/**
	 * Check if option results information is available.
	 * @return <i>true</i> if available
	 */
	bool hasOptionResults() const;
	
	/**
	 * Get the array of option results.
	 * First call #hasOptionResults to know if this information is available.
	 * @return the array of option results
	 */
	bool * getOptionResults() const;
	
	/**
	 * Get the length of the array of option results.
	 * First call #hasOptionResults to know if this information is available.
	 * @return the length of the array of option results
	 */
	word32 getNumOfOptions() const;
};

/** @class RTArray
 *  Carries an user's answer.
 *
 *  @author ADETTI
 *  @date 2002
 */
class RTArray : public BitField
{
private:
	word32 id;
	word32 subId;
	byte * promptText;
	word32 promptTextSize;

	// clear all internal data
	virtual void clear();

public:
	/**
	 * Constructor.
	 */
	RTArray();
	/**
	 * Destructor.
	 */
	~RTArray();

	/**
	 * Copy constructor.
	 */
	RTArray(const RTArray &);

	/**
	 * Method for writing the message using the data of the class.
	 * @param buf storage location for the byte sequence. 
	 */
	virtual int toBuffer(byte * buf);		
	/**
	 * Method for reading the contents of the message.
	 * @param buf storage location of the byte sequence. 
	 */
	virtual int fromBuffer(const byte * buf);

	/**
	 * Checks that the syntax is correct. This can be used before sending
	 * the message through an interface.
	 * @return <i>true</i> if successful
	 */
	virtual bool checkSyntax() const;

	/**
	 * Get the size in bytes of the binary message generated by the
	 * toBuffer method.
	 * @return the size (in bytes) of this class.
	 */
	virtual unsigned int getSize() const;

	/**
	 * Set the 16 bit serial number to be associated with the contained
	 * data or to associate contained data with other data.
	 * @param id a 16 bit identifier
	 */
	void setId(word32 id);
	/**
	 * Clear (i.e. set to 0) the 16 bit identifier.
	 */
	void clearId();
	/**
	 * Get the 16 bit serial number associated with the contained
	 * data or to associating contained data with other data.
	 * @return the 16 bit identifier.
	 */
	word32 getId() const;

	/**
	 * Set the 16 bit serial number to associate items within an
	 * id group.
	 * @param subId a 16 bit sub-identifier
	 */
	void setSubId(word32 subId);
	/**
	 * Clear (i.e. set to 0) the 16 bit sub-identifier.
	 */
	void clearSubId();
	/**
	 * Get the 16 bit serial number that associates items within
	 * the id group.
	 * @return the 16 bit sub-identifier
	 */
	word32 getSubId() const;

	/**
	 * Set the user's answer.
	 * @param promptText user's answer
	 * @param promptTextSize length of the user's answer
	 */
	void setPromptText(byte * promptText, word32 promptTextSize);
	/**
	 * Clear the user's answer.
	 */
	void clearPromptText();
	/**
	 * Get the user's answer.
	 * @return the user's answer
	 */
	byte * getPromptText() const;
	/**
	 * Get the size of the user's answer.
	 * @return the size of the user's answer
	 */
	word32 getPromptTextSize() const;
};

END_MOSES_NAMESPACE

#endif

